/*
 * Copyright (C) 1994, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */
/*------------------------------------------------------------------------------
 *
 * OORT - background.c - Routines for main view background.
 *
 * $Id: background.c,v 1.2 1994/01/28 00:16:23 mtj Exp $
 * 
 * Chris Fouts - April, 1993.
 * 
 *----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <pf.h>
#include <prmath.h>

#include "oort.h"
#include "background.h"

/* BEGIN PROTOTYPES -S background.c */
static void     drawSun( int nSteps ) ;
static void     genStars( int nStars, float r, float theta0, float theta1,
                    float phi0, float phi1 ) ;
static float    randomNumber( float lo, float hi ) ;
/* END PROTOTYPES -S background.c */


/*------------------------------------------------------------------------------
 * Make background object.
 *----------------------------------------------------------------------------*/
Object
makeBackground(
	void
	)
{
	int		i ;
	int		j ;
	int		nPhi = 11 ;
	int		nTheta = 21 ;
	int		nStars = 5000 ;
	int		totalStars = 0 ;
	int		n ;
	void		*arena;
	float		dTheta ;
	float		dPhi ;
	float		phi0 ;
	float		phi1 ;
	float		theta0 ;
	float		theta1 ;
	float		dS ;
	float		S ;
	float		totalS = 0.0f ;
	float		r = 0.5f * ASTEROID_RADIUS ;
	Object		objectNo ;

	dTheta = 360.0f / (float)nTheta ;
	dPhi = 178.0f / (float)nPhi ;

	S = 4.0f * PF_PI ;	/* Total surface area of a unit sphere */

	objectNo = genobj() ;

	makeobj( objectNo ) ;
		cpack( 0x00ffffff ) ;
		zfunction( ZF_ALWAYS ) ;
		zwritemask( 0x0 ) ;
		bgnpoint() ;
		for( i = 0 ; i < nTheta ; i++ )
		{
			theta0 = ( i     ) * dTheta ;
			theta1 = ( i + 1 ) * dTheta ;
			for( j = 0 ; j < nPhi ; j++ )
			{
				phi0 = ( j     ) * dPhi - 89.0f ;
				phi1 = ( j + 1 ) * dPhi - 89.0f ;

				dS = ( sinf( PF_DEG2RAD( phi1 ) ) -
					sinf( PF_DEG2RAD( phi0 ) ) ) *
					PF_DEG2RAD( theta1 - theta0 ) ;

				n = (int)( (float)nStars * dS / S ) ;

				genStars( n, r, theta0, theta1, phi0, phi1 ) ;
				totalStars += n ;
				totalS += dS ;
			}
		}
		endpoint() ;

		/*
		 * Draw sun.
		 */
		cpack( 0x0088ffff ) ;
		drawSun( 4 ) ;
		zfunction( ZF_LEQUAL ) ;
		zwritemask( 0xffffffff ) ;
	closeobj() ;

	return( objectNo ) ;
}



/*------------------------------------------------------------------------------
 * Generate points within a specified area.
 *----------------------------------------------------------------------------*/
static void
genStars(
	int	nStars,
	float	r,
	float	theta0,
	float	theta1,
	float	phi0,
	float	phi1
	)
{
	int	i ;
	pfVec3	star ;
	float	phi ;
	float	theta ;
	float	cp ;
	float	sp ;
	float	ct ;
	float	st ;

	for( i = 0 ; i < nStars ; i++ )
	{

		phi = randomNumber( phi0, phi1 ) ;
		theta = randomNumber( theta0, theta1 ) ;
		pfSinCos( phi, &sp, &cp ) ;
		pfSinCos( theta, &st, &ct ) ;

		PFSET_VEC3( star, r*ct*cp, r*st*cp, r*sp ) ;
		v3f( star ) ;
	}
}



/*------------------------------------------------------------------------------
 * Generate a random number in the range [lo,hi).
 *----------------------------------------------------------------------------*/
static float
randomNumber(
	float	lo,
	float	hi
	)
{
	return( lo + drand48() * ( hi - lo ) ) ;
}



/*------------------------------------------------------------------------------
 * Draw the sun.
 *----------------------------------------------------------------------------*/
static void
drawSun(
	int	nSteps
	)
{
	int	i ;
	float	theta ;
	float	r = 0.005f ;
	float	d = -0.4f ;
	float	c ;
	float	s ;
	pfVec3	p ;

	nSteps++ ;
	bgntmesh() ;
		PFSET_VEC3( p, 0.0f, d, r ) ;
		v3f( p ) ;
		for( i = 1 ; i < nSteps ; i++ )
		{
			theta = 180.0f * (float)i / (float)nSteps ;
			pfSinCos( theta, &s, &c ) ;
			PFSET_VEC3( p,  r*s, d, r*c ) ;
			v3f( p ) ;
			PFSET_VEC3( p, -r*s, d, r*c ) ;
			v3f( p ) ;
		}
		PFSET_VEC3( p, 0.0f, d, -r ) ;
		v3f( p ) ;
	endtmesh() ;
}
