/*
 * Copyright (C) 1994, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */
/*------------------------------------------------------------------------------
 *
 * OORT - oort.h - Global header.
 *
 * $Id: oort.h,v 1.4 1994/01/28 00:20:38 mtj Exp $
 * 
 * Chris Fouts - April, 1993.
 * 
 *----------------------------------------------------------------------------*/

#ifndef OORT_H
#define OORT_H

#ifndef	__EXTENSIONS__
#define	__EXTENSIONS__
#define UN_EXTEN
#endif /* __EXTENSIONS__ */
#include <math.h>
#ifdef UN_EXTEN
#undef	__EXTENSIONS__
#endif /* UN_EXTEN */

#include <Performer/pf.h>
#include "validate.h"

#define	NAMELEN		12
#define	MAX_MSG_SIZE	80
#define	MAX_SPEED	(22.22f)

/*
 * Center of gravity structure.
 */
typedef struct {
	pfVec3	xyz ;	/* location */
	pfVec3	vel ;	/* wheel velocity */
	pfVec3	n ;	/* normal at ground point */
	int	inAir ;	/* indicate if in air */
	} Wheel ;

typedef struct {
        pfVec3  cg ;            /* Center of gravity */
        pfVec3  vel ;           /* Vehicle velocity */
        pfVec4  rot ;      	/* Quaternion of current rotation */
	} Motion ;

typedef struct {
	ngMagic		magic ;
	long		id ;
	unsigned short	key ;
	char		team ;
	char		status ;
	unsigned char	laserStrength ;
	char		name[NAMELEN] ;
	int		score ;
	long		hitId ;
	long		hitById ;
	float		time ;
	pfVec3		xyz ;
	pfVec3		ahead ;
	pfVec3		right ;
	pfVec3		up ;
	} Player ;

typedef struct {
	ngMagic		magic ;
	long		id ;
	unsigned short	key ;
	char		team ;
	char		msg[MAX_MSG_SIZE] ;
	} Notice ;

typedef union {
	ngMagic		magic ;
	Player		player ;
	Notice		notice ;
	struct ngAck	ack ;
	struct ngCensus	census ;
	} GamePacket ;

typedef struct {
	Screencoord	left ;
	Screencoord	right ;
	Screencoord	bottom ;
	Screencoord	top ;
	} ViewPort ;

typedef struct {
	unsigned long	fg ;
	unsigned long	bg ;
	pfVec3		v0 ;
	pfVec3		v1 ;
	pfVec3		v2 ;
	pfVec3		v3 ;
	pfVec3		v4 ;
	pfVec3		v5 ;
	} BarGraph ;


typedef void	(*PacketFunction)( void ) ;


/* ID's */
#define	OORT_PLAYER	0x4f6f5000	/* OoP0 - Oort Player v. 0 */
#define	OORT_ACK	0x4f6f4100	/* OoA0 - Oort acknowledge v. 0 */
#define	OORT_CENSUS	0x4f6f4300	/* OoC0 - Oort census v. 0 */
#define	OORT_DROP	0x4f6f4400	/* OoD0 - Oort drop v. 0 */
#define	OORT_INVALIDATE	0x4f6f4900	/* OoI0 - Oort invalidate v. 0 */
#define	OORT_NOTICE	0x4f6f4e00	/* OoN0 - Oort notice v. 0 */
#define	OORT_QUIT	0x4f6f5100	/* OoQ0 - Oort quit v. 0 */

#define OORT_SERVICE		"sgi-oortnet"
#define	OORT_DEF_PORT		5137

#define	MAX_TEXT_LINES	20
#define	N_TEAMS		4
#define	MAIL_ALL	(0x7f)

#define	FRONT_WHEEL	0
#define	RREAR_WHEEL	1
#define	LREAR_WHEEL	2

#define	SELF		0
#define	ENEMY		1

#define	CAR_WIDTH	2.00f
#define	CAR_LENGTH	2.00f
#define	CAR_HEIGHT	1.00f
#define SHIELD_RADIUS	2.0f	/* Meters */

#define ASTEROID_RADIUS	1115.0f	/* Meters */

#define	SQR_LENGTH_VEC3(x)	((x)[0]*(x)[0]+(x)[1]*(x)[1]+(x)[2]*(x)[2] )
#define	GROUND_MASK	0x01
#define	VEHICLE_MASK	0x02
#define ADDSCALE_VEC3(_dst, _v1, _s2, _v2)   \
    (((_dst)[0] = (_v1)[0] + (_s2)*(_v2)[0]), \
     ((_dst)[1] = (_v1)[1] + (_s2)*(_v2)[1]), \
     ((_dst)[2] = (_v1)[2] + (_s2)*(_v2)[2]))
#define OFFSET_VEC3(_dst, _s2, _v2)   \
    (((_dst)[0] += (_s2)*(_v2)[0]), \
     ((_dst)[1] += (_s2)*(_v2)[1]), \
     ((_dst)[2] += (_s2)*(_v2)[2]))

#define MAIN_VIEW_BOTTOM	(0.4f)

#define	OORT_ST_EXIT		0
#define	OORT_ST_IN_ORBIT	0x001
#define	OORT_ST_ON_GROUND	0x002
#define	OORT_ST_CONFIGURING	0x004
#define	OORT_ST_DEAD		0x008

#define	OORT_ST_SHIELDS		0x010
#define	OORT_ST_EXPLODING	0x020
#define	OORT_ST_CLOAK		0x040
#define	OORT_ST_IDLE_HIT	0x080

#define	SHIELD_ON(p)		( (p)->status & OORT_ST_SHIELDS )
#define	EXPLODING(p)		( (p)->status & OORT_ST_EXPLODING )
#define	CLOAKED(p)		( (p)->status & OORT_ST_CLOAK )

#if !defined( DATA_DIR )
#define	DATA_DIR		"/usr/local/data/oort"
#endif /* !defined( DATA_DIR ) */

/* BEGIN PROTOTYPES oort.c */
int     addPlayer( Player *input ) ;
void    blowUpSelf( int hitId ) ;
void    checkId( int n ) ;
void    copyPlayer( int k, Player *psrc ) ;
void    deletePlayer( int i, char *verb ) ;
void    endProgram( int status ) ;
char *  findFile( char *file ) ;
int     findPlayer( long id ) ;
float   getHeading( void ) ;
void    listScores( void ) ;
void    main( int argc, char **argv ) ;
void *  myMalloc( size_t nbytes ) ;
void    printMat( char *label, pfMatrix m ) ;
void    printVec3( char *label, pfVec3 v ) ;
void    reset( void ) ;
void    setMainDrawFunction( void ) ;
void    shieldDamage( float strength, int id ) ;
float   speedPowerReq( void ) ;
void    toggleCloaking( void ) ;
int     toggleTexturing( void ) ;
/* END PROTOTYPES oort.c */

#define	ABS(x)		( ( (x) > 0 ) ? (x) : -(x) )

#endif /* OORT_H */
