/*
 * Copyright (C) 1994, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */
#include <stdio.h>
#include <gl.h>

#include "creatures.h"

/* 
          Creates and draws the playing chips
*/

extern CREATURE creatures[MAX_PLAYERS];
CHIPS *chips[MAX_PLAYERS][CHIPS_PER_PLAYER];
extern short num_players;

chips_position()
{
  int i, j;
  
  VEC_SET(chips[0][0]->position, -5, 0, 57); 
  VEC_SET(chips[0][1]->position, -1, 0, 57); 
  VEC_SET(chips[0][2]->position, 3, 0, 57); 
  VEC_SET(chips[0][3]->position, 7, 0, 57); 

  VEC_SET(chips[1][0]->position, 51, 0, -21); 
  VEC_SET(chips[1][1]->position, 50, 0, -24); 
  VEC_SET(chips[1][2]->position, 49, 0, -27); 
  VEC_SET(chips[1][3]->position, 48, 0, -30); 

  VEC_SET(chips[2][0]->position, -51, 0, -21); 
  VEC_SET(chips[2][1]->position, -50, 0, -24); 
  VEC_SET(chips[2][2]->position, -49, 0, -27); 
  VEC_SET(chips[2][3]->position, -48, 0, -30); 

  for(i=0; i<MAX_PLAYERS; i++)
    for(j=0; j<4; j++)
      chips[i][j]->status = ON_GROUND;
  
}  
/* this creates a bunch of playing chips */

chips_compute()
{
  VECTOR base[6];
  VECTOR top_point;
  short player,chip;
  short i;
  TRIANGLE current[6];

  for (player = 0; player < MAX_PLAYERS; player++)
    {
      for (chip = 0; chip < CHIPS_PER_PLAYER; chip++)
      {
	/* create the base of the chip */
	create_border(BASE_BOUNDS, 0.0, base);
  
	/* determine the topmost point of the chip */
	VEC_SET(top_point, random_num(-0.2, 0.2), random_num(0.8, 1.2), 
		random_num(-0.2, 0.2));

  
	/* triangulate it */
	chips[player][chip] = (CHIPS *) malloc(sizeof(struct chips));
	chips[player][chip]->status = ON_GROUND; 
	chips[player][chip]->indexi = player;
	chips[player][chip]->indexj = chip;
	chips[player][chip]->yrot = 0.0;
  
#define	current chips[player][chip]->list

	for (i=0; i<5; i++)
	  {
	    VEC_COPY(current[i].v[0], base[i]);
	    VEC_COPY(current[i].v[1], top_point);
	    VEC_COPY(current[i].v[2], base[i+1]);
	    calc_lighting(base[i], base[i+1], top_point, current[i].n, 
			  colors[player], current[i].c);
	  }

	/* last triangle special */
	VEC_COPY(current[i].v[0], base[5]);
	VEC_COPY(current[i].v[1], top_point);
	VEC_COPY(current[i].v[2], base[0]);
	calc_lighting(base[5], base[0], top_point, current[i].n, 
		      colors[player], current[i].c);
      }
    }
  
  chips_position();
  
}


/* draws the playing chips */
chips_draw()
{
  CHIPS *chip;
  
  short player, i;
  for (player = 0; player<num_players; player++)
    {
      if(creatures[player].alive)
      {
	for (i = 0; i<4; i++)
	{
	  chip = chips[player][i];
	  
	  /* move the chip down a little more if dropping */
	  if (chip->status == DROPPING)
	  {
	    chip->position[1] -= 1;
	    chip->position[0] += (chip->final_position[0] - chip->position[0])/3;
	    chip->position[2] += (chip->final_position[2] - chip->position[2])/3;
	    if (chip->position[1] < 
		chip->final_position[1])
	    {
	      VEC_COPY(chip->position, chip->final_position);
	      if (chip->position[1] == -3)
		chip->status = IN_HOLE;
	      else
		chip->status = ON_GROUND;
	    }
	    /* tell the server about the changes */
	    client_update_chip(chip);
	  }
	  
	  draw_chip(chips[player][i]);
	}
      }
    }
}

/* draws a single chip */
draw_chip(CHIPS *chip)
{
  short c[3];
  short i;

  zbuffer(TRUE);
  
  pushmatrix();
  translate(VEC_EXPAND(chip->position));
  rot(chip->yrot, 'y');
  
#define list chip->list

  for (i = 0; i<6; i++)
    {
      /* check if it's the restart chip */
      if (creatures[chip->indexi].done && (chip->status == ON_GROUND)) 
	VEC_COPY(c, crestart);
      else
	VEC_COPY(c, list[i].c);
      add_fog(c, chip->position);
      c3s(c);
      bgnpolygon();
      v3f(list[i].v[0]);
      v3f(list[i].v[1]);
      v3f(list[i].v[2]);
      endpolygon();
    }

  popmatrix();
  zbuffer(FALSE);
}








