/*
 * Copyright (C) 1994, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */
#include <stdio.h>
#include <bstring.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include "net.h"
#include "opcodes.h"
#include "packet.h"
#include "game.h"
#include "client.h"
#include "creatures.h"

static struct sockaddr_in addr;
static NET_SELECT *net_select;
static int fd_client;

static int waiting_for_next_game;
static int joined_game;

extern CREATURE creatures[MAX_PLAYERS];
extern short me;

quit()
{
  PLAYER player;
  
  printf("tagh: (client) Terminated.\n");

  creatures[me].alive = 0;
  update_player_info(&player, me);
  
  player_send_update(fd_client, &addr, &player);
}

  
static void client_read_handler(int fd)
{
  struct sockaddr_in addr;
  SERVER_PACKET server_packet;

  server_packet_recv(fd, &addr, &server_packet);

  switch (server_packet.opcode)
  {
  case OP_GAME_IS_FULL :
    fprintf(stderr, "tagh: Game is already full.\n");
    exit(-1);
    break;
  case OP_GAME_IN_PROGRESS : /* not used */
    printf("tagh: Game is already in progress, waiting for next game...\n");
    net_select_timeout(net_select, 60, 0);
    waiting_for_next_game = 1;
    break;
  case OP_WORLD_UPDATE :
    world_receive_update(&server_packet.world);
    if (joined_game != 1)
      me = server_packet.world.new_player;
    joined_game = 1;
    break;
  }
}

int client_init(struct sockaddr_in* server_addr)
{
  int done;

  bcopy(server_addr, &addr, sizeof(struct sockaddr_in));

  fd_client = net_open(0);
  if (fd_client == NET_ERROR)
    return CLIENT_ERROR;

  net_select = net_select_new();
  if (net_select == NULL)
    return CLIENT_ERROR;

  net_select_timeout(net_select, 1, 0);
  net_select_set(net_select, fd_client, client_read_handler);

  joined_game = 0;
  waiting_for_next_game = 0;

  while (!joined_game) 
  {
    if (!waiting_for_next_game)
      client_opcode_send(fd_client, &addr, OP_JOIN_REQUEST);

    while (net_select_loop(net_select) == NET_ERROR)
      return CLIENT_ERROR;
  }

  net_select_timeout(net_select, 0, 0);

  return CLIENT_OK;
}

void client_send_endgame()
{
  client_opcode_send(fd_client, &addr, OP_END_GAME);
}

void client_poll_network()
{
  net_select_loop(net_select);
}

void client_update_player(PLAYER *player)
{
  player_send_update(fd_client, &addr, player);
}

void client_update_game(int whosit)
{
  player_send_game_update(fd_client, &addr, whosit);
}

void client_update_chip(CHIPS *chip)
{
  CHIP_INFO stuff;

  stuff.indexi = chip->indexi;
  stuff.indexj = chip->indexj;
  VEC_COPY(stuff.position, chip->position);
  VEC_COPY(stuff.final_position, chip->final_position);
  stuff.yrot = chip->yrot;
  stuff.status = chip->status;

  player_send_chip_update(fd_client, &addr, stuff);
}







