/*
 * Copyright (C) 1994, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */
/***
  Contains data structures and definitions related to the creatures 
****/
#include "camera.h"
#include "chips.h"
#include "matrix.h"
#include "defs.h"

/* creature status stuff */
#define C_NORMAL     0
#define C_SWOOP_DOWN 1
#define C_SWOOP_UP   2
#define C_TURN       3

#define C_NOTIT         0
#define C_IT            1
#define C_HIT           2    /* creature tagged another */
#define C_MISS          3    /* creature missed another */

typedef struct wing {
  VECTOR poly1[4], poly2[3];
  COLOR c;
  VECTOR position;
  float p1yrot, p1zrot, p2zrot;
  float direction;
} WING;

typedef struct tail {
  VECTOR poly[4];
  COLOR c;
  VECTOR position;
  float pxrot;
  short direction;
} TAIL;

typedef struct eye {
  TRIANGLE white[3], lid[3], pupil[3];
  float xrot, yrot;
} EYE;

typedef struct beak
{
  TRIANGLE top[2], bottom[2];
  float xrot, yrot;
  MATRIX yrot_mat;
  float top_xrot, bottom_xrot;
} BEAK;

typedef struct creature {
  short alive;  /* am i still playing? */
  short done;   /* true if all my chips are in the hole */
  WING wings[2];
  TAIL tail;
  EYE eyes[2];
  BEAK beak;
  VECTOR position;
  float xrot, yrot, zrot;
  short aimat;  /* index of creature you're aiming at */
  int status;         /* current state of being */
  float speed;        /* how fast the creature is going */
  short holding_chip; /* is the creature holding a chip right now? */
  CHIPS *chip;         /* chip that it's holding or aiming for */
  short it;             /* am i it? */
  float timeit;       /* how long you've been it. can't tag back in first sec*/
} CREATURE;

void   draw_tractor(VECTOR zero, VECTOR one);
