/*
 * Copyright (C) 1994, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */
#include <stdio.h>
#include <stdlib.h>
#include <gl.h>
#include <device.h>
#include <sys/types.h>
#include <unistd.h>

#include "game.h"
#include "cursor.xbm"
#include "defs.h"
#include "display.h"
#include "vector.h"

#define FOV		40.0	/* field of view, in degrees */
#define CLIP_NEAR	0.1	/* near clipping plane */
#define CLIP_FAR	200.0	/* far clipping plane */

long win_id;			/* window ID of main viewport */

long win_size_x;		/* size of main viewport */
long win_size_y;

long scr_size_x;                /* size of actual screen */
long scr_size_y;

long origin_x;			/* origin of main viewport on the screen */
long origin_y;

long mouse_x;			/* current mouse coordinates */
long mouse_y;

unsigned short cursor[16];
float mouse_vector[2] = { 0.0, 0.0 };

extern void creatures_update(float dx, float dy);
extern short me;

/*
     borrowed from pods - creates our cursor 
*/
void cursor_convert(unsigned long *buf, char *curs)
{
  int i, j;
  int t;

  for (i = 31; i >= 0; i--)
  {
    buf[i] = *curs + (*(curs + 1) << 8)
      + (*(curs + 2) << 16) + (*(curs + 3) << 24);
    t = 0;
    for (j = 0; j < 32; j++)
      t |= ((buf[i] >> j) & 1) << (31 - j);
    buf[i] = t;
    curs += 4;
  }
}

/*
	Open the main viewport window.
*/

void display_init()
{
  static unsigned long cursor[32];
  
  scr_size_x = getgdesc(GD_XPMAX);
  scr_size_y = getgdesc(GD_YPMAX);

/*  win_size_x = scr_size_x;
  win_size_y = scr_size_y; */

  foreground();
  prefposition(0, scr_size_x -1, 0, scr_size_y -1);
  win_id = winopen("Tagh");
  getorigin(&origin_x, &origin_y); 
  getsize(&win_size_x, &win_size_y);
  
  mouse_x = getvaluator(MOUSEX) - origin_x;
  mouse_y = getvaluator(MOUSEY) - origin_y;
  
  doublebuffer();
  RGBmode();
  gconfig();
  
  swapinterval(2);
  
  shademodel(GOURAUD);
  subpixel(TRUE);
  glcompat(GLC_OLDPOLYGON, 0);

  czclear(BACKGROUND_COLOR, getgdesc(GD_ZMAX));
  
  qdevice(REDRAW);
  qdevice(MOUSEX);
  qdevice(MOUSEY);
  qdevice(LEFTMOUSE);
  qdevice(ESCKEY);
  
  /* make a cross-hair cursor */
  curstype(C32X1);
  cursor_convert(cursor, cursor_bits);
  defcursor(1, (unsigned short *) cursor);
  curorigin(1, cursor_x_hot, cursor_height - cursor_y_hot + 1);
  setcursor(1, 0, 0);
  drawmode(CURSORDRAW);
  mapcolor(1, 0, 0, 0);
  drawmode(NORMALDRAW);
  
  setvaluator(MOUSEX, origin_x + win_size_x/2, 0, getgdesc(GD_XPMAX));
  setvaluator(MOUSEY, origin_y + win_size_y/2, 0, getgdesc(GD_YPMAX));
  
  
  reshapeviewport();
  mmode(MVIEWING);
  perspective((int) (FOV*10), (1.0*win_size_x)/win_size_y,
	      CLIP_NEAR, CLIP_FAR);
  
  srand48(getpid()); /* seed for our random numbers */
}

/*
	Display the next frame of animation.
*/

void display_redraw()
{
    czclear(BACKGROUND_COLOR, getgdesc(GD_ZMAX));

    wings_update();
    creatures_update(mouse_vector[0], mouse_vector[1]);
    camera_aim();

    /* don't stuff this full of crap, just make
       references to other functions in other files */

    field_draw();		/* draw playing field */
    posts_draw();               /* draw goalposts */
    chips_draw();               /* draw playing chips */
    creatures_draw();           /* draw the players */

    
    swapbuffers();
}

/*
	The event loop.
*/


void adjust_mouse_vector()
{
  int mouse_x;
  int mouse_y;
  float v[2];
  float len;

  getorigin(&origin_x, &origin_y);
  getsize(&win_size_x, &win_size_y);
  
  mouse_x = getvaluator(MOUSEX) - origin_x;
  mouse_y = getvaluator(MOUSEY) - origin_y;
  
  v[0] = mouse_x - (origin_x + win_size_x/2);
  v[1] = mouse_y - (origin_y + win_size_y/2);
  
  if (v[0] != 0 || v[1] != 0) {
    mouse_vector[0] = v[0]/(1.0*win_size_x);
    mouse_vector[1] = v[1]/(1.0*win_size_y);
  } 
}


void display_loop()
{
  short val;
  int done;
  int moved;
  float dx, dy;
  long old;
  PLAYER player;

  
  done = 0;
  while (!done) { 
    update_player_info(&player, me);
    client_update_player(&player);   /* send new info to the server */
    client_poll_network();   /* get new info from the server */

    /* camera_update(); */
    display_redraw();
    adjust_mouse_vector();
    creatures_move();
    update_player_info(&player, me);
    client_update_player(&player);   /* send new info to the server */

    moved = 0;
    while (qtest()) {
      switch (qread(&val)) {
      case REDRAW :
	getorigin(&origin_x, &origin_y);
	getsize(&win_size_x, &win_size_y);
	break;
      case MOUSEX :
	mouse_x = val - origin_x;
	break;
      case MOUSEY :
	mouse_y = val - origin_y;
	break;
      case LEFTMOUSE:
	if (val == TRUE)
	{	  
	  creature_swoop(); /* this function also tags if you're it */
	}
	break;
      case ESCKEY :
	done = TRUE;
	break;
      }

    }
    
  }

  quit();
}














