/*
 * Copyright (C) 1994, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */
#include <stdio.h>
#include <gl.h>

#include "vector.h"
#include "defs.h"
#include "camera.h"
#include "creatures.h"

static VECTOR light = {0.0, 0.0, -1.0};
extern CREATURE creatures[MAX_PLAYERS];
extern short me;

/* this calculates a normal vector and changes the base color to a lighted
 * color given two vectors on a polygon and the point between them */

void calc_lighting(VECTOR v1,
		   VECTOR v2,
		   VECTOR o,
		   VECTOR normal,
		   short cstart[3],
		   short cnew[3])
{
  float index;
  VECTOR l;
  VECTOR w1, w2;

  VEC_SUB(l, light, o);
  VEC_NORM(l);

  VEC_SUB(w1, v1, o);
  VEC_SUB(w2, v2, o);

  VEC_CROSS(normal, w2, w1);
  VEC_NORM(normal);
  index = VEC_DOT(normal,l);
  VEC_COPY(cnew, cstart);
  if (index > 0) 
    VEC_MULT(cnew, index);
}

/*
	adds fog to the color 'c'.
	'v' is a vertex on the polygon, used to find the
	distance to the camera (so we know how much
	fog to add
*/
void add_fog(short c[3], VECTOR x)
{
  short d[3];
  float len;
  float f;
  VECTOR v;

  VEC_SUB(v, camera_from, x);
  len = VEC_LEN(v);
  VEC_SET(d, BACKGROUND_COLOR & 0xff,
	  (BACKGROUND_COLOR >> 8) & 0xff,
	  (BACKGROUND_COLOR >> 16) & 0xff);
  len -= 35.0;
  if (len < 0.0)
    len = 0.0;
  f = len/70.0;
  if (f > 1.0)
    f = 1.0;
  f = 1.0 - f;
  VEC_MULT(c, f);
  f = 1.0 - f;
  VEC_ADDS(c, c, f, d);

  /* ground fog */
  if(x[1] < .25)
  {
    f = v[1]/20.0;
    if (f<0) f = 0;
    if (f > 1.0)
      f = 1.0;
    f = 1.0 - f;
    VEC_MULT(c, f);
    f = 1.0 - f;
    VEC_ADDS(c, c, f, d);
  }
  
}



