/*
 * Copyright (C) 1994, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */
#include <stdio.h>
#include <gl.h>

#include "defs.h"
#include "field.h"
#include "creatures.h"


/*
	Creates and draws the playing field.
*/

extern CREATURE creatures[MAX_PLAYERS];
extern short num_players;
extern short me;

TRIANGLE hole_edge[12], hole_center[12];
VECTOR edge_border[6], center_border[6], tri_edges[12];
VERT_LIST *cracks[6]; /* 6 cracks emmanating from center */
VECTOR bound[4][2] = {       /* how far the creature can go */
  { {90, 0, 90}, {90, 5, 90} },
  { {-90, 0, 90}, {-90, 5, 90} },
  { {-90, 0, -90}, {-90, 5, -90} },
  { {90, 0, -90}, {90, 5, -90} }
};

      
/* makes a crack in the ground from start to the edge of the board, in
   the general direction of d */
void
make_crack(VECTOR start, VECTOR d, VERT_LIST **cpoint)
{
  VERT_LIST *crack;
  int n, i;
  short sign;
  VECTOR x, y, seg; /* temp */
  float len, crack_len, dist_left, seg_len;

  if ((random() % 2) == 0) sign = 1;
  else sign = -1;
  
  /* find full length of crack */
  VEC_COPY(x, d);
  VEC_MULT(x, 150);
  VEC_ADD(y, x, start);
  if (y[0] < -90) y[0] = -90;
  if (y[0] > 90) y[0] = 90;
  if (y[2] < -90) y[2] = -90;
  if (y[2] > 90) y[2] = 90;
  VEC_SUB(y, y, start);
  crack_len = dist_left = VEC_LEN(y);

  /* now create the crack */
  n = random_num(7, 14); /* between 7 and 14 segs */
  seg_len = crack_len/n;
  
  crack = (VERT_LIST *) malloc(sizeof(VERT_LIST));
  *cpoint = crack;
  
  VEC_COPY(crack->v, start);

  for (i=1; i<n; i++)
  {
    VEC_COPY(seg, d);
    len = random_num(seg_len/2, seg_len*2);
    if (len > dist_left) len = dist_left;
    dist_left -= len;
    VEC_MULT(seg, len);  /* size of this segment is set */

    crack->next = (VERT_LIST *) malloc(sizeof(VERT_LIST));
    VEC_ADD(crack->next->v, crack->v, seg); /* create the next segment */
    crack = crack->next;
    
    /* now offset it */
    crack->v[0] += random_num(sign*len/5, sign*len*.6);
    crack->v[2] += random_num(sign*len/5, sign*len*.6);

    sign *= -1;

    if (crack->v[0] < -90) { crack->v[0] = -90; i=n; }
    if (crack->v[0] > 90) { crack->v[0] = 90; i=n; }
    if (crack->v[2] < -90) { crack->v[2] = -90; i=n; }
    if (crack->v[2] > 90) { crack->v[2] = 90; i=n; }
  }

/*  crack->next = (VERT_LIST *) malloc(sizeof(VERT_LIST));
  crack->next->last = crack;
  VEC_ADD(crack->next->v, start, y); /* create the next segment */
/*  crack = crack->next; */
  
  crack->next = NULL;

}

  
	       
/* makes the hole, represented as a bunch of triangles, and a bottom polygon */

void field_compute()
{
  int i;
  short edge;
  VECTOR v;
  
  /* making the hole first */

  create_border(BIG_BOUNDS, 0.0, edge_border);
  create_border(SMALL_BOUNDS, -3.0, center_border);

  /* now create cracks emmanating from each vertex on the edge_border */
  for (i=0; i<6; i++)
  {
    VEC_SUB(v, edge_border[i], center_border[i]); /* directional */
    v[1] = 0; /* flat */
    VEC_NORM(v);
    make_crack(edge_border[i], v, &cracks[i]);
  }    
  
  /* we'll triangulate each four sided polygon in one of two ways for
   * coolness */
  for(edge = 0; edge < 5; edge++)
    {
      if(random() %2)
	{
	  VEC_COPY(hole_edge[edge*2].v[0], edge_border[edge]);
	  VEC_COPY(hole_edge[edge*2].v[1], edge_border[edge+1]);
	  VEC_COPY(hole_edge[edge*2].v[2], center_border[edge+1]);

	  /* now calculate the lighting and normal  for this triangle given 
	   * the original color and the two vectors */
	  calc_lighting(hole_edge[edge*2].v[0], hole_edge[edge*2].v[2], 
			hole_edge[edge*2].v[1], hole_edge[edge*2].n, 
			cground, hole_edge[edge*2].c);
	  
	  VEC_COPY(hole_edge[edge*2 + 1].v[0], center_border[edge]);
	  VEC_COPY(hole_edge[edge*2 + 1].v[1], edge_border[edge]);
	  VEC_COPY(hole_edge[edge*2 + 1].v[2], center_border[edge+1]);
	  calc_lighting(hole_edge[edge*2 + 1].v[0], hole_edge[edge*2 + 1].v[2],
			hole_edge[edge*2 + 1].v[1], hole_edge[edge*2 + 1].n, 
			cground, hole_edge[edge*2 + 1].c);
	}
      else
	{
	  VEC_COPY(hole_edge[edge*2].v[0], center_border[edge]);
	  VEC_COPY(hole_edge[edge*2].v[1], edge_border[edge+1]);
	  VEC_COPY(hole_edge[edge*2].v[2], center_border[edge+1]);
	  calc_lighting(hole_edge[edge*2].v[0], hole_edge[edge*2].v[2], 
			hole_edge[edge*2].v[1], hole_edge[edge*2].n, 
			cground, hole_edge[edge*2].c);

	  VEC_COPY(hole_edge[edge*2 + 1].v[0], center_border[edge]);
	  VEC_COPY(hole_edge[edge*2 + 1].v[1], edge_border[edge]);
	  VEC_COPY(hole_edge[edge*2 + 1].v[2], edge_border[edge+1]);
	  calc_lighting(hole_edge[edge*2 + 1].v[0], hole_edge[edge*2 + 1].v[2],
			hole_edge[edge*2 + 1].v[1], hole_edge[edge*2 + 1].n, 
			cground, hole_edge[edge*2 + 1].c);

	}	  
    }
  /* last triangles special */
  if(random() %2)
    {
      VEC_COPY(hole_edge[10].v[0], edge_border[5]);
      VEC_COPY(hole_edge[10].v[1], edge_border[0]);
      VEC_COPY(hole_edge[10].v[2], center_border[0]);
      calc_lighting(hole_edge[10].v[0], hole_edge[10].v[2], hole_edge[10].v[1],
		    hole_edge[10].n, cground, hole_edge[10].c);

      VEC_COPY(hole_edge[11].v[0], center_border[5]);
      VEC_COPY(hole_edge[11].v[1], edge_border[5]);
      VEC_COPY(hole_edge[11].v[2], center_border[0]);
      calc_lighting(hole_edge[11].v[0], hole_edge[11].v[2], hole_edge[11].v[1],
		    hole_edge[11].n, cground, hole_edge[11].c);
    }
  else
    {
      VEC_COPY(hole_edge[10].v[0], center_border[5]);
      VEC_COPY(hole_edge[10].v[1], edge_border[0]);
      VEC_COPY(hole_edge[10].v[2], center_border[0]);
      calc_lighting(hole_edge[10].v[0], hole_edge[10].v[2], hole_edge[10].v[1],
		    hole_edge[10].n, cground, hole_edge[10].c);

      VEC_COPY(hole_edge[11].v[0], center_border[5]);
      VEC_COPY(hole_edge[11].v[1], edge_border[5]);
      VEC_COPY(hole_edge[11].v[2], edge_border[0]);
      calc_lighting(hole_edge[11].v[0], hole_edge[11].v[2], hole_edge[11].v[1],
		    hole_edge[11].n, cground, hole_edge[11].c);

    }	  

}

draw_crack(VERT_LIST *crack, int player)
{
  int line = 7;
  COLOR c;

  player += 1;
  if (player == 3) player = 0;
  
  
  do
  {
    linewidth(line);
    if (line > 2) line-=2;
    bgnline();
    VEC_COPY(c, cblack);
    add_fog(c, crack->v);
    c3s(c);
    v3f(crack->v);

    VEC_COPY(c, cblack);
    add_fog(c, crack->next->v);
    c3s(c);
    v3f(crack->next->v);
    endline();

    crack = crack->next;
  } while(crack->next != NULL);
  
}

draw_region(VERT_LIST *crack1, VERT_LIST *crack2, int player)
{
  COLOR c;
  int i;
  

  player += 1;
  if (player == 3) player = 0;

  linewidth(2);
  
  bgnpolygon();
  
  crack1 = crack1->next;
  crack2 = crack2->next;
  
  for (i=0; i<2; i++)
  {
    VEC_COPY(c, groundplane[player]);
    add_fog(c, crack1->v);
    c3s(c);
    v3f(crack1->v);
    crack1 = crack1->next;
  }

  VEC_COPY(c, groundplane[player]);
  add_fog(c, crack2->next->v);
  c3s(c);
  v3f(crack2->next->v);

  VEC_COPY(c, groundplane[player]);
  add_fog(c, crack2->v);
  c3s(c);
  v3f(crack2->v);

  crack2 = crack2->next;
 
  endpolygon();

}


void
  draw_line(VECTOR zero, VECTOR one)
{
  COLOR c;
  
  bgnline();
  VEC_COPY(c, cborder);
  add_fog(c, zero);
  c3s(c);
  v3f(zero);

  VEC_COPY(c, cborder);
  add_fog(c, one);
  c3s(c);
  v3f(one);
  endline();
}

/* this checks for a collision with the wall, and if it finds one, lights up
   that wall */
void draw_collisions()
{
  COLOR c;
  int bound1, bound2;
  VECTOR x1, x2;
  
  int i; /* which creature we're testing */

  for (i=0; i<num_players; i++)
  {
    if((random() % 5) == 0)
      if (border_test(&bound1, &bound2, i))
      {
      /* draw two gouraud-shaded polygons, so that the brightest yellow is at 
	 the creature */
	
	VEC_COPY(x1, bound[bound1][0]);
	VEC_COPY(x2, bound[bound1][1]);
	
	if (ABS(creatures[i].position[0]) == 80)
	{
	  x1[2] = creatures[i].position[2];
	  x2[2] = creatures[i].position[2];
	}	
	else
	{
	  x1[0] = creatures[i].position[0];
	  x2[0] = creatures[i].position[0];
	}	
	
	
	bgnpolygon();
	
	VEC_COPY(c, cborder);
	add_fog(c, bound[bound1][0]);
	c3s(c);
	v3f(bound[bound1][0]);
	
	VEC_COPY(c, cborder);
	add_fog(c, bound[bound2][0]);
	c3s(c);
	v3f(bound[bound1][1]);
	
	VEC_COPY(c, cborder);
	add_fog(c, x2);
	c3s(c);
	v3f(x2);
	
	VEC_COPY(c, cborder);
	add_fog(c, x1);
	c3s(c);
	v3f(x1);
	
	endpolygon();


	bgnpolygon();
	
	VEC_COPY(c, cborder);
	add_fog(c, x1);
	c3s(c);
	v3f(x1);

	VEC_COPY(c, cborder);
	add_fog(c, x2);
	c3s(c);
	v3f(x2);
	
	VEC_COPY(c, cborder);
	add_fog(c, bound[bound2][1]);
	c3s(c);
	v3f(bound[bound2][1]);
	
	VEC_COPY(c, cborder);
	add_fog(c, bound[bound2][0]);
	c3s(c);
	v3f(bound[bound2][0]);
	
	endpolygon(); 
      }
    }

}


  
  
  
void field_draw()
{
  VECTOR v, camera_vector;
  short edge, tri;
  int i;
  COLOR c;

  /* bottom of hole */
  VEC_COPY(c, cground_shadow); 
  add_fog(c, center_border[0]);
  c3s(c);
  bgnpolygon();
  for(edge = 0; edge < 6; edge++)
    v3f(center_border[edge]);
  endpolygon();
  
  zbuffer(TRUE);

  linewidth(5);
  /* perimeters */
  bgnclosedline();
  for(edge = 0; edge < 6; edge++)
  {
    VEC_COPY(c, cblack);
    add_fog(c, edge_border[edge]);
    c3s(c);
    v3f(edge_border[edge]);
  }
  endclosedline();

  /* lighted triangles */
  for(tri = 0; tri < 12; tri++)
  {
    VEC_SUB(camera_vector, camera_from, hole_edge[tri].v[0]);
    if(VEC_DOT(hole_edge[tri].n, camera_vector) >= 0.0) 
    {
      bgnpolygon();
      VEC_COPY(c, hole_edge[tri].c);
      add_fog(c, hole_edge[tri].v[0]);
      c3s(c);
      v3f(hole_edge[tri].v[0]);

      VEC_COPY(c, hole_edge[tri].c);
      add_fog(c, hole_edge[tri].v[1]);
      c3s(c);
      v3f(hole_edge[tri].v[1]);

      VEC_COPY(c, hole_edge[tri].c);
      add_fog(c, hole_edge[tri].v[2]);
      c3s(c);
      v3f(hole_edge[tri].v[2]);
      endpolygon();
    } 
    else 
    {
      cpack(BACKGROUND_COLOR);
      bgnpolygon();
      v3f(hole_edge[tri].v[0]);
      v3f(hole_edge[tri].v[1]);
      v3f(hole_edge[tri].v[2]);
      endpolygon();
    }
    
  }


  /* border fence */
  linewidth(2);
  draw_line(bound[0][0], bound[0][1]);  /* posts */
  draw_line(bound[1][0], bound[1][1]);
  draw_line(bound[2][0], bound[2][1]);
  draw_line(bound[3][0], bound[3][1]);

  draw_line(bound[0][1], bound[1][1]);    /* railings */
  draw_line(bound[1][1], bound[2][1]);
  draw_line(bound[2][1], bound[3][1]);
  draw_line(bound[3][1], bound[0][1]);


  zbuffer(FALSE);

  draw_line(bound[0][0], bound[1][0]);  /* on ground */
  draw_line(bound[1][0], bound[2][0]);
  draw_line(bound[2][0], bound[3][0]);
  draw_line(bound[3][0], bound[0][0]);

  /* collisions with fence */
  draw_collisions();
  
  /* cracks */
  for(i=0; i<6; i+=2)
    draw_region(cracks[i], cracks[i+1], i/2);
  for(i=0; i<6; i++)
    draw_crack(cracks[i], i/2);
}





