/*
 * Copyright (C) 1994, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */
#include <stdio.h>

#include "world.h"
#include "game.h"
#include "creatures.h"

/*
	Extrapolate the state of the specified world
	by 't' game ticks.
*/

void game_update_world(WORLD *world, float t)
{

}

/*
	Update the world to reflect new player control information.
*/

void game_update_player(WORLD *world, PLAYER *player)
{
  short i, j, chipi, chipj;
  CHIPS *chip;
  int alive=0;
  int n_its=0;
  int n_chips;
  
  i = player->index;
  VEC_COPY(world->creat_posit[i], player->creat_posit);
  world->creat_xrot[i] = player->creat_xrot;
  world->creat_yrot[i] = player->creat_yrot;
  world->creat_zrot[i] = player->creat_zrot;
  world->creat_beak_xrot[i] = player->creat_beak_xrot;
  world->creat_eyes_yrot[i][0] = player->creat_eyes_yrot[0];
  world->creat_eyes_yrot[i][1] = player->creat_eyes_yrot[1];
  world->creat_speed[i] = player->creat_speed;
  world->creat_alive[i] = player->alive;
  world->creat_status[i] = player->creat_status;
  world->creat_aimat[i] = player->creat_aimat;
  if(player->creat_it)
    world->creat_it[i] = player->creat_it;
  
  if (player->chip)
  { /* worry about chip being dropped */
    chipi = player->chip_indexi;
    chipj = player->chip_indexj;

    VEC_COPY(world->chip_position[chipi][chipj], player->chip_position);
    world->chip_yrot[chipi][chipj] = player->chip_yrot;
    world->chip_status[chipi][chipj] = player->chip_status;
  }
 
  /* has the player finished dropping chips? */
  n_chips = 0; /* counting number of chips in the hole */
  for (j=0; j<4; j++)
    if (world->chip_status[i][j] == IN_HOLE)
      n_chips++;
  if (n_chips == 4) /* all four chips are in the hole */
  {
    world->creat_done[i] = TRUE;

    /* set up a restart chip */
    if (i==0)  VEC_SET(world->chip_position[i][0], 0, 0, 57);
    else if (i==1) VEC_SET(world->chip_position[i][0], 50, 0, -26);
    else if (i==2) VEC_SET(world->chip_position[i][0], -50, 0, -26);
    world->chip_status[i][0] = ON_GROUND;
  }
  
  /* worry about the possibility of being two its or none */
  for (i=0; i<world->num_players; i++)
  {
    if (world->creat_alive[i])
    {
      alive++;
      if (world->creat_it[i])
      {
	n_its++;
	if (n_its > 1)  /* more than one it */
	  if (world->creat_it[i] != C_IT) world->creat_it[i] = C_NOTIT;
	  else 
	    if(world->creat_it[0]) world->creat_it[0] = C_NOTIT;
	    else if (world->creat_it[1]) world->creat_it[1] = C_NOTIT;
      }
    }      
  }

  if ((n_its == 0) && (alive > 1))
  {
    i = 0;
    while(!world->creat_alive[i] && (i<world->num_players)) i++;
    world->creat_it[i] = C_IT;
  }
  else if (alive == 1)
  {
    i = 0;
    while(!world->creat_alive[i] && (i<world->num_players)) i++;
    world->creat_it[i] = C_NOTIT;
  }
  
}

    






