/*
 * Copyright (C) 1994, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */
#include <stdio.h>
#include <gl.h>

#include "goals.h"
#include "creatures.h"
/* 
    Creates and draws the goalposts 
*/

VERT_LIST *list1[3], *list2[3];
TRI_LIST *first_triangle[MAX_PLAYERS*2];
TRI_LIST *first_triangle_intersect[MAX_PLAYERS*2];
VECTOR player_lines[3][2] = {  /* the top of the goals */
/*  {{5.0, 15.0, 54.0},
   {3.0, 18.0, 53.0}},
  {{45.0, 10.0, -20.0},
   {36.0, 13.0, -23.0}}, 
  {{-47.0, 20.0, -26.0},
   {-47.0, 15.0, -28.0}}};  */
  {{7.0, 23.0, 54.0},
   {-3.0, 16.0, 48.0}},
  {{45.0, 25.0, -23.0},
   {36.0, 18.0, -28.0}}, 
  {{-49.0, 25.0, -26.0},
   {-45.0, 20.0, -34.0}}}; 
VECTOR goal_start[3][2] = {
/*  {{15, 0, 50},
   {-13, 0, 50}},
  {{50, 0, -12},
   {35, 0, -31}}, 
  {{-52, 0, -12},
   {-42, 0, -38}}}; */
 {{26, 0, 50},
   {-24, 0, 50}},
  {{65, 0, -3},
   {45, 0, -48}}, 
  {{-62, 0, -5},
   {-42, 0, -49}}};
      
/* for calculating intersections */
VECTOR goal_bounds[3][2][3]; /* 3 sets, 2 goals per set, 3 points per goal */

extern CREATURE creatures[MAX_PLAYERS];
extern short me;
extern short num_players;

/* this gives the outer bounds for the goals */
bounds_compute()
{
  int i, j, k;
  
  for (i=0; i<3; i++) /* player pairs */
  {
    for (j=0; j<2; j++) /* two goals */
    {
      for (k=0; k<3; k++) /* three points per goal */
      {
	VEC_COPY(goal_bounds[i][j][k], goal_start[i][j]);
      }
      goal_bounds[i][j][0][0] += 7;
      goal_bounds[i][j][0][2] -= 2;
      goal_bounds[i][j][1][0] -= 7;
      goal_bounds[i][j][1][2] -= 2;
      goal_bounds[i][j][2][2] += 7;

/*      if ((turn(goal_bounds[i][j][2], goal_bounds[i][j][0], 
		player_lines[i][j]) == CLOCKWISE) &&
	  (turn(goal_bounds[i][j][2], goal_bounds[i][j][1], 
		player_lines[i][j]) == COUNTERCLOCK))
      {
	goal_bounds[i][j][0][2] -= 4;
	goal_bounds[i][j][1][2] -= 4;
      }
      else if ((turn(goal_bounds[i][j][0], goal_bounds[i][j][1], 
		     player_lines[i][j]) == CLOCKWISE) &&
	       (turn(goal_bounds[i][j][0], goal_bounds[i][j][2], 
		     player_lines[i][j]) == COUNTERCLOCK))
      {
	goal_bounds[i][j][1][0] -= 2;
	goal_bounds[i][j][2][0] -= 2;
	goal_bounds[i][j][1][2] -= 2;
	goal_bounds[i][j][2][2] += 2;
      }
      else /*if ((turn(goal_bounds[i][j][1], goal_bounds[i][j][2], 
		player_lines[i][j]) == CLOCKWISE) &&
	  (turn(goal_bounds[i][j][1], goal_bounds[i][j][0], 
		player_lines[i][j]) == COUNTERCLOCK)) */
/*      {
	goal_bounds[i][j][2][0] += 2;
	goal_bounds[i][j][0][0] += 2;
	goal_bounds[i][j][2][2] += 2;
	goal_bounds[i][j][0][2] -= 2;
      }
*/	  
    }
  }
}

/* this creates 3 sets of goals. changing the vectors changes the 
   orientation of the posts */
posts_compute()
{
  bounds_compute(); /* boundaries for intersections */
  
  create_post(goal_start[0][0], player_lines[0][0], &first_triangle[0], 
	      &first_triangle_intersect[0]);
  create_post(goal_start[0][1], player_lines[0][1], &first_triangle[1],
	      &first_triangle_intersect[1]);
  create_post(goal_start[1][0], player_lines[1][0], &first_triangle[2],
	      &first_triangle_intersect[2]);
  create_post(goal_start[1][1], player_lines[1][1], &first_triangle[3],
	      &first_triangle_intersect[3]);
  create_post(goal_start[2][0], player_lines[2][0], &first_triangle[4],
	      &first_triangle_intersect[4]);
  create_post(goal_start[2][1], player_lines[2][1], &first_triangle[5],
	      &first_triangle_intersect[5]);
}

/* this creates a post, given a line around which to orient itself,and a pointer
   to the triangles that will store it. I was originally going to
   subdivide each post about 8 times, but just one subdivision was way cool.
   subdividing sorta bends the post randomly. the triangulation might be
   screwy, but it looks cool!
*/

create_post(VECTOR start, VECTOR end, TRI_LIST **first_tri, 
	    TRI_LIST **first_tri_intersect)
{
  VERT_LIST *verts[3], *first[3];
  TRI_LIST *tris, *tris2;
  VECTOR v;
  short i;
  short p1[2] = {-1,1};
  short p2[2] = {1,-1};

  verts[0] = (VERT_LIST *) malloc(sizeof(VERT_LIST)); 
  verts[1] = (VERT_LIST *) malloc(sizeof(VERT_LIST)); 
  verts[2] = (VERT_LIST *) malloc(sizeof(VERT_LIST)); 
  first[0] = verts[0]; first[1] = verts[1]; first[2] = verts[2];
  VEC_SET(verts[0]->v, start[0] + random_num(2,4), start[1], start[2]);
  VEC_SET(verts[1]->v, start[0], start[1], start[2] + random_num(2,4));
  VEC_SET(verts[2]->v, start[0] - random_num(1,3), start[1], 
	  start[2] );

  verts[0]->next = (VERT_LIST *) malloc(sizeof(VERT_LIST)); 
  verts[1]->next = (VERT_LIST *) malloc(sizeof(VERT_LIST)); 
  verts[2]->next = (VERT_LIST *) malloc(sizeof(VERT_LIST)); 
  VEC_COPY(verts[0]->next->v, end);
  VEC_COPY(verts[1]->next->v, end);
  VEC_COPY(verts[2]->next->v, end);
 
  /* *the* subdivision, cut it in half (sorta) */
  subdivide(verts[0],p1);
  subdivide(verts[1],p2);
  subdivide(verts[2],p1);

  /* create a list of triangles to draw */
  tris = (TRI_LIST *) malloc(sizeof(TRI_LIST));
  tris->wireframe = FALSE;
  *first_tri = tris;

  /* create a list of triangles to calculate intersections */
  tris2 = (TRI_LIST *) malloc(sizeof(TRI_LIST));
  *first_tri_intersect = tris2;

/*  while(verts[0]->next != NULL) */
  for (i=0; i<2; i++)
    {
      /* six triangles per layer */
      tris->t = (TRIANGLE *) malloc(sizeof(TRIANGLE));
      VEC_COPY(tris->t->v[0], verts[0]->v);
      VEC_COPY(tris->t->v[1], verts[0]->next->v);
      VEC_COPY(tris->t->v[2], verts[2]->v);
      calc_lighting(tris->t->v[0], tris->t->v[2], tris->t->v[1],
		    tris->t->n, cbackground, tris->t->c);
      tris->next = (TRI_LIST *) malloc(sizeof(TRI_LIST));
      
      if (i==0)
      {
	tris2->t = (TRIANGLE *) malloc(sizeof(TRIANGLE));
	VEC_COPY(tris2->t->v[0], verts[0]->v);
	VEC_HALF(v, verts[0]->next->v, verts[0]->v);
	VEC_COPY(tris2->t->v[1], v);
	VEC_COPY(tris2->t->v[2], verts[2]->v);
	tris2->t->v[0][1] = 0;	tris2->t->v[1][1] = 0;	tris2->t->v[2][1] = 0;
	tris2->next = (TRI_LIST *) malloc(sizeof(TRI_LIST));
      }	
	
      tris = tris->next;
      tris->t = (TRIANGLE *) malloc(sizeof(TRIANGLE));
      VEC_COPY(tris->t->v[0], verts[0]->next->v);
      VEC_COPY(tris->t->v[1], verts[2]->next->v);
      VEC_COPY(tris->t->v[2], verts[2]->v);
      calc_lighting(tris->t->v[0], tris->t->v[2], tris->t->v[1],
		    tris->t->n, cbackground, tris->t->c);
      tris->next = (TRI_LIST *) malloc(sizeof(TRI_LIST));

      if (i==0)
      {
	tris2 = tris2->next;
	tris2->t = (TRIANGLE *) malloc(sizeof(TRIANGLE));
	VEC_HALF(v, verts[0]->next->v, verts[2]->v);
	VEC_COPY(tris2->t->v[0], v);
	VEC_HALF(v, verts[2]->next->v, verts[2]->v);
	VEC_COPY(tris2->t->v[1], v);
	VEC_COPY(tris2->t->v[2], verts[2]->v);
	tris2->t->v[0][1] = 0;	tris2->t->v[1][1] = 0;	tris2->t->v[2][1] = 0;
	tris2->next = (TRI_LIST *) malloc(sizeof(TRI_LIST));
      }
      
      tris = tris->next;
      tris->t = (TRIANGLE *) malloc(sizeof(TRIANGLE));
      VEC_COPY(tris->t->v[0], verts[2]->v);
      VEC_COPY(tris->t->v[1], verts[2]->next->v);
      VEC_COPY(tris->t->v[2], verts[1]->v);
      calc_lighting(tris->t->v[0], tris->t->v[2], tris->t->v[1],
		    tris->t->n, cbackground, tris->t->c);
      tris->next = (TRI_LIST *) malloc(sizeof(TRI_LIST));

      if (i==0)
      {
	tris2 = tris2->next;
	tris2->t = (TRIANGLE *) malloc(sizeof(TRIANGLE));
	VEC_COPY(tris2->t->v[0], verts[2]->v);
	VEC_HALF(v, verts[2]->next->v, verts[2]->v);
	VEC_COPY(tris2->t->v[1], v);
	VEC_COPY(tris2->t->v[2], verts[1]->v);
	tris2->t->v[0][1] = 0;	tris2->t->v[1][1] = 0;	tris2->t->v[2][1] = 0;
	tris2->next = (TRI_LIST *) malloc(sizeof(TRI_LIST));
      }	

      tris = tris->next;
      tris->t = (TRIANGLE *) malloc(sizeof(TRIANGLE));
      VEC_COPY(tris->t->v[0], verts[2]->next->v);
      VEC_COPY(tris->t->v[1], verts[1]->next->v);
      VEC_COPY(tris->t->v[2], verts[1]->v);
      calc_lighting(tris->t->v[0], tris->t->v[2], tris->t->v[1],
		    tris->t->n, cbackground, tris->t->c);
      tris->next = (TRI_LIST *) malloc(sizeof(TRI_LIST));

      if (i==0)
      {
	tris2 = tris2->next;
	tris2->t = (TRIANGLE *) malloc(sizeof(TRIANGLE));
	VEC_HALF(v, verts[2]->next->v, verts[1]->v);
	VEC_COPY(tris2->t->v[0], v);
	VEC_HALF(v, verts[1]->next->v, verts[1]->v);
	VEC_COPY(tris2->t->v[1], v);
	VEC_COPY(tris2->t->v[2], verts[1]->v);
	tris2->t->v[0][1] = 0;	tris2->t->v[1][1] = 0;	tris2->t->v[2][1] = 0;
	tris2->next = (TRI_LIST *) malloc(sizeof(TRI_LIST));
      }
      
      tris = tris->next;
      tris->t = (TRIANGLE *) malloc(sizeof(TRIANGLE));
      VEC_COPY(tris->t->v[0], verts[1]->v);
      VEC_COPY(tris->t->v[1], verts[1]->next->v);
      VEC_COPY(tris->t->v[2], verts[0]->v);
      calc_lighting(tris->t->v[0], tris->t->v[2], tris->t->v[1],
		    tris->t->n, cbackground, tris->t->c);
      tris->next = (TRI_LIST *) malloc(sizeof(TRI_LIST));

      if (i==0)
      {
	tris2 = tris2->next;
	tris2->t = (TRIANGLE *) malloc(sizeof(TRIANGLE));
	VEC_COPY(tris2->t->v[0], verts[1]->v);
	VEC_HALF(v, verts[1]->next->v, verts[1]->v);
	VEC_COPY(tris2->t->v[1], v);
	VEC_COPY(tris2->t->v[2], verts[0]->v);
	tris2->t->v[0][1] = 0;	tris2->t->v[1][1] = 0;	tris2->t->v[2][1] = 0;
	tris2->next = (TRI_LIST *) malloc(sizeof(TRI_LIST));
      }
      
      tris = tris->next;
      tris->t = (TRIANGLE *) malloc(sizeof(TRIANGLE));
      VEC_COPY(tris->t->v[0], verts[1]->next->v);
      VEC_COPY(tris->t->v[1], verts[0]->next->v);
      VEC_COPY(tris->t->v[2], verts[0]->v);
      calc_lighting(tris->t->v[0], tris->t->v[2], tris->t->v[1],
		    tris->t->n, cbackground, tris->t->c);
      tris->next = (TRI_LIST *) malloc(sizeof(TRI_LIST));
      
      if (i==0)
      {
	tris2 = tris2->next;
	tris2->t = (TRIANGLE *) malloc(sizeof(TRIANGLE));
	VEC_HALF(v, verts[1]->next->v, verts[0]->v);
	VEC_COPY(tris2->t->v[0], v);
	VEC_HALF(v, verts[0]->next->v, verts[0]->v);
	VEC_COPY(tris2->t->v[1], v);
	VEC_COPY(tris2->t->v[2], verts[0]->v);
	tris2->t->v[0][1] = 0;	tris2->t->v[1][1] = 0;	tris2->t->v[2][1] = 0;
	tris2->next = (TRI_LIST *) malloc(sizeof(TRI_LIST));

	tris2 = tris2->next;
	tris2->t = (TRIANGLE *) malloc(sizeof(TRIANGLE));
	VEC_COPY(tris2->t->v[0], verts[0]->v);
	VEC_COPY(tris2->t->v[1], verts[1]->v);
	VEC_COPY(tris2->t->v[2], verts[2]->v);
      }
      
      verts[0] = verts[0]->next;
      verts[1] = verts[1]->next;
      verts[2] = verts[2]->next;
    }

}


/* this subdivides a line segment randomly */

subdivide(VERT_LIST *vert, short perp[2])
{
  float maxd, mind;
  float len, dist;
  VECTOR direction;
  VECTOR new_point;
  VERT_LIST *temp_vert;
  float temp;

  len = sqrt(VEC_DOT(vert->v, vert->next->v));
  maxd = len/4;
  mind = len/8;
  /* get a random distance */
  dist = random_num(mind, maxd);

  /* new point is perpendicular to our directional vector */
  VEC_SUB(direction, vert->next->v, vert->v);
  VEC_COPY(new_point, direction);
  VEC_NORM(new_point);
  
  temp = new_point[0];
  new_point[0] = -new_point[2]*perp[0];
  new_point[2] = temp*perp[1];

  /* now set it our distance away */
  VEC_MULT(new_point, dist); 

  VEC_DIV(direction, 2+random_num(-.5,.5)); 
  VEC_ADD(new_point, new_point, direction);
  VEC_ADD(new_point, new_point, vert->v);

  /* insert it into the list */
  temp_vert = (VERT_LIST *) malloc(sizeof(VERT_LIST)); 
  VEC_COPY(temp_vert->v, new_point);
  temp_vert->next = vert->next;
  vert->next = temp_vert;

}



/* draws an electronic-like spark between the tips of the goalposts */

/* loop through the set of goalposts, drawing each one. change the number
   of goalposts drawn by changing the global value for number of players */
/* borrowed the basic stuff for this from pods */
void
  draw_spark(VECTOR zero, VECTOR one, int player)
{
  int i;
  VECTOR s, w, x, v;
  float len;
  COLOR c;
  
  VEC_COPY(c, colors[player]);
  add_fog(c, zero);
  c3s(c);

  for (i=0; i < 3; i++) 
  {
    if ((random() % 3) == 0)
      linewidth(2);
    else
      linewidth(1);

    VEC_COPY(w, one);
    VEC_COPY(x, zero);
    
    bgnline();
    do
    {
      VEC_SUB(v, w, x);
      len = VEC_LEN(v);
      VEC_NORM(v);
      VEC_ADDS(x, x, 0.1, v);
      x[0] += ((random() % 1000)/500.0 - 1.0)*0.065;
      x[1] += ((random() % 1000)/500.0 - 1.0)*0.065;
      x[2] += ((random() % 1000)/500.0 - 1.0)*0.065;
      v3f(x);
    }
    while (len > 0.05);
    endline();
  }
}

posts_draw()
{
  short i, j;
  short c[3];

  
  zbuffer(TRUE);


  for (i=0; i<MAX_PLAYERS; i++)
    {
      linewidth(3);
      draw_post(first_triangle[i*2], first_triangle[i*2]->wireframe, 11); 
      draw_post(first_triangle[i*2 +1], first_triangle[i*2+1]->wireframe, 11);

/*      draw_post(first_triangle_intersect[i*2], TRUE, 6); 
      draw_post(first_triangle_intersect[i*2 +1], TRUE, 6);      

      for (j=0; j<2; j++) 
      {
	VEC_COPY(c, cwhite);
	add_fog(c, goal_bounds[i][j][0]);
	c3s(c);
	bgnclosedline();
	v3f(goal_bounds[i][j][0]);
	v3f(goal_bounds[i][j][1]);
	v3f(goal_bounds[i][j][2]);
	endclosedline();
      }
*/      
      /* reset the wireframe stuff */
      first_triangle[i*2]->wireframe = FALSE; 
      first_triangle[i*2 + 1]->wireframe = FALSE; 

      /* draw a spark between them */
      draw_spark(player_lines[i][0], player_lines[i][1], i);
      
/*      VEC_COPY(c, colors[i]);
      c3s(c);
      bgnline();
      v3f(player_lines[i][0]);
      v3f(player_lines[i][1]);
      endline();
*/
    }

  zbuffer(FALSE);

}

/* this draws a single post represented as a list of triangles. the border
   of the posts flicker, we might want to look into how to fix that */

draw_post(TRI_LIST *triangles, short wireframe, int num_tris)
{
  short c[3];
  short i, j;
  

/*  do */
  for (i=0; i<num_tris; i++)
    {
      VEC_COPY(c, cground_shadow);
      add_fog(c, triangles->t->v[0]);
      c3s(c);
      bgnclosedline();
      v3f(triangles->t->v[0]);
      v3f(triangles->t->v[1]);
      v3f(triangles->t->v[2]);
      endclosedline();

      
      /*** only fill in the polygon if not in wireframe mode ***/
      if (!wireframe) 
      {
	VEC_COPY(c, triangles->t->c);
	if (creatures[me].done)  VEC_COPY(c, colors[me]);
	else
	  for (j=0; j<num_players; j++)
	    if ((creatures[j].done) && (creatures[j].alive))
	      VEC_COPY(c, colors[j]);
	add_fog(c, triangles->t->v[0]);
	c3s(c);
	bgnpolygon();
	v3f(triangles->t->v[0]);
	v3f(triangles->t->v[1]);
	v3f(triangles->t->v[2]);
	endpolygon();
      }

      triangles = triangles->next;
    }
/*  while(triangles->next != NULL); */
  
}
    











