/*
 * Copyright (C) 1994, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */
#include <math.h>

#include "misc.h"
#include "vector.h"
#include "matrix.h"

void mat_rot_apply(VECTOR y, VECTOR x, MATRIX a)
{
  y[0] = x[0]*a[0][0] + x[1]*a[1][0] + x[2]*a[2][0];
  y[1] = x[0]*a[0][1] + x[1]*a[1][1] + x[2]*a[2][1];
  y[2] = x[0]*a[0][2] + x[1]*a[1][2] + x[2]*a[2][2];
}

/* create an identity matrix */
void mat_identity(MATRIX a)
{
  a[0][0] = 1.; a[0][1] = 0.; a[0][2] = 0.; a[0][3] = 0.;
  a[1][0] = 0.; a[1][1] = 1.; a[1][2] = 0.; a[1][3] = 0.;
  a[2][0] = 0.; a[2][1] = 0.; a[2][2] = 1.; a[2][3] = 0.;
  a[3][0] = 0.; a[3][1] = 0.; a[3][2] = 0.; a[3][3] = 1.;
}

/* matrix multiply: c=a*b */
void mat_mul(MATRIX a, MATRIX b, MATRIX c)
{
    c[0][0] = a[0][0]*b[0][0] + a[0][1]*b[1][0] + a[0][2]*b[2][0];
    c[0][1] = a[0][0]*b[0][1] + a[0][1]*b[1][1] + a[0][2]*b[2][1];
    c[0][2] = a[0][0]*b[0][2] + a[0][1]*b[1][2] + a[0][2]*b[2][2];
    c[1][0] = a[1][0]*b[0][0] + a[1][1]*b[1][0] + a[1][2]*b[2][0];
    c[1][1] = a[1][0]*b[0][1] + a[1][1]*b[1][1] + a[1][2]*b[2][1];
    c[1][2] = a[1][0]*b[0][2] + a[1][1]*b[1][2] + a[1][2]*b[2][2];
    c[2][0] = a[2][0]*b[0][0] + a[2][1]*b[1][0] + a[2][2]*b[2][0];
    c[2][1] = a[2][0]*b[0][1] + a[2][1]*b[1][1] + a[2][2]*b[2][1];
    c[2][2] = a[2][0]*b[0][2] + a[2][1]*b[1][2] + a[2][2]*b[2][2];
}

matrix_print(char *str, MATRIX a)
{
    int i;

    printf("%s:\n", str);
    for (i=0; i<3; i++)
	printf("  %9.4f %9.4f %9.4f\n", a[i][0], a[i][1], a[i][2]);
}

/* create a matrix to multiply angle around x */
void matrix_rotx(float angle, MATRIX m)
{
  float c, s;
  
  c = cos(angle);
  s = sin(angle);
  m[0][0] = 1.; m[0][1] = 0.; m[0][2] = 0.; m[0][3] = 0.;
  m[1][0] = 0.; m[1][1] = c;  m[1][2] = -s; m[1][3] = 0.;
  m[2][0] = 0.; m[2][1] = s;  m[2][2] = c;  m[2][3] = 0.;
  m[3][0] = 0.; m[3][1] = 0.; m[3][2] = 0.; m[3][3] = 1.;
}  

/* create a matrix to multiply angle around y */
void matrix_roty(float angle, MATRIX m)
{
  float c, s;
  
  c = cos(angle);
  s = sin(angle);
  m[0][0] = c;  m[0][1] = 0;  m[0][2] = s;  m[0][3] = 0.;
  m[1][0] = 0.; m[1][1] = 1.; m[1][2] = 0.; m[1][3] = 0.;
  m[2][0] = -s; m[2][1] = 0;  m[2][2] = c;  m[2][3] = 0.;
  m[3][0] = 0.; m[3][1] = 0.; m[3][2] = 0.; m[3][3] = 1.;
}  

/* create a matrix to multiply angle around z */
void matrix_rotz(float angle, MATRIX m)
{
  float c, s;
  
  c = cos(angle);
  s = sin(angle);
  m[0][0] = c;  m[0][1] = s;  m[0][2] = 0.; m[0][3] = 0.;
  m[1][0] = -s; m[1][1] = c;  m[1][2] = 0.; m[1][3] = 0.;
  m[2][0] = 0.; m[2][1] = 0.; m[2][2] = 1.; m[2][3] = 0.;
  m[3][0] = 0.; m[3][1] = 0.; m[3][2] = 0.; m[3][3] = 1.;
}  
  
