/*
 * Copyright (C) 1994, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */
#include <stdio.h>
#include <gl.h>
#include <device.h>
#include "creatures.h"
#include "game.h"

extern CREATURE creatures[MAX_PLAYERS];
extern CHIPS *chips[MAX_PLAYERS][CHIPS_PER_PLAYER];

extern short me;  /* index of my creature */
extern short num_players; /* how many creatures are there */

float howfast = 1.0;
float bestrot;

float facing; /* angle the creature is facing */
VECTOR looking; /* which way the camera is looking */
		  
void drop_chip(CREATURE *creature);
  
/****************
  creature movement! 
********************/

/* eyes try to look at spot_posit */
float
  creature_lookat(VECTOR spot_posit, EYE eyes[2], float other_rot)
{
  VECTOR spot, spot_dir;
  VECTOR eye_vector;
  float angle;

  VEC_COPY(spot,spot_posit);

  spot[0] = ABS(spot[0]);
  spot[2] = ABS(spot[2]);
  VEC_NORM(spot); 
  
/*  if (spot[0] < .001 || spot[2] < .001)
    return(0);
*/

  /* get the angle of rotation from the origin to the mouse direction...
     remembering that the creature doesn't really start on the x-axis but
     at "facing" degrees */

  angle = DEGREES(asinf(spot[2]/VEC_LEN(spot)));  
/*  angle = DEGREES(asinf(spot[2])); */

  if (spot_posit[2] <= 0)
    if (spot_posit[0] <= 0)  /* 3rd quadrant */
      angle += 180;
    else                    /* 4th quadrant */
      angle *= -1; /*angle -= 90;*/
  else 
    if (spot_posit[0] < 0)   /* second quandrant */
      angle = 180 - angle; /* not += 90 */

  /* add in initial rotation (by camera position, etc) */
  angle += facing; 
    
  /* compensate for body and head rotations */
  angle -= other_rot;
  
  /* keep it in a certain range */
  while(angle > 180)
    angle -= 360;
  while(angle < -180)
    angle += 360;

  /* can only look so far */
  if (angle > 80)
    angle = 80;
  else if (angle < -80)
    angle = -80;
  
/*  eyes[0].yrot = angle;
  eyes[1].yrot = angle;
*/
  return (angle);
}

  
/* called each time in event loop. updates bird movement. speed of everything
   regulated with a timer */
void 
  creatures_update(float dx,float dy)
{
  VECTOR spot;
  float rot;
  float len;
  float time;

  if (creatures[me].it)
  {
    if (creatures[me].chip != NULL) /* if it has a chip drop it */
      drop_chip(&creatures[me]);
    
  }
  
  {
    
    time = time_it(STOP_TIMER, TIMER_MOVE);
    if (time > 1) /* ridiculous */ 
      howfast = 1.0;
    else /* move 'speed' units per tenth of a second */
      howfast = time/.1;
    
    /* reset the timer now */
    time_it(START_TIMER, TIMER_MOVE);
    
    /* if the creature is swooping special turns */
    if ((creatures[me].status == C_SWOOP_DOWN) || 
	(creatures[me].status == C_SWOOP_UP))
    {
    } 
    
    else
    {
    
      VEC_SET(spot, dx, 0, dy);
      
      /* look at the point */  
      if (sqrt(dx*dx + dy*dy) > .01) /* if the mouse has moved significantly */
	rot = creature_lookat(spot, creatures[me].eyes, 
			      creatures[me].beak.yrot + creatures[me].yrot);
      else rot = 0;
      
      creatures[me].eyes[0].yrot = rot;
      creatures[me].eyes[1].yrot = rot;
      
      /* swing the head if it's not easily seen */
      if (ABS(rot) > 25) 
      {
	creatures[me].beak.yrot += rot/6;  
	creatures[me].eyes[0].yrot /= 4.;
	creatures[me].eyes[1].yrot /= 4.;
      }
    
      /* swing the body if the head can't rotate anymore */
      if (ABS(creatures[me].beak.yrot) > 25) /*(30)*/
      {
	if(ABS(creatures[me].beak.yrot) < 35)  /* precise turn */
	{
	  creatures[me].yrot += creatures[me].beak.yrot/10; /* swinging body */
	  creatures[me].zrot = creatures[me].beak.yrot/3;/*tilting in the 
							   direction of the turn */
	  creatures[me].beak.yrot = SIGN(creatures[me].beak.yrot)*25;
	}
	else /* sharp turn */
	{
	  creatures[me].yrot += creatures[me].beak.yrot/8; /* swinging body */
	  creatures[me].zrot = creatures[me].beak.yrot/3;/*tilting in the 
							   direction of the turn */
	  creatures[me].beak.yrot = SIGN(creatures[me].beak.yrot)*35;      
	}
	creatures[me].status = C_TURN;  /* turning */
	
	
      }
      else   /* creature not turning */
      {
	/* speed of bird determined here */
	len = sqrt(dx*dx + dy*dy); 
	if (dy < 0) 
	  len *= -1;
	creatures[me].speed += len - .25;
	if (creatures[me].speed > 2.0) creatures[me].speed = 2.0;
	if (creatures[me].speed < .75)  creatures[me].speed = .75;
	
	creatures[me].zrot -= creatures[me].zrot/4;
	creatures[me].status = C_NORMAL;
      }
    }
  

    /* create rotation matrices for results */
    matrix_roty(RADIANS(creatures[me].beak.yrot), creatures[me].beak.yrot_mat);
  }
  
}

/* keeps the wings flapping */
void
wing_update(WING wing[2], int who)
{
  float wing0d, wing1d; /* change in wing rotation */
  
  if (wing[0].p1yrot >= 16.5 || wing[0].p1yrot <= -4.5)
  {
    wing[0].direction *= -1.;
    wing[1].direction *= -1.; 
  }
  
  wing0d = wing[0].direction*howfast*creatures[who].speed;
  wing1d = wing[1].direction*howfast*creatures[who].speed;
    
  if((wing0d + wing[0].p1yrot) > 16.5)
    wing0d = wing1d = 16.5 - wing[0].p1yrot;
  else if((wing0d + wing[0].p1yrot) < -4.5)
    wing0d = wing1d = -4.5 - wing[0].p1yrot;
    
    
  wing[0].p1yrot += wing0d;
  wing[1].p1yrot -= wing1d;
  
  wing[0].p1zrot += 2.*wing0d;
  wing[1].p1zrot -= 2.*wing1d;
  
  wing[0].p2zrot += 4.*wing0d;
  wing[1].p2zrot -= 4.*wing1d;
 
  
}


wings_update()
{
  short i;

  for (i=0; i<num_players; i++)
  {
    {
      /* Wings should be up and stationary */
      if (creatures[i].status == C_SWOOP_DOWN 
	  || creatures[i].status == C_TURN) 
      {
	if (creatures[i].wings[0].p1yrot >= 0.)  /* wings are low */
	{
	  if ((creatures[i].wings[0].direction > 0) /* if flapping down */
	    && (creatures[i].wings[0].p1yrot < 16)) /* but won't go up on own*/
	  {
	    creatures[i].wings[0].direction *= -1;  /* flap up */
	    creatures[i].wings[1].direction *= -1;
	  }
	  wing_update(creatures[i].wings, i);
	}
      }
      else if (creatures[i].status == C_SWOOP_UP)
      {
	if (creatures[i].wings[0].p1yrot <= 15.0)
	{
	  if (creatures[i].wings[0].direction < 0)
	  {
	    creatures[i].wings[0].direction *= -1;
	    creatures[i].wings[1].direction *= -1;
	  }
	  wing_update(creatures[i].wings, i);
	}
      }

      else
	wing_update(creatures[i].wings, i); 
    }
  }
  
}


/* move 'howmuch' in the direction the creature is facing */
void 
  creature_forward(CREATURE *creature, float howmuch)
{
  VECTOR direction;
  
  {
    /* keep degree measures in a certain range */
    while(creature->yrot > 180)
      creature->yrot -= 360;
    while(creature->yrot < -180)
      creature->yrot += 360;
    
    while(facing > 180)
      facing -= 360;
    while(facing < -180)
      facing += 360;
    
    
    /* get vector of current direction */
    /* correcting for initial rotation with -90 */
    VEC_SET(direction, fcos(RADIANS(-creature->yrot -90)), 0, 
	    fsin(RADIANS(-creature->yrot -90)));
    
    VEC_MULT(direction, howmuch); 
    VEC_ADD(creature->position, creature->position, direction);
    
    /* if a goal obstructs the view, make it wireframe. if the creature is 
       hitting a goal, don't move it */
    if (find_viewblocks()) 
    {
      VEC_SUB(creature->position, creature->position, direction); 
    }
    
    /* don't go beyond the boundaries! */
    if(creature->position[0] < -80)
      creature->position[0] = -80;
    else if (creature->position[0] > 80)
      creature->position[0] = 80;
    if(creature->position[2] < -80)
      creature->position[2] = -80;
    else if (creature->position[2] > 80)
      creature->position[2] = 80;
  }
  
}

void
drop_chip(CREATURE *creature)
{
  creature->chip->status = DROPPING;
  VEC_SET(creature->chip->final_position, creature->position[0], 0.0,
	  creature->position[2]);
  creature->chip->final_position[1] = 0.0;
  
  /* move the chip differently if it's in the hole or hitting something*/
  find_chipblocks(creature->chip, creature->chip->final_position);

  client_update_chip(creature->chip); /* tell the server 'bout our changes */
  
  creature->chip = NULL;   /* drop the chip */
  creature->holding_chip = FALSE;

}	

/* move 'howmuch' down */
void
creature_down(CREATURE *creature, float howmuch)
{
  /* if it's down low enough then begin to move back up */
  if (creature->position[1] < 2.0)
  {
    creature->status = C_SWOOP_UP; 
    creature->xrot = 0.0;
    creature->beak.xrot = 0.0;
    creature->position[1] = 2.0;
    if (creature->chip != NULL)
      /* pick up the chip if you don't have one */
      if (!creature->holding_chip)
      {
	creature->chip->status = CARRIED;
	VEC_COPY(creature->chip->position, creature->position);
	creature->chip->yrot = creature->yrot;
	creature->holding_chip = TRUE;
      }
      else  /* drop the chip */
	drop_chip(creature);
      
    return;
  }
  
  /* if just starting tilt it down a lot */
  if (creature->position[1] > 3.0)
  {
    creature->beak.xrot -= howmuch * 2;
    creature->xrot -= howmuch;
  }
  else   /* else begin to tilt it back up */
  {
    creature->beak.xrot += howmuch * 2;
    creature->xrot += howmuch;
  }
  
  /* move it down a bit */
  creature->position[1] -= (.2*howmuch);
}

/* move 'howmuch' up */
void
  creature_up(CREATURE *creature, float howmuch)
{
  /* if it's up high enough then return to normal flight */
  if (creature->position[1] > 3.7)
  {
    creature->status = C_NORMAL; 
    creature->xrot = 0.0;
    creature->beak.xrot = 0.0;
    creature->position[1] = 4.0;
    return;
  }

  if (creature->position[1] < 3.0)
  {
    creature->beak.xrot += howmuch*2;
    creature->xrot += howmuch;
  }
  else   /* else begin to tilt it back up */
  {
    creature->beak.xrot -= howmuch *2;
    creature->xrot -= howmuch;
  }
  
  /* move it up a bit */
  creature->position[1] += (.2*howmuch);
}
  
/* close beak 'howmuch */
void 
  beak_close(CREATURE *creature, float howmuch)
{
  creature->beak.top_xrot -= howmuch*2;
  creature->beak.bottom_xrot += howmuch*2;
}

/* open beak 'howmuch */
void 
  beak_open(CREATURE *creature, float howmuch)
{
  creature->beak.top_xrot += howmuch*2;
  creature->beak.bottom_xrot -= howmuch*2;
}

/* find how much a creature has to rotate to face the chip or creature */
float rot_to_this(float creature_rot, VECTOR chip_posit)
{
  float angle, answer;
  float new_crot;

  new_crot = creature_rot;
  while (new_crot > 360) new_crot -= 360; 
  while (new_crot < 0) new_crot += 360;
  
  angle = DEGREES(atanf(ABS(chip_posit[2]/chip_posit[0])));
  if (chip_posit[2] > 0)
  {
    if (chip_posit[0] < 0) /* 2nd quadrant */
      angle = 180 - angle;
  }
  else
  {
    if (chip_posit[0] < 0)  /* 3rd quadrant */
      angle += 180;
    else                /* 4th quadrant */
      angle = 360 - angle;  
  }
  
  answer = angle - new_crot;
  
  if (answer > 180) answer = answer - 360; /* we want the little angle */
  else if (answer < -180) answer = answer + 360;
  
  return(-answer); /* creature's coordinate system is backwards */
}



/* cycle through the chips seeing if any are close, and if so, which
   of these is closest */
float
close_to_chip(CREATURE *creature, CHIPS **chip)
{
  short i, j;
  VECTOR chip_direction;
  float closest = INFINITY;
  float dist;
  float rot;

  bestrot = 80;
  
  for (i = 0; i < num_players; i++)
  {
    for (j = 0; j < CHIPS_PER_PLAYER; j++)
    {
      /* ignore chips in the hole and on other players */
      if (chips[i][j]->status == ON_GROUND)
      {
	dist = VEC_DIST(creature->position, chips[i][j]->position);     
	if ((dist < 15) && (dist < closest))
	{
	  VEC_SUB(chip_direction, chips[i][j]->position, creature->position);
	  chip_direction[1] = 0;
	  if(ABS(rot = rot_to_this(-(creature->yrot + 90),
				   chip_direction)) < 50)
	  {
	    
	    if (ABS(creature->beak.yrot - rot) < 
		ABS(creature->beak.yrot - bestrot))
	    {
	      bestrot = rot;
	      closest = dist;
	      *chip = chips[i][j];
	    }
	  }
	}
      }
    }
  }

  
  if (closest < 15)
    return(bestrot);

  return(80);
}

      
void
creature_swoop()
{
  CREATURE *creature = &creatures[me];
  float rot;
  int i;
  PLAYER player;
  
  if( !creature->it)
  {
    
    if(creature->status != C_SWOOP_UP)
      creature->status = C_SWOOP_DOWN;
    
    /* if creature is close enough to a chip to pick it up */
    if ((creature->holding_chip == FALSE) && 
	((rot = close_to_chip(creature, &creature->chip)) < 80))
    {
      creature->beak.yrot = rot;

      /* see if it's the restart chip */
      if (creatures[creature->chip->indexi].done) 
      {
	/* tell the server to reset the chips */
	client_send_endgame();
      }
	
      if (ABS(rot) > 5)  
      {
	/*      creature->beak.yrot = 0; */
	
	creature->yrot += creature->beak.yrot/4; /* swinging body */
	creature->zrot = creature->beak.yrot/6; /* tilting in the 
						   direction of the turn */
	creature->beak.yrot = SIGN(creature->beak.yrot)*30;

      } 
    }
  }
  else /* you're it */
  {
    /* you can only tagh if you've been it for a decent length of time */
    if (time_it(STOP_TIMER, TIMER_IT) > 4) 
    {
      for (i=0; i<num_players; i++)
      {
	if ((i != me) && creatures[i].alive)
	{
	  if((VEC_DIST(creatures[i].position, creatures[me].position) < 20) &&
	     (ABS(rot_to_this(creatures[me].yrot + creatures[me].beak.yrot + 90,
			  creatures[i].position)) < 120))
	  {
	    creatures[me].aimat = i;
	    creatures[me].it = C_HIT;
	    creatures[i].it = C_IT;
	  }
	  else creatures[me].it = C_MISS;
	}
      }
    }
    
  }
  
  
}  

void 
creatures_move()
{
  if(creatures[me].status == C_SWOOP_DOWN)
  {
    creature_down(&creatures[me], (creatures[me].speed*howfast));
  }    
  else if (creatures[me].status == C_SWOOP_UP)
  {    
    creature_up(&creatures[me], (creatures[me].speed*howfast));
  }
  
  /* now move forward */
  /* if it's picking up a chip, head for it */
  if ((creatures[me].status == C_SWOOP_DOWN) && 
      (creatures[me].chip != NULL) && !creatures[me].holding_chip
      && !creatures[me].it)
  {
    creature_forward(&creatures[me], 
		     ((VEC_DIST(creatures[me].position, 
				creatures[me].chip->position)/10)*howfast));
  }

  /* move faster if wings going down */
  else if (creatures[me].wings[0].direction < 0)
    creature_forward(&creatures[me], (creatures[me].speed*howfast)*.75); 
  else
    creature_forward(&creatures[me], (creatures[me].speed*howfast)); 
  
  /* move any chips the creature might be holding */
  if (creatures[me].holding_chip)
  {
    VEC_COPY(creatures[me].chip->position, creatures[me].position);
    creatures[me].chip->yrot = creatures[me].yrot;
  }
  
  /* fix the camera so it's always looking at the creature's back */
  facing = creatures[me].yrot - 90;
  VEC_SET(looking, fcos(RADIANS(-facing)), 0, fsin(RADIANS(-facing)));
  camera_update(creatures[me].position[0], creatures[me].position[2]);
  
}
  








