/*
 * Copyright (C) 1994, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */
#include <stdio.h>
#include <malloc.h>
#include <bstring.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/param.h>
#include <netdb.h>
#include <netinet/in.h>

#include "client.h"
#include "config.h"
#include "server.h"
#include "net.h"

int server_pid = -1;

static char name_local[MAXHOSTNAMELEN + 1];

int net_open(int port)
{
  int fd;
  struct sockaddr_in addr;

  fd = socket(AF_INET, SOCK_DGRAM, PF_UNSPEC);
  if (fd == -1)
    return NET_ERROR;

  addr.sin_family = AF_INET;
  addr.sin_port = htons(port);
  addr.sin_addr.s_addr = 0;

  if (bind(fd, &addr, sizeof(struct sockaddr_in)) == -1)
    return NET_ERROR;

  return fd;
}

NET_SELECT *net_select_new()
{
  NET_SELECT *s;

  s = (NET_SELECT *) malloc(sizeof(NET_SELECT));
  if (s == NULL)
    return NULL;

  s->count = 0;
  s->max = 0;
  s->sfd = NULL;
  FD_ZERO(&s->fds_read);

  s->timeout.tv_sec = 0;
  s->timeout.tv_usec = 0;

  return s;
}

void net_select_free(NET_SELECT *s)
{
  free(s->sfd);
  free(s);
}

void net_select_timeout(NET_SELECT *s, int sec, int usec)
{
  s->timeout.tv_sec = sec;
  s->timeout.tv_usec = usec;
}

int net_select_set(NET_SELECT *s, int fd, void (* cb_read)())
{
  int index;
  int i;

  index = -1;
  for (i = 0; i < s->count; i++)
  {
    if (s->sfd[i].fd == -1)
      index = i;

    if (s->sfd[i].fd == fd)
    {
      if (cb_read == NULL)
      {
	s->sfd[i].fd = -1;
	FD_CLR(fd, &s->fds_read);

	return NET_OK;
      }
      else
      {
	index = i;
	break;
      }
    }
  }

  if (cb_read == NULL)
    return NET_OK;

  if (index == -1)
  {
    index = s->count;

    if (s->max == 0)
    {
      s->max = 8;
      s->sfd = (NET_SELECT_FD *) malloc(sizeof(NET_SELECT_FD)*s->max);
    }
    else if (s->count == s->max)
    {
      s->max += 8;
      s->sfd = (NET_SELECT_FD *) realloc(s->sfd, 
	sizeof(NET_SELECT_FD)*s->max);
    }

    s->count++;
  }

  s->sfd[index].fd = fd;
  s->sfd[index].cb_read = cb_read;
  FD_SET(fd, &s->fds_read);

  return NET_OK;
}

int net_select_loop(NET_SELECT *s)
{
  fd_set read_fds;
  struct timeval timeout;
  int count;
  int i;
  NET_SELECT_FD *sfd;

  for (;;)
  {
    read_fds = s->fds_read;

    timeout = s->timeout;
    count = select(FD_SETSIZE, &read_fds, NULL, NULL, &timeout);

    /* timeout */
    if (count == 0)
      return NET_TIMEOUT;

    if (count == -1)
      return NET_ERROR;

    for (i = 0; i < s->count; i++)
    {
      sfd = &s->sfd[i];

      if (FD_ISSET(sfd->fd, &read_fds))
	sfd->cb_read(sfd->fd);
    }
  }
}

int net_addr_lookup(struct sockaddr_in* addr, char *name)
{
  struct hostent *hp, *hp_local;
  int local;

  gethostname(name_local, MAXHOSTNAMELEN);
  if (strcasecmp(name, "localhost") == 0)
    name = name_local;
  hp_local = gethostbyname(name_local);
  if (hp_local == NULL)
    return NET_ERROR;
  strcpy(name_local, hp_local->h_name);
  hp = gethostbyname(name);
  if (hp == NULL)
    return NET_ERROR;

  addr->sin_family = AF_INET;
  addr->sin_port = htons(SERVER_PORT);
  bcopy(hp->h_addr, &addr->sin_addr, sizeof(addr->sin_addr));
  
  if (strcmp(hp->h_name, name_local) == 0)
    return NET_LOCAL;

  return NET_REMOTE;
}

void net_init(char *server_name)
{
  struct sockaddr_in addr;
  int fd;
  int result;

  result = net_addr_lookup(&addr, server_name);

  if (result == NET_ERROR)
  {
    fprintf(stderr, "tagh: Can't find address for \"%s\".\n", server_name);
    exit(-1);
  }

  printf("tagh: Trying to connect to game on \"%s\"...\n", server_name);

  if (result == NET_LOCAL)
  {
    fd = net_open(SERVER_PORT);
    if (fd != NET_ERROR)
    {
      close(fd);
      if (server_init() == SERVER_ERROR) 
      {
	fprintf(stderr, "tagh: Can't initialize server.\n");
	exit(-1);
      }
    }
  }

  if (client_init(&addr) == CLIENT_ERROR)
  {
    fprintf(stderr, "tagh: Can't initialize client.\n");
    exit(-1);
  }
}






