/*
 * Copyright (C) 1994, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>

#include "packet.h"

int server_packet_send(int fd, struct sockaddr_in *addr,
	SERVER_PACKET *server_packet)
{
  return sendto(fd, server_packet, sizeof(SERVER_PACKET), 0, addr,
		sizeof(struct sockaddr_in));
}

int server_packet_recv(int fd, struct sockaddr_in *addr,
	SERVER_PACKET *server_packet)
{
  int fromlen;
  
  fromlen = sizeof(struct sockaddr_in);
  if (recvfrom(fd, server_packet, sizeof(SERVER_PACKET), 0, addr, &fromlen) == -1)
    return PACKET_ERROR;

  return PACKET_OK;
}

int client_packet_send(int fd, struct sockaddr_in *addr,
	CLIENT_PACKET *client_packet)
{
  return sendto(fd, client_packet, sizeof(CLIENT_PACKET), 0, addr,
		sizeof(struct sockaddr_in));
}

int client_packet_recv(int fd, struct sockaddr_in *addr,
	CLIENT_PACKET *client_packet)
{
  int fromlen;

  fromlen = sizeof(struct sockaddr_in);
  if (recvfrom(fd, client_packet, sizeof(CLIENT_PACKET), 0, addr, &fromlen) == -1)
    return PACKET_ERROR;

  return PACKET_OK;
}

int server_opcode_send(int fd, struct sockaddr_in *addr, int opcode)
{
  SERVER_PACKET server_packet;

  server_packet.opcode = opcode;
  return server_packet_send(fd, addr, &server_packet);
}

int server_opcode_recv(int fd, struct sockaddr_in *addr)
{
  SERVER_PACKET server_packet;

  if (server_packet_recv(fd, addr, &server_packet) == PACKET_ERROR)
    return PACKET_ERROR;

  return server_packet.opcode;
}

int client_opcode_send(int fd, struct sockaddr_in *addr, int opcode)
{
  CLIENT_PACKET client_packet;

  client_packet.opcode = opcode;
  return client_packet_send(fd, addr, &client_packet);
}

int client_opcode_recv(int fd, struct sockaddr_in *addr)
{
  CLIENT_PACKET client_packet;

  if (client_packet_recv(fd, addr, &client_packet) == PACKET_ERROR)
    return PACKET_ERROR;

  return client_packet.opcode;
}
