/*
 * Copyright (C) 1994, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */
#ifndef _VECTOR
#define _VECTOR

#include <math.h>

#define VEC_EXPAND(x) x[0], x[1], x[2]

#define VEC_DOT(x, y) (x[0]*y[0] + x[1]*y[1] + x[2]*y[2])
#define VEC_LEN(x) (sqrt(VEC_DOT(x, x)))
#define VEC_DIST(x, y) (sqrt(SQR(x[0]-y[0]) + SQR(x[2]-y[2])))  
#define VEC_DIST_SQR(x, y) (SQR(x[0]-y[0]) + SQR(x[2]-y[2]))  
#define VEC_SET(x, a, b, c) x[0] = a, x[1] = b, x[2] = c
#define VEC_ZERO(x) VEC_SET(x, 0.0, 0.0, 0.0)
#define VEC_COPY(y, x) y[0] = x[0], y[1] = x[1], y[2] = x[2]
#define VEC_NEG(x) x[0] = -x[0], x[1] = -x[1], x[2] = -x[2]
#define VEC_ADD(z, x, y) z[0] = x[0] + y[0], z[1] = x[1] + y[1], \
                z[2] = x[2] + y[2]
#define VEC_SUB(z, x, y) z[0] = x[0] - y[0], z[1] = x[1] - y[1], \
                z[2] = x[2] - y[2]
#define VEC_MULT(x, a) x[0] *= a, x[1] *= a, x[2] *= a
#define VEC_DIV(x, a) x[0] /= a, x[1] /= a, x[2] /= a
#define VEC_ADDS(z, x, a, y) z[0] = x[0] + (a)*y[0], z[1] = x[1] + (a)*y[1], \
                z[2] = x[2] + (a)*y[2]
#define VEC_NORM(x) { float len = VEC_LEN(x); \
			if (len < EPSILON) \
			  VEC_SET(x, 0.0, 1.0, 0.0); \
			else \
			  VEC_DIV(x, len); }
#define VEC_CROSS(z, x, y) z[0] = x[1]*y[2] - x[2]*y[1], \
                z[1] = x[2]*y[0] - x[0]*y[2], z[2] = x[0]*y[1] - x[1]*y[0]

typedef float VECTOR[3];

#endif
