/*
 * Copyright (C) 1994, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include "config.h"
#include "packet.h"
#include "opcodes.h"
#include "clock.h"
#include "game.h"
#include "world.h"
#include "defs.h"
#include "creatures.h"

WORLD world;		/* the state of the world */
extern CREATURE creatures[MAX_PLAYERS];
extern CHIPS *chips[MAX_PLAYERS][CHIPS_PER_PLAYER];
extern short me;
extern short num_players;

static float last_time;

void update_player_info(PLAYER *player, short index)
{
  CHIPS *chip;

  player->index = index;
  player->alive = creatures[index].alive;

  VEC_COPY(player->creat_posit, creatures[index].position);
  player->creat_xrot = creatures[index].xrot;
  player->creat_yrot = creatures[index].yrot;
  player->creat_zrot = creatures[index].zrot;
  player->creat_beak_xrot = creatures[index].beak.xrot;
  player->creat_eyes_yrot[0] = creatures[index].eyes[0].yrot;
  player->creat_eyes_yrot[1] = creatures[index].eyes[1].yrot;
  player->creat_speed = creatures[index].speed;
  player->creat_status = creatures[index].status;
  player->creat_aimat = creatures[index].aimat;
  player->creat_it    = creatures[index].it; 

  if (creatures[index].chip != NULL)
  {
    chip = creatures[index].chip;
    player->chip = 1;
    player->chip_indexi = chip->indexi;
    player->chip_indexj = chip->indexj;
    VEC_COPY(player->chip_position, chip->position);
    player->chip_yrot = chip->yrot;
    player->chip_status = chip->status;
  }
  else player->chip = 0; 

}

    
  
void world_init()
{
  int i, j;
  
/*  last_time = system_clock(); */
/*  world.creat_it[1] = C_IT; */

  for (i=0; i<MAX_PLAYERS; i++)
  {
    VEC_COPY(world.creat_posit[i], creatures[i].position);
    world.creat_xrot[i] = creatures[i].xrot;
    world.creat_yrot[i] = creatures[i].yrot;
    world.creat_zrot[i] = creatures[i].zrot;
    world.creat_beak_xrot[i] = creatures[i].beak.xrot;
    world.creat_eyes_yrot[i][0] = creatures[i].eyes[0].yrot;
    world.creat_eyes_yrot[i][1] = creatures[i].eyes[1].yrot;
    world.creat_speed[i] = creatures[i].speed;
    world.creat_status[i] = creatures[i].status;
    world.creat_alive[i] = creatures[i].alive;

    for (j=0; j<4; j++)
    {
      VEC_COPY(world.chip_position[i][j], chips[i][j]->position);
      VEC_COPY(world.chip_final_position[i][j], chips[i][j]->final_position);
      world.chip_yrot[i][j] = chips[i][j]->yrot;
      world.chip_status[i][j] = chips[i][j]->status;
    }
  }
}


void world_send_update(int fd, struct sockaddr_in *addr)
{
  SERVER_PACKET server_packet;

  server_packet.opcode = OP_WORLD_UPDATE;
  server_packet.world = world;
/*  game_update_world(&server_packet.world, 
	(system_clock() - last_time)*TICKS_PER_SECOND); */
  server_packet_send(fd, addr, &server_packet);
}

void world_receive_update(WORLD *new_world)
{
  int i,j;
  int alive = 0;
  world = *new_world;

  num_players = world.num_players;
  
  for (i=0; i<num_players; i++)
  {
    if (i!=me)
    {
      VEC_COPY(creatures[i].position, world.creat_posit[i]);
      creatures[i].alive = world.creat_alive[i];
      creatures[i].xrot = world.creat_xrot[i];
      creatures[i].yrot = world.creat_yrot[i];
      creatures[i].zrot = world.creat_zrot[i];
      creatures[i].beak.xrot = world.creat_beak_xrot[i];
      creatures[i].beak.yrot = world.creat_beak_yrot[i];
      creatures[i].eyes[0].yrot = world.creat_eyes_yrot[i][0];
      creatures[i].eyes[1].yrot = world.creat_eyes_yrot[i][1];
      creatures[i].speed = world.creat_speed[i];
      creatures[i].status = world.creat_status[i];
      creatures[i].aimat = world.creat_aimat[i];
    }
    creatures[i].done = world.creat_done[i];

    for (j=0; j<num_players; j++)  /* how many are still up and going? */
      if (creatures[i].alive) alive++;

    if (alive > 1)
    {
      if (world.creat_it[i] == C_IT)
	/* if my creature wasn't it before start the timer */
	if ((!creatures[i].it) && (i==me))
	  time_it(START_TIMER, TIMER_IT);
      creatures[i].it = world.creat_it[i];
    }
    
    else
      creatures[i].it = C_NOTIT;
    
    for (j=0; j<4; j++)
    {
      VEC_COPY(chips[i][j]->position, world.chip_position[i][j]); 
      VEC_COPY(chips[i][j]->final_position, world.chip_final_position[i][j]); 
      chips[i][j]->yrot = world.chip_yrot[i][j];
      chips[i][j]->status = world.chip_status[i][j]; 
    }

  }
  
}

void player_send_update(int fd, struct sockaddr_in *addr, PLAYER *player)
{
  CLIENT_PACKET client_packet;
  
  client_packet.opcode = OP_PLAYER_UPDATE;
  client_packet.item.player = *player;
  client_packet_send(fd, addr, &client_packet);
}

void player_send_game_update(int fd, struct sockaddr_in *addr, int whosit)
{
  CLIENT_PACKET client_packet;

  client_packet.opcode = OP_GAME_UPDATE;
  client_packet.item.it = whosit; 
  client_packet_send(fd, addr, &client_packet); 
}  

void player_send_chip_update(int fd, struct sockaddr_in *addr, CHIP_INFO info)
{
  CLIENT_PACKET client_packet;

  client_packet.opcode = OP_CHIP_UPDATE;
  client_packet.item.chip = info;
  client_packet_send(fd, addr, &client_packet); 
}
  
void player_receive_update(PLAYER *player)
{
  game_update_player(&world, player);
}

void world_update()
{
  float time_now;

  time_now = system_clock();
  game_update_world(&world, (time_now - last_time)*TICKS_PER_SECOND);
  last_time = time_now;
}

/* game has ended. reset the chips */
void world_renew()
{
  int i, j;
  
  chips_position();
  for (i=0; i<MAX_PLAYERS; i++)
  {
    world.creat_done[i] = FALSE;
    for (j=0; j<4; j++)
    {
      VEC_COPY(world.chip_position[i][j], chips[i][j]->position);
      world.chip_status[i][j] = chips[i][j]->status;
    }
  }
  
}


