/*
 * Copyright (C) 1994, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>

#include <X11/Intrinsic.h>

#include <GL/gl.h>

#include "vroom.h"
#include "body.h"
#include "car.h"
#include "track.h"
#include "ogl.h"
#include "quickcull.h"


#define	NOZ_X	( 15.0f)
#define	NOZ_Y	(-12.0f)
#define	NOZ_Z	( 13.0f)
#define	NOZ_L	( 30.0f)
#define	NOZ_R	(  9.0f)

#define	NUM_CHARGES	(16)

extern PlayerStruct	player[MAX_PLAYERS] ;
extern Car		cars[] ;
extern float		collisionRadius2 ;
extern float		speedFactor ;
extern float		currentTime ;
extern int		baseLod ;
extern int		debugOn ;

static GLuint	carMaterial ;
static GLuint	carBody ;
static GLuint	windShieldMaterial ;
static GLuint	windShield ;
static GLuint	exhaust ;
static GLuint	road ;
static GLuint	bolt ;
static qcDetail	carCullDetail ;

#define	WS_WIDTH	(80.0f)
#define	WS_LENGTH	(70.0f)
#define	WS_HEIGHT	(36.0f)

float	carLength ;
float	carWidth ;

static GLfloat3	carSideLine[] = {
			{ -21.3334,  48.47620, 7.52011 },
			{ -25.6667,  42.69330, 12.7771 },
			{ -28.3334,  35.52250, 17.1930 },
			{ -30.0000,  26.96370, 18.6649 },
			{ -30.0000,  16.26880, 17.8238 },
			{ -30.0000,   4.47162, 16.7724 },
			{ -30.0000,  -4.08708, 16.7724 },
			{ -30.0000, -13.55400, 17.8752 },
			{ -30.0000, -22.16880, 19.9163 },
			{ -30.0000, -31.03360, 20.6000 },
			{ -28.6667, -38.86560, 20.1369 },
			{ -28.0000, -41.62280, 18.9163 },
			{ -27.3334, -43.58540, 16.3827 },
			{ -26.5000, -44.58540, 13.0827 },
			} ;
static GLfloat3	tailProfile[4] ;
static GLfloat3	noseProfile[4] ;

/* BEGIN PROTOTYPES -S body.c */
static void     averageArray3( GLfloat3 *dst, GLfloat3 *src0, GLfloat3 *src1,
                    GLfloat3 *src2, int n ) ;
static void     calculateEllipseNorm( GLfloat3 v, GLfloat A, GLfloat B,
                    GLfloat C, GLfloat3 n ) ;
static void     createBody( GLfloat3 *sideLine, int n, int smoothCorner ) ;
static void     createExhaust( int nPieces, GLfloat length, GLfloat radius ) ;
static void     createHubcap( GLfloat r, GLfloat width, int n ) ;
static void     createNozzle( int nPieces, GLfloat xStart, GLfloat yStart,
                    GLfloat zStart, GLfloat length, GLfloat radius ) ;
static void     createStaticCharge( void ) ;
static void     createTailAndNose( int smoothCorner ) ;
static void     createTheBody( int iSteps, int jSteps ) ;
static void     createTheWindShield( int iSteps, int jSteps ) ;
static void     createTire( GLfloat r, GLfloat width, int n, int hasHubcap ) ;
static void     createWindshield( int iSteps, int jSteps ) ;
static void     doDrawAllCars( int n, int *carIndex, int *lod ) ;
static void     moveArray3( GLfloat3 *dst, GLfloat x, GLfloat y, GLfloat z,
                    GLfloat3 *src, int n ) ;
static GLfloat  normalizeVec3( GLfloat3 v ) ;
static void     printArray3( char *label, GLfloat3 *v, int n ) ;
static void     scaleArray3( GLfloat3 *dst, GLfloat x, GLfloat y, GLfloat z,
                    GLfloat3 *src, int n ) ;
static void     windShieldNormal( GLfloat3 n, GLfloat u, GLfloat theta ) ;
static void     windShieldVertex( GLfloat3 p, GLfloat u, GLfloat theta ) ;
/* END PROTOTYPES -S body.c */


/*------------------------------------------------------------------------------
 * Create the display lists for the car.
 *----------------------------------------------------------------------------*/
void
createCar(
	void
	)
{
	int		i ;
	int		k ;

	carLength = carSideLine[0][1] -
			carSideLine[XtNumber( carSideLine ) - 1][1] ;

	carWidth = -2.0f * carSideLine[XtNumber( carSideLine ) / 2][0] ;
	collisionRadius2 = 0.5f * ( carLength * carLength +
					carWidth * carWidth ) ;
	collisionRadius2 *= 1.1f ;

	carBody = glGenLists( MAX_LOD ) ;
	glNewList( carBody, GL_COMPILE ) ;
	createTheBody( 8, 10 ) ;
	createNozzle( 8,  NOZ_X, NOZ_Y, NOZ_Z, NOZ_L, NOZ_R ) ;
	createNozzle( 8, -NOZ_X, NOZ_Y, NOZ_Z, NOZ_L, NOZ_R ) ;
	glEndList() ;

	glNewList( carBody+1, GL_COMPILE ) ;
	createTheBody( 6, 8 ) ;
	createNozzle( 6,  NOZ_X, NOZ_Y, NOZ_Z, NOZ_L, NOZ_R ) ;
	createNozzle( 6, -NOZ_X, NOZ_Y, NOZ_Z, NOZ_L, NOZ_R ) ;
	glEndList() ;

	glNewList( carBody+2, GL_COMPILE ) ;
	createTheBody( 5, 7 ) ;
	createNozzle( 5,  NOZ_X, NOZ_Y, NOZ_Z, NOZ_L, NOZ_R ) ;
	createNozzle( 5, -NOZ_X, NOZ_Y, NOZ_Z, NOZ_L, NOZ_R ) ;
	glEndList() ;

	glNewList( carBody+3, GL_COMPILE ) ;
	createTheBody( 4, 6 ) ;
	createNozzle( 4,  NOZ_X, NOZ_Y, NOZ_Z, NOZ_L, NOZ_R ) ;
	createNozzle( 4, -NOZ_X, NOZ_Y, NOZ_Z, NOZ_L, NOZ_R ) ;
	glEndList() ;

	windShield = glGenLists( MAX_LOD ) ;
	glNewList( windShield, GL_COMPILE ) ;
	createTheWindShield( 8, 10 ) ;
	glEndList() ;
	glNewList( windShield+1, GL_COMPILE ) ;
	createTheWindShield( 6, 8 ) ;
	glEndList() ;
	glNewList( windShield+2, GL_COMPILE ) ;
	createTheWindShield( 4, 6 ) ;
	glEndList() ;
	glNewList( windShield+3, GL_COMPILE ) ;
	createTheWindShield( 4, 4 ) ;
	glEndList() ;

	exhaust = glGenLists( MAX_LOD ) ;
	glNewList( exhaust, GL_COMPILE ) ;
	createExhaust( 8, NOZ_L, NOZ_R ) ;
	glEndList() ;
	glNewList( exhaust+1, GL_COMPILE ) ;
	createExhaust( 6, NOZ_L, NOZ_R ) ;
	glEndList() ;
	glNewList( exhaust+2, GL_COMPILE ) ;
	createExhaust( 5, NOZ_L, NOZ_R ) ;
	glEndList() ;
	glNewList( exhaust+3, GL_COMPILE ) ;
	createExhaust( 4, NOZ_L, NOZ_R ) ;
	glEndList() ;

	if( ( carCullDetail = qcCreateDetail( 0, MAX_LOD - 1, 600.0f, 1750.0f,
		50.0f ) ) == NULL )
	{
		fatalError( "could not allocate memory for culling detail" ) ;
	}

	bolt = glGenLists( NUM_CHARGES ) ;
	for( i = 0 ; i < NUM_CHARGES ; i++ )
	{
		glNewList( bolt + i, GL_COMPILE ) ;
		createStaticCharge() ;
		glEndList() ;
	}
}



/*------------------------------------------------------------------------------
 * Create the car body.
 *----------------------------------------------------------------------------*/
static void
createBody(
	GLfloat3	*sideLine,
	int		n,
	int		smoothCorner
	)
{
	int		i ;
	GLfloat3	*line[5] ;
	GLfloat3	*norms[3] ;
	GLfloat3	v ;

	line[0] = myMalloc( n * sizeof( GLfloat3 ) ) ;
	line[1] = myMalloc( n * sizeof( GLfloat3 ) ) ;
	line[2] = myMalloc( n * sizeof( GLfloat3 ) ) ;
	line[3] = myMalloc( n * sizeof( GLfloat3 ) ) ;
	line[4] = myMalloc( n * sizeof( GLfloat3 ) ) ;
	norms[0] = myMalloc( n * sizeof( GLfloat3 ) ) ;
	norms[1] = myMalloc( n * sizeof( GLfloat3 ) ) ;
	norms[2] = myMalloc( n * sizeof( GLfloat3 ) ) ;

	/*
	 * Top of car.
	 */
	if( smoothCorner )
	{
		scaleArray3( line[0], 0.85f, 1.0f, 1.0f, sideLine, n ) ;
	}
	else
	{
		scaleArray3( line[0], 1.0f, 1.0f, 1.0f, sideLine, n ) ;
	}
	scaleArray3( line[1], -1.0f, 1.0f, 1.0f, line[0], n ) ;
	COPY_VEC3( tailProfile[0], line[0][n-1] ) ;
	COPY_VEC3( noseProfile[0], line[0][0] ) ;

	SUB_VEC3( v, line[0][1], line[0][0] ) ;
	SET_VEC3( norms[0][0], 0.0f, v[2], -v[1] ) ;
	(void)normalizeVec3( norms[0][0] ) ;
	for( i = 1 ; i < n - 1 ; i++ )
	{
		SUB_VEC3( v, line[0][i+1], line[0][i-1] ) ;
		SET_VEC3( norms[0][i], 0.0f, v[2], -v[1] ) ;
		(void)normalizeVec3( norms[0][i] ) ;
	}
	SET_VEC3( norms[0][n-1], 0.0f, -1.0f, 0.0f ) ;

	glBegin( GL_TRIANGLE_STRIP ) ;
	for( i = 0 ; i < n ; i++ )
	{
		glNormal3fv( norms[0][i] ) ;
		glVertex3fv( line[1][i] ) ;
		glVertex3fv( line[0][i] ) ;
	}
	glEnd() ;

	/*
	 * Right side of car.
	 */
	if( smoothCorner )
	{
		scaleArray3( line[2], -1.0f, 1.0f, 0.85f, sideLine, n ) ;
		scaleArray3( line[3], 1.0f, 1.0f, 0.0f, line[2], n ) ;
		moveArray3( line[3], 0.0f, 0.0f, line[2][0][2] - 0.5f, line[3],
				n ) ;
	}
	else
	{
		scaleArray3( line[2], -1.0f, 1.0f, 1.0f, sideLine, n ) ;
		scaleArray3( line[3], 1.0f, 1.0f, 0.0f, line[2], n ) ;
		moveArray3( line[3], 0.0f, 0.0f, 0.85f * line[2][0][2] - 0.5f,
				line[3], n ) ;
	}

	SUB_VEC3( v, line[3][1], line[3][0] ) ;
	SET_VEC3( norms[1][0], -v[1], v[0], 0.0f ) ;
	(void)normalizeVec3( norms[1][0] ) ;
	for( i = 1 ; i < n - 1 ; i++ )
	{
		SUB_VEC3( v, line[3][i+1], line[3][i-1] ) ;
		SET_VEC3( norms[1][i], -v[1], v[0], 0.0f ) ;
		(void)normalizeVec3( norms[1][i] ) ;
	}
	SUB_VEC3( v, line[3][n-1], line[3][n-2] ) ;
	SET_VEC3( norms[1][n-1], -v[1], v[0], 0.0f ) ;
	(void)normalizeVec3( norms[1][n-1] ) ;

	glBegin( GL_TRIANGLE_STRIP ) ;
	for( i = 0 ; i < n ; i++ )
	{
		glNormal3fv( norms[1][i] ) ;
		glVertex3fv( line[3][i] ) ;
		glVertex3fv( line[2][i] ) ;
	}
	glEnd() ;

	scaleArray3( line[2], -1.0f, 1.0f, 1.0f, line[2], n ) ;
	scaleArray3( line[3], -1.0f, 1.0f, 1.0f, line[3], n ) ;
	scaleArray3( norms[1], -1.0f, 1.0f, 1.0f, norms[1], n ) ;
	COPY_VEC3( tailProfile[2], line[2][n-1] ) ;
	COPY_VEC3( tailProfile[3], line[3][n-1] ) ;
	COPY_VEC3( noseProfile[2], line[2][0] ) ;
	COPY_VEC3( noseProfile[3], line[3][0] ) ;

	glBegin( GL_TRIANGLE_STRIP ) ;
	for( i = 0 ; i < n ; i++ )
	{
		glNormal3fv( norms[1][i] ) ;
		glVertex3fv( line[2][i] ) ;
		glVertex3fv( line[3][i] ) ;
	}
	glEnd() ;

	/*
	 * Smooth corner.
	 */
	averageArray3( line[4], sideLine, line[0], line[2], n ) ;
	COPY_VEC3( tailProfile[1], line[4][n-1] ) ;
	COPY_VEC3( noseProfile[1], line[4][0] ) ;

	if( smoothCorner )
	{
		for( i = 0 ; i < n ; i++ )
		{
			norms[2][i][0] = 0.5f * ( norms[0][i][0] +
						  norms[1][i][0] ) ;
			norms[2][i][1] = 0.5f * ( norms[0][i][1] +
						  norms[1][i][1] ) ;
			norms[2][i][2] = 0.5f * ( norms[0][i][2] +
						  norms[1][i][2] ) ;
			(void)normalizeVec3( norms[2][i] ) ;
		}

		glBegin( GL_TRIANGLE_STRIP ) ;
		for( i = 0 ; i < n ; i++ )
		{
			glNormal3fv( norms[0][i] ) ;
			glVertex3fv( line[0][i] ) ;
			glNormal3fv( norms[2][i] ) ;
			glVertex3fv( line[4][i] ) ;
		}
		glEnd() ;
		glBegin( GL_TRIANGLE_STRIP ) ;
		for( i = 0 ; i < n ; i++ )
		{
			glNormal3fv( norms[2][i] ) ;
			glVertex3fv( line[4][i] ) ;
			glNormal3fv( norms[1][i] ) ;
			glVertex3fv( line[2][i] ) ;
		}
		glEnd() ;

		scaleArray3( line[2], -1.0f, 1.0f, 1.0f, line[2], n ) ;
		scaleArray3( line[4], -1.0f, 1.0f, 1.0f, line[4], n ) ;
		scaleArray3( norms[1], -1.0f, 1.0f, 1.0f, norms[1], n ) ;
		scaleArray3( norms[2], -1.0f, 1.0f, 1.0f, norms[2], n ) ;

		glBegin( GL_TRIANGLE_STRIP ) ;
		for( i = 0 ; i < n ; i++ )
		{
			glNormal3fv( norms[2][i] ) ;
			glVertex3fv( line[4][i] ) ;
			glNormal3fv( norms[0][i] ) ;
			glVertex3fv( line[1][i] ) ;
		}
		glEnd() ;
		glBegin( GL_TRIANGLE_STRIP ) ;
		for( i = 0 ; i < n ; i++ )
		{
			glNormal3fv( norms[1][i] ) ;
			glVertex3fv( line[2][i] ) ;
			glNormal3fv( norms[2][i] ) ;
			glVertex3fv( line[4][i] ) ;
		}
		glEnd() ;
	}

	free( line[0] ) ;
	free( line[1] ) ;
	free( line[2] ) ;
	free( line[3] ) ;
	free( line[4] ) ;
	free( norms[0] ) ;
	free( norms[1] ) ;
	free( norms[2] ) ;
}



/*------------------------------------------------------------------------------
 * Create the windshield.
 *----------------------------------------------------------------------------*/
static void
createWindshield(
	int	iSteps,
	int	jSteps
	)
{
	int		i ;
	int		k ;
	int		n = iSteps ;
	GLfloat		theta ;
	GLfloat		*u ;
	GLfloat3	*line[2] ;
	GLfloat3	*norms[2] ;

	u = myMalloc( n * sizeof( GLfloat ) ) ;
	line[0] = myMalloc( n * sizeof( GLfloat3 ) ) ;
	line[1] = myMalloc( n * sizeof( GLfloat3 ) ) ;
	norms[0] = myMalloc( n * sizeof( GLfloat3 ) ) ;
	norms[1] = myMalloc( n * sizeof( GLfloat3 ) ) ;

	for( i = 0 ; i < n ; i++ )
	{
		u[i] = (GLfloat)i / (GLfloat)n ;
		windShieldVertex( line[0][i], u[i], 0.0f ) ;
		windShieldNormal( norms[0][i], u[i], 0.0f ) ;
	}

	for( k = 1 ; k <= jSteps ; k++ )
	{
		theta = k * M_PI / (GLfloat)jSteps ;
		glBegin( GL_TRIANGLE_STRIP ) ;
		glNormal3fv( norms[0][0] ) ;
		glVertex3fv( line[0][0] ) ;
		for( i = 1 ; i < n ; i++ )
		{
			windShieldVertex( line[1][i], u[i], theta ) ;
			windShieldNormal( norms[1][i], u[i], theta ) ;
			glNormal3fv( norms[1][i] ) ;
			glVertex3fv( line[1][i] ) ;
			glNormal3fv( norms[0][i] ) ;
			glVertex3fv( line[0][i] ) ;
			line[0][i][0] = line[1][i][0] ;
			line[0][i][1] = line[1][i][1] ;
			line[0][i][2] = line[1][i][2] ;
			norms[0][i][0] = norms[1][i][0] ;
			norms[0][i][1] = norms[1][i][1] ;
			norms[0][i][2] = norms[1][i][2] ;
		}
		glEnd() ;
	}

	free( line[0] ) ;
	free( line[1] ) ;
	free( norms[0] ) ;
	free( norms[1] ) ;
}



/*------------------------------------------------------------------------------
 * Create the tail and nose pieces.
 *----------------------------------------------------------------------------*/
static void
createTailAndNose(
	int	smoothCorner
	)
{
return ;
	glBegin( GL_TRIANGLE_STRIP ) ;
	glNormal3f( 0.0f, -1.0f, 0.0f ) ;
	glVertex3f( -tailProfile[0][0], tailProfile[0][1], tailProfile[0][2] ) ;
	glVertex3fv( tailProfile[0] ) ;
	if( smoothCorner )
	{
		glVertex3f( -tailProfile[1][0], tailProfile[1][1],
				tailProfile[1][2] ) ;
		glVertex3fv( tailProfile[1] ) ;
		glVertex3f( -tailProfile[2][0], tailProfile[2][1],
				tailProfile[2][2] ) ;
		glVertex3fv( tailProfile[2] ) ;
	}
	glVertex3f( -tailProfile[3][0], tailProfile[3][1], tailProfile[3][2] ) ;
	glVertex3fv( tailProfile[3] ) ;
	glEnd() ;

	glBegin( GL_TRIANGLE_STRIP ) ;
	glNormal3f( 0.0f, -1.0f, 0.0f ) ;
	glVertex3fv( noseProfile[0] ) ;
	glVertex3f( -noseProfile[0][0], noseProfile[0][1], noseProfile[0][2] ) ;
	if( smoothCorner )
	{
		glVertex3fv( noseProfile[1] ) ;
		glVertex3f( -noseProfile[1][0], noseProfile[1][1],
				noseProfile[1][2] ) ;
		glVertex3fv( noseProfile[2] ) ;
		glVertex3f( -noseProfile[2][0], noseProfile[2][1],
				noseProfile[2][2] ) ;
	}
	glVertex3fv( noseProfile[3] ) ;
	glVertex3f( -noseProfile[3][0], noseProfile[3][1], noseProfile[3][2] ) ;
	glEnd() ;
}



/*------------------------------------------------------------------------------
 * Create a tire.
 *----------------------------------------------------------------------------*/
static void
createTire(
	GLfloat	r,
	GLfloat	width,
	int	n,
	int	hasHubcap
	)
{
	int		i ;
	int		j ;
	GLfloat		f = 0.5f ;
	GLfloat		theta ;
	GLfloat3	*coords ;
return ;

	coords = myMalloc( n * sizeof( GLfloat3 ) ) ;

	for( i = 0 ; i < n ; i++ )
	{
		theta = ( i * 2.0f * (float)M_PI ) / n ;

		coords[i][0] = width ;
		coords[i][1] = r * cosf( theta ) ;
		coords[i][2] = r * sinf( theta ) ;
	}

	glShadeModel( GL_FLAT ) ;
	glBegin( GL_TRIANGLE_STRIP ) ;
	glNormal3f( 1.0f, 0.0f, 0.0f ) ;
	if( hasHubcap )
	{
		for( j = 0 ; j <= n ; j++ )
		{
			i = j % n ;
			glVertex3f( coords[i][0], f * coords[i][1],
					f * coords[i][2] ) ;
			glVertex3fv( coords[i] ) ;
		}
		glEnd() ;
	}
	else
	{
		glVertex3fv( coords[0] ) ;
		for( i = 1 ; i < n / 2 ; i++ )
		{
			glVertex3fv( coords[i] ) ;
			glVertex3fv( coords[n-i] ) ;
		}
		if( ( n % 2 ) == 0 )
		{
			glVertex3fv( coords[n/2] ) ;
		}

		glEnd() ;
	}

	glShadeModel( GL_SMOOTH ) ;
	glBegin( GL_TRIANGLE_STRIP ) ;
	for( j = 0 ; j <= n ; j++ )
	{
		i = j % n ;
		theta = ( i * 2.0f * (float)M_PI ) / n ;
		glNormal3f( 0.0f, cosf( theta ), sinf( theta ) ) ;

		glVertex3fv( coords[i] ) ;
		glVertex3f( -coords[i][0], coords[i][1],
				coords[i][2] ) ;
	}
	glEnd() ;

	free( coords ) ;
}



/*------------------------------------------------------------------------------
 * Create a hubcap.
 *----------------------------------------------------------------------------*/
static void
createHubcap(
	GLfloat	r,
	GLfloat	width,
	int	n
	)
{
	int		i ;
	int		j ;
	GLfloat		theta ;
	GLfloat3	*coords ;
return ;

	r *= 0.5f ;

	coords = myMalloc( n * sizeof( GLfloat3 ) ) ;

	for( i = 0 ; i < n ; i++ )
	{
		theta = ( i * 2.0f * (float)M_PI ) / n ;

		coords[i][0] = 0.45f * width ;
		coords[i][1] = r * cosf( theta ) ;
		coords[i][2] = r * sinf( theta ) ;
	}

	glShadeModel( GL_FLAT ) ;
	glBegin( GL_TRIANGLE_FAN ) ;
	glNormal3f( 1.0f, 0.0f, 0.0f ) ;
	glVertex3f( 0.45f * width, 0.0f, 0.0f ) ;
	for( j = 0 ; j <= n ; j++ )
	{
		i = j % n ;
		glVertex3fv( coords[i] ) ;
	}
	glEnd() ;

	glShadeModel( GL_SMOOTH ) ;
	glBegin( GL_TRIANGLE_STRIP ) ;
	for( j = 0 ; j <= n ; j++ )
	{
		i = j % n ;
		theta = ( i * 2.0f * (float)M_PI ) / n ;
		glNormal3f( 0.0f, -cosf( theta ), -sinf( theta ) ) ;

		glVertex3fv( coords[i] ) ;
		glVertex3f( width, coords[i][1], coords[i][2] ) ;
	}
	glEnd() ;

	free( coords ) ;
}



/*------------------------------------------------------------------------------
 * Normalize a vector (return the original length).
 *----------------------------------------------------------------------------*/
static GLfloat
normalizeVec3(
	GLfloat3	v
	)
{
	GLfloat	l ;

	l = sqrtf( v[0]*v[0] + v[1]*v[1] + v[2]*v[2] ) ;
	if( l > 0.0f )
	{
		v[0] /= l ;
		v[1] /= l ;
		v[2] /= l ;
	}

	return( l ) ;
}



/*------------------------------------------------------------------------------
 * Scale an array non-uniformly.
 *----------------------------------------------------------------------------*/
static void
scaleArray3(
	GLfloat3	*dst,
	GLfloat		x,
	GLfloat		y,
	GLfloat		z,
	GLfloat3	*src,
	int		n
	)
{
	int	i ;

	for( i = 0 ; i < n ; i++ )
	{
		dst[i][0] = x * src[i][0] ;
		dst[i][1] = y * src[i][1] ;
		dst[i][2] = z * src[i][2] ;
	}
}



/*------------------------------------------------------------------------------
 * Move an array of GLfloat3's.
 *----------------------------------------------------------------------------*/
static void
moveArray3(
	GLfloat3	*dst,
	GLfloat		x,
	GLfloat		y,
	GLfloat		z,
	GLfloat3	*src,
	int		n
	)
{
	int	i ;

	for( i = 0 ; i < n ; i++ )
	{
		dst[i][0] = x + src[i][0] ;
		dst[i][1] = y + src[i][1] ;
		dst[i][2] = z + src[i][2] ;
	}
}



/*------------------------------------------------------------------------------
 * Average 3 arrays of GLfloat3's.
 *----------------------------------------------------------------------------*/
static void
averageArray3(
	GLfloat3	*dst,
	GLfloat3	*src0,
	GLfloat3	*src1,
	GLfloat3	*src2,
	int		n
	)
{
	int	i ;

	for( i = 0 ; i < n ; i++ )
	{
		dst[i][0] = ( src0[i][0] + src1[i][0] + src2[i][0] ) / 3.0f ;
		dst[i][1] = ( src0[i][1] + src1[i][1] + src2[i][1] ) / 3.0f ;
		dst[i][2] = ( src0[i][2] + src1[i][2] + src2[i][2] ) / 3.0f ;
	}
}



/*------------------------------------------------------------------------------
 * Print out an array of GLfloat3's.
 *----------------------------------------------------------------------------*/
static void
printArray3(
	char		*label,
	GLfloat3	*v,
	int		n
	)
{
	int	i ;

	printf( "%s:\n", label ) ;
	for( i = 0 ; i < n ; i++ )
	{
		printf( "   %f %f %f\n", v[i][0], v[i][1], v[i][2] ) ;
	}
}



/*------------------------------------------------------------------------------
 * Define car materials.
 *----------------------------------------------------------------------------*/
void
defineCarMaterials(
	void
	)
{
	static GLfloat	matSpec[] = { 0.4, 0.4, 0.4, 1.0 } ;
	static GLfloat	matShiny[] = { 50.0 } ;
	static GLfloat	matAmbient[] = { 0.2, 0.0, 0.0, 1.0 } ;
	static GLfloat	matDiffuse[] = { 0.8, 0.0, 0.0, 1.0 } ;

	static GLfloat	wsMatSpec[] = { 0.9, 0.9, 0.9, 1.0 } ;
	static GLfloat	wsMatShiny[] = { 100.0 } ;
	static GLfloat	wsMatAmbient[] = { 0.0, 0.0, 0.0, 1.0 } ;
	static GLfloat	wsMatDiffuse[] = { 0.1, 0.1, 0.1, 1.0 } ;

	/*
	 * Define material values.
	 */
	carMaterial = glGenLists( 1 ) ;
	glNewList( carMaterial, GL_COMPILE ) ;
	glMaterialfv( GL_FRONT, GL_SPECULAR, matSpec ) ;
	glMaterialfv( GL_FRONT, GL_SHININESS, matShiny ) ;
	glMaterialfv( GL_FRONT, GL_AMBIENT, matAmbient ) ;
	glMaterialfv( GL_FRONT, GL_DIFFUSE, matDiffuse ) ;
	glEndList() ;

	windShieldMaterial = glGenLists( 1 ) ;
	glNewList( windShieldMaterial, GL_COMPILE ) ;
	glMaterialfv( GL_FRONT, GL_SPECULAR, wsMatSpec ) ;
	glMaterialfv( GL_FRONT, GL_SHININESS, wsMatShiny ) ;
	glMaterialfv( GL_FRONT, GL_AMBIENT, wsMatAmbient ) ;
	glMaterialfv( GL_FRONT, GL_DIFFUSE, wsMatDiffuse ) ;
	glEndList() ;
}



/*------------------------------------------------------------------------------
 * Calculate a point on the windshield.
 *----------------------------------------------------------------------------*/
static void
windShieldVertex(
	GLfloat3	p,
	GLfloat		u,
	GLfloat		theta
	)
{
	GLfloat	r ;
	GLfloat	c ;
	GLfloat	s ;

	c = cosf( theta ) ;
	s = sinf( theta ) ;

	p[0] = WS_WIDTH * u * ( 1.0f - u ) * c ;
	p[1] = WS_LENGTH * ( u * u * ( 4.0f * u * u - 13.0f ) / 9.0f ) +
			WS_LENGTH * 0.5f - 10.0f ;
	p[2] = WS_HEIGHT * u * ( 1.0f - u ) * s + 15.5f ;
}



/*------------------------------------------------------------------------------
 * Calculate a normal on the windshield.
 *----------------------------------------------------------------------------*/
static void
windShieldNormal(
	GLfloat3	n,
	GLfloat		u,
	GLfloat		theta
	)
{
	GLfloat3	du ;
	GLfloat3	dt ;
	GLfloat		c ;
	GLfloat		s ;

	if( u == 0.0f )
	{
		n[0] = 0.0f ;
		n[1] = 1.0f ;
		n[2] = 0.0f ;
		return ;
	}

	c = cosf( theta ) ;
	s = sinf( theta ) ;

	du[0] = WS_WIDTH * ( 1.0f - 2.0f * u ) * c ;
	du[1] = WS_LENGTH * u * ( 16.0f * u * u - 26.0f ) / 9.0f ;
	du[2] = WS_HEIGHT * ( 1.0f - 2.0f * u ) * s ;

	if( u == 1.0f )
	{
		u = 0.99f ;
	}

	dt[0] = WS_WIDTH * u * ( 1.0f - u ) * s ;
	dt[1] = 0.0f ;
	dt[2] = -WS_HEIGHT * u * ( 1.0f - u ) * c ;

	n[0] = du[1] * dt[2] - du[2] * dt[1] ;
	n[1] = du[2] * dt[0] - du[0] * dt[2] ;
	n[2] = du[0] * dt[1] - du[1] * dt[0] ;
	(void)normalizeVec3( n ) ;
}



/*------------------------------------------------------------------------------
 * Cull and draw all cars.
 *----------------------------------------------------------------------------*/
void
cullDrawAllCars(
	int	start,
	int	n
	)
{
	int	i ;
	int	k ;
	int	maxCars = 0 ;
	int	index[MAX_PLAYERS] ;
	int	lod[MAX_PLAYERS] ;
	float	v[3] ;

	n += start ;
	for( i = start ; i < n ; i++ )
	{
		v[0] = player[i].x ;
		v[1] = player[i].y ;
		v[2] = player[i].z ;
		if( ( lod[maxCars] = qcCull( v, carCullDetail ) ) != QC_CULLED )
		{
			index[maxCars++] = i ;
		}
	}

	doDrawAllCars( maxCars, index, lod ) ;
}



/*------------------------------------------------------------------------------
 * Draw all cars (no culling).
 *----------------------------------------------------------------------------*/
void
drawAllCars(
	int	start,
	int	n
	)
{
	int		i ;
	int		lod[MAX_PLAYERS] ;
	static int	index[MAX_PLAYERS] = { 0, 1, 2, 3, 4, 5, 6, 7 } ;

	n += start ;
	for( i = start ; i < n ; i++ )
	{
		lod[i] = baseLod ;
	}

	doDrawAllCars( n, &index[start], &lod[start] ) ;
}



/*------------------------------------------------------------------------------
 * Draw all cars.
 *----------------------------------------------------------------------------*/
static void
doDrawAllCars(
	int	n,
	int	*carIndex,
	int	*lod
	)
{
	int	i ;
	int	k ;
	int	l ;
	int	nLo = 0 ;
	int	nHi = 0 ;
	int	loIndex[MAX_PLAYERS] ;
	int	hiIndex[MAX_PLAYERS] ;
	int	loLod[MAX_PLAYERS] ;
	int	hiLod[MAX_PLAYERS] ;
	float	f ;
	float	t ;

	/*
	 * Sort into low and high lod lists for later.
	 */
	for( i = 0 ; i < n ; i++ )
	{
		if( lod[i] < 3 )
		{
			hiLod[nHi] = lod[i] ;
			hiIndex[nHi++] = carIndex[i] ;
		}
		else
		{
			loLod[nLo] = lod[i] ;
			loIndex[nLo++] = carIndex[i] ;
		}
	}

	glEnable( GL_LIGHTING ) ;
	glCallList( carMaterial ) ;
        glEnable( GL_COLOR_MATERIAL ) ;
	for( k = 0 ; k < n ; k++ )
	{
		i = carIndex[k] ;
		glPushMatrix() ;
		glTranslatef( player[i].x, player[i].y, player[i].z ) ;
		glRotatef( player[i].thetaDeg, 0.0f, 0.0f, 1.0f ) ;
		glRotatef( player[i].roll, 0.0f, 1.0f, 0.0f ) ;
		glColor3fv( player[i].color ) ;
		glCallList( carBody+lod[k] ) ;
		glPopMatrix() ;
	}
        glDisable( GL_COLOR_MATERIAL ) ;

	/*
	 * Draw lighted windshield for higher level-of-detail cars.
	 */
	if( nHi > 0 )
	{
		glCallList( windShieldMaterial ) ;
		for( k = 0 ; k < nHi ; k++ )
		{
			i = hiIndex[k] ;
			glPushMatrix() ;
			glTranslatef( player[i].x, player[i].y, player[i].z ) ;
			glRotatef( player[i].thetaDeg, 0.0f, 0.0f, 1.0f ) ;
			glRotatef( player[i].roll, 0.0f, 1.0f, 0.0f ) ;
			glCallList( windShield+hiLod[k] ) ;
			glPopMatrix() ;
		}
	}

	glDisable( GL_LIGHTING ) ;

	/*
	 * Draw flat shaded windshield for low level-of-detail cars.
	 */
	if( nLo > 0 )
	{
		glColor3f( 0.0f, 0.0f, 0.0f ) ;
		for( k = 0 ; k < nLo ; k++ )
		{
			i = loIndex[k] ;
			l = loLod[k] ;
			glPushMatrix() ;
			glTranslatef( player[i].x, player[i].y, player[i].z ) ;
			glRotatef( player[i].thetaDeg, 0.0f, 0.0f, 1.0f ) ;
			glRotatef( player[i].roll, 0.0f, 1.0f, 0.0f ) ;
			glCallList( windShield+l ) ;
			glPopMatrix() ;
		}
	}

	/*
	 * Check for changing lanes.
	 */
	glColor3f( 0.50f, 1.0f, 1.0f ) ;
	glLineWidth( 3.0f ) ;
	for( k = 0 ; k < n ; k++ )
	{
		float	r ;
		float	l ;
		int	nc = rand() % NUM_CHARGES ;
		i = carIndex[k] ;
		if( cars[i].status & CHANGE_LANES )
		{
			glPushMatrix() ;
			glTranslatef( player[i].x, player[i].y, player[i].z ) ;
			glRotatef( player[i].thetaDeg, 0.0f, 0.0f, 1.0f ) ;
			l = floorf( cars[i].lane ) ;
			r = cars[i].track->laneWidth( cars[i].track,
				cars[i].s, cars[i].lane, l ) ;
			glPushMatrix() ;
			glScalef( -r, 1.0f, player[i].z ) ;
			glCallList( bolt + nc ) ;
			glPopMatrix() ;
			l = ceilf( cars[i].lane ) ;
			r = cars[i].track->laneWidth( cars[i].track,
				cars[i].s, cars[i].lane, l ) ;
			glScalef( r, 1.0f, player[i].z ) ;
			glCallList( bolt + nc ) ;
			glPopMatrix() ;
		}
	}
	glLineWidth( 1.0f ) ;

	/*
	 * Draw exhaust.
	 */
	t = 1.5f * ( fmodf( currentTime, 0.2f ) - 0.1f ) ;
	t = ABSFUNC( t ) ;
	glEnable( GL_BLEND ) ;
	for( k = 0 ; k < n ; k++ )
	{
		i = carIndex[k] ;
		f = t + 0.1f + 0.9f*cars[i].desiredSpeed*speedFactor ;
		glColor4f( 1.0f, 0.7f * f, 0.0f, 0.5f ) ;
		glPushMatrix() ;
		glTranslatef( player[i].x, player[i].y, player[i].z ) ;
		glRotatef( player[i].thetaDeg, 0.0f, 0.0f, 1.0f ) ;
		glRotatef( player[i].roll, 0.0f, 1.0f, 0.0f ) ;
		glTranslatef( -NOZ_X, NOZ_Y-NOZ_L, NOZ_Z ) ;
		glScalef( 1.0f, f, 1.0f ) ;
		glCallList( exhaust+lod[k] ) ;
		glTranslatef( NOZ_X+NOZ_X, 0.0f, 0.0f ) ;
		glCallList( exhaust+lod[k] ) ;
		glPopMatrix() ;
	}
	glDisable( GL_BLEND ) ;
}



/*------------------------------------------------------------------------------
 * Set the base lod for the cars.
 *----------------------------------------------------------------------------*/
void
setCarBaseLod(
	int	lod
	)
{
	qcSetLodRange( carCullDetail, lod, MAX_LOD - 1 ) ;
}



/*------------------------------------------------------------------------------
 * Create the ellipsoid body.
 *----------------------------------------------------------------------------*/
static void
createTheBody(
	int	iSteps,
	int	jSteps
	)
{
	int		i ;
	int		j ;
	const GLfloat	W = 30.0f ;
	const GLfloat	L = 50.0f ;
	const GLfloat	H = 20.0f ;
	GLfloat		r ;
	GLfloat		theta ;
	GLfloat		*phi ;
	GLfloat3	*line[2] ;
	GLfloat3	*norms[2] ;

	phi = myMalloc( jSteps * sizeof( GLfloat3 ) ) ;
	line[0] = myMalloc( jSteps * sizeof( GLfloat3 ) ) ;
	line[1] = myMalloc( jSteps * sizeof( GLfloat3 ) ) ;
	norms[0] = myMalloc( jSteps * sizeof( GLfloat3 ) ) ;
	norms[1] = myMalloc( jSteps * sizeof( GLfloat3 ) ) ;

	theta = 0.0f ;
	for( j = 0 ; j < jSteps ; j++ )
	{
		phi[j] = (GLfloat)M_PI * ( 0.1f + 0.8f *
			(GLfloat)j / (GLfloat)(jSteps - 1) ) ;
		r = sinf( phi[j] ) ;
		line[0][j][0] = W * r * cosf( theta ) ;
		line[0][j][1] = L * cosf( phi[j] ) ;
		line[0][j][2] = H * r * sinf( theta ) ;
		calculateEllipseNorm( line[0][j], W, L, H, norms[0][j] ) ;
	}

	for( i = 1 ; i <= iSteps ; i++ )
	{
		theta = i * (GLfloat)M_PI / (GLfloat)iSteps ;
		glBegin( GL_TRIANGLE_STRIP ) ;
		for( j = 0 ; j < jSteps ; j++ )
		{
			phi[j] = (GLfloat)M_PI * ( 0.1f + 0.8f *
				(GLfloat)j / (GLfloat)(jSteps - 1) ) ;
			r = sinf( phi[j] ) ;
			line[1][j][0] = W * r * cosf( theta ) ;
			line[1][j][1] = L * cosf( phi[j] ) ;
			line[1][j][2] = H * r * sinf( theta ) ;
			calculateEllipseNorm( line[1][j], W, L, H,
						norms[1][j] ) ;
			glNormal3fv( norms[0][j] ) ;
			glVertex3fv( line[0][j] ) ;
			glNormal3fv( norms[1][j] ) ;
			glVertex3fv( line[1][j] ) ;
			line[0][j][0] = line[1][j][0] ;
			line[0][j][1] = line[1][j][1] ;
			line[0][j][2] = line[1][j][2] ;
			norms[0][j][0] = norms[1][j][0] ;
			norms[0][j][1] = norms[1][j][1] ;
			norms[0][j][2] = norms[1][j][2] ;
		}
		glEnd() ;
	}

	free( phi ) ;
	free( line[0] ) ;
	free( line[1] ) ;
	free( norms[0] ) ;
	free( norms[1] ) ;
}



/*------------------------------------------------------------------------------
 * Create the ellipsoid windshield.
 *----------------------------------------------------------------------------*/
static void
createTheWindShield(
	int	iSteps,
	int	jSteps
	)
{
	int		i ;
	int		j ;
	const GLfloat	W = 20.0f ;
	const GLfloat	L = 28.0f ;
	const GLfloat	H = 15.0f ;
	const GLfloat	X =  0.0f ;
	const GLfloat	Y = 10.0f ;
	const GLfloat	Z = 10.0f ;
	GLfloat		r ;
	GLfloat		theta ;
	GLfloat		*phi ;
	GLfloat3	*line[2] ;
	GLfloat3	*norms[2] ;

	phi = myMalloc( jSteps * sizeof( GLfloat3 ) ) ;
	line[0] = myMalloc( jSteps * sizeof( GLfloat3 ) ) ;
	line[1] = myMalloc( jSteps * sizeof( GLfloat3 ) ) ;
	norms[0] = myMalloc( jSteps * sizeof( GLfloat3 ) ) ;
	norms[1] = myMalloc( jSteps * sizeof( GLfloat3 ) ) ;

	theta = 0.0f ;
	for( j = 0 ; j < jSteps ; j++ )
	{
		phi[j] = (GLfloat)M_PI * ( 0.0f + 1.0f *
			(GLfloat)j / (GLfloat)(jSteps - 1) ) ;
		r = sinf( phi[j] ) ;
		line[0][j][0] = W * r * cosf( theta ) ;
		line[0][j][1] = L * cosf( phi[j] ) ;
		line[0][j][2] = H * r * sinf( theta ) ;
		calculateEllipseNorm( line[0][j], W, L, H, norms[0][j] ) ;
		line[0][j][0] += X ;
		line[0][j][1] += Y ;
		line[0][j][2] += Z ;
	}

	for( i = 1 ; i <= iSteps ; i++ )
	{
		theta = i * (GLfloat)M_PI / (GLfloat)iSteps ;
		glBegin( GL_TRIANGLE_STRIP ) ;
		for( j = 0 ; j < jSteps ; j++ )
		{
			phi[j] = (GLfloat)M_PI * ( 0.1f + 0.8f *
				(GLfloat)j / (GLfloat)(jSteps - 1) ) ;
			r = sinf( phi[j] ) ;
			line[1][j][0] = W * r * cosf( theta ) ;
			line[1][j][1] = L * cosf( phi[j] ) ;
			line[1][j][2] = H * r * sinf( theta ) ;
			calculateEllipseNorm( line[1][j], W, L, H,
						norms[1][j] ) ;
			line[1][j][0] += X ;
			line[1][j][1] += Y ;
			line[1][j][2] += Z ;
			glNormal3fv( norms[0][j] ) ;
			glVertex3fv( line[0][j] ) ;
			glNormal3fv( norms[1][j] ) ;
			glVertex3fv( line[1][j] ) ;
			line[0][j][0] = line[1][j][0] ;
			line[0][j][1] = line[1][j][1] ;
			line[0][j][2] = line[1][j][2] ;
			norms[0][j][0] = norms[1][j][0] ;
			norms[0][j][1] = norms[1][j][1] ;
			norms[0][j][2] = norms[1][j][2] ;
		}
		glEnd() ;
	}

	free( phi ) ;
	free( line[0] ) ;
	free( line[1] ) ;
	free( norms[0] ) ;
	free( norms[1] ) ;
}



/*------------------------------------------------------------------------------
 * Calculate the normal at a point on an ellipse.
 *----------------------------------------------------------------------------*/
static void
calculateEllipseNorm(
	GLfloat3	v,
	GLfloat		A,
	GLfloat		B,
	GLfloat		C,
	GLfloat3	n
	)
{
	GLfloat	s ;

	n[0] = v[0]/ ( A * A ) ;
	n[1] = v[1]/ ( B * B ) ;
	n[2] = v[2]/ ( C * C ) ;

	s = sqrtf( n[0]*n[0] + n[1]*n[1] + n[2]*n[2] ) ;
	if( s > 0.0f )
	{
		s = 1.0f / s ;
		n[0] *= s ;
		n[1] *= s ;
		n[2] *= s ;
	}
}



/*------------------------------------------------------------------------------
 * Create the jet nozzles.
 *----------------------------------------------------------------------------*/
static void
createNozzle(
	int	nPieces,
	GLfloat	xStart,
	GLfloat	yStart,
	GLfloat	zStart,
	GLfloat	length,
	GLfloat	radius
	)
{
	int		i ;
	GLfloat3	v ;
	GLfloat3	n ;
	float		theta ;
	float		denom ;

	denom = radius * sqrtf( 1.0f + radius * radius / ( length * length ) ) ;
	denom = 1.0f / denom ;

	v[1] = yStart - length ;
	n[1] = radius / sqrtf( length * length + radius * radius ) ;
	glBegin( GL_TRIANGLE_STRIP ) ;
	for( i = 0 ; i <= nPieces ; i++ )
	{
		theta = 2.0f * (float)M_PI * (float)i / (float)( nPieces ) ;
		v[0] = radius * cosf( theta ) ;
		v[2] = radius * sinf( theta ) ;
		n[0] = v[0] * denom ;
		n[2] = v[2] * denom ;
		v[0] += xStart ;
		v[2] += zStart ;
		glNormal3fv( n ) ;
		glVertex3fv( v ) ;
		theta = 2.0f * (float)M_PI * ( (float)i + 0.5f ) /
			(float)( nPieces ) ;
		n[0] = radius * cosf( theta ) * denom ;
		n[2] = radius * sinf( theta ) * denom ;
		glNormal3fv( n ) ;
		glVertex3f( xStart, yStart, zStart ) ;
	}
	glEnd() ;
}



/*------------------------------------------------------------------------------
 * Create the jet exhaust.
 *----------------------------------------------------------------------------*/
static void
createExhaust(
	int	nPieces,
	GLfloat	length,
	GLfloat	radius
	)
{
	int		i ;
	GLfloat3	v ;
	float		theta ;

	v[1] = 0.0f ;

	glBegin( GL_TRIANGLE_FAN ) ;
	glVertex3f( 0.0f, 0.0f, 0.0f ) ;
	for( i = 0 ; i <= nPieces ; i++ )
	{
		theta = 2.0f * (float)M_PI * (float)i / (float)( nPieces ) ;
		v[0] = radius * cosf( theta ) ;
		v[2] = radius * sinf( theta ) ;
		glVertex3fv( v ) ;
	}
	glEnd() ;

	glBegin( GL_TRIANGLE_STRIP ) ;
	for( i = 0 ; i <= nPieces ; i++ )
	{
		glVertex3f( 0.0f, -length, 0.0f ) ;
		theta = 2.0f * (float)M_PI * (float)i / (float)( nPieces ) ;
		v[0] = radius * cosf( theta ) ;
		v[2] = radius * sinf( theta ) ;
		glVertex3fv( v ) ;
	}
	glEnd() ;
}



/*------------------------------------------------------------------------------
 * Create a static discharge.
 *----------------------------------------------------------------------------*/
static void
createStaticCharge(
	void
	)
{
	int		i ;
	const int	N = 20 ;
	const float	Y_OFF = 35.0f ;
	float		d1 = 1.0f / (float)( N - 1 ) ;
	float		d2 = 5.0f / (float)( 0x080 ) ;
	float		d3 = 1.0f / (float)( 0x080 ) ;
	float		x ;
	float		y ;
	float		z ;

	glBegin( GL_LINE_STRIP ) ;
	for( i = 0 ; i < N - 1 ; i++ )
	{
		x = (float)i * d1 ;
		y = (float)( ( rand() % 0x0ff ) - 0x080 ) * d2 -
			( 1.0f - x ) * Y_OFF ;
		z = -x + ( 1.0f - x ) * 0.3f *
				(float)( ( rand() % 0x0ff ) - 0x080 ) * d3 ;
		glVertex3f( x, y, z ) ;
	}
	glVertex3f( 1.0f, 0.0f, -1.0f ) ;
	for( i = N - 2 ; i >= 0 ; i-- )
	{
		x = (float)i * d1 ;
		y = (float)( ( rand() % 0x0ff ) - 0x080 ) * d2 +
			( 1.0f - x ) * Y_OFF ;
		z = -x + ( 1.0f - x ) * 0.3f *
				(float)( ( rand() % 0x0ff ) - 0x080 ) * d3 ;
		glVertex3f( x, y, z ) ;
	}
	glEnd() ;
}
