/*
 * Copyright (C) 1994, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */
#include <stdio.h>
#include <math.h>

#include "vroom.h"
#include "car.h"
#include "ogl.h"
#include "track.h"
#include "solo.h"
#include "sound.h"

#define	WRECK_WAIT_TIME		(3.0f)		/* Seconds */
#define	RESET_DISTANCE		(250.0f)

Car	cars[MAX_PLAYERS] ;


/* BEGIN PROTOTYPES -S car.c */
static void     resetCar( int n, int timeTrial ) ;
static void     stopRoll( Car *car, float dt ) ;
/* END PROTOTYPES -S car.c */

extern float		currentTime ;
extern int		self ;
extern int		nCars ;
extern int		nTracks ;
extern int		nPlayers ;
extern int		debugOn ;
extern PlayerStruct	player[] ;
extern int		updateStatWindow ;
extern Track		*introTrack[] ;
extern Sfx		changeLaneSfx ;
extern Sfx		cheerSfx ;
extern Sfx		crashSfx ;
extern Sfx		motorSfx ;
extern Sfx		squealSfx ;

static const float	rollRate = 270.0f ;

GLfloat			carColors[MAX_PLAYERS][3] = {
				{ 1.0f, 0.0f, 0.0f },
				{ 0.0f, 1.0f, 0.0f },
				{ 0.0f, 0.0f, 1.0f },
				{ 1.0f, 1.0f, 0.0f },
				{ 1.0f, 0.0f, 1.0f },
				{ 0.0f, 1.0f, 1.0f },
				{ 1.0f, 0.3f, 0.0f },
				{ 0.9f, 0.9f, 0.9f },
				} ;

char			*positionString[MAX_PLAYERS] = {
				"first", "second", "third", "fourth",
				"fifth", "sixth", "seventh", "eighth",
				} ;

/*------------------------------------------------------------------------------
 * Handle a wipe out.
 *----------------------------------------------------------------------------*/
void
wipeOutCar(
	Car	*car,
	float	x,
	float	y,
	int	spinDirection,
	int	cause
	)
{
	float	f ;

	if( car->status & CAR_SKID )
	{
		stopSkid( car ) ;
	}
	car->status &= ~( CHANGE_LANE_LEFT | CHANGE_LANE_RIGHT ) ;
	car->status |= CAR_WIPE_OUT ;
	f = 1.0f - 0.75f * car->speed / TOP_SPEED ;
	f = f * f * car->speed ;
	car->wreckVel[0] = f * x ;
	car->wreckVel[1] = f * y ;
	car->spinRate = ( spinDirection == 1 ? -360.0f : 360.0f ) ;
	car->spinAngle = 0.0f ;
	car->wreckTime = currentTime ;
	car->speed = 0.0f ;
	if( car->lane < 1.0f )
	{
		car->lane = 1.0f ;
	}
	else if( car->lane > car->track->lanes )
	{
		car->lane = car->track->lanes ;
	}
	else
	{
		car->lane = (float)rint( (double)car->lane ) ;
	}

	car->status |= cause ;

	if( car->number == self )
	{
		showReason( car->status & WIPE_OUT_REASON ) ;
		sfxPlay( crashSfx ) ;
		sfxDisable( motorSfx ) ;
	}
}



/*------------------------------------------------------------------------------
 * Display a message indicating why car wiped out.
 *----------------------------------------------------------------------------*/
void
showReason(
	int	reason
	)
{
	switch( reason )
	{
		case WIPE_OUT_SPEED :
			setMessage( "You were going too fast." ) ;
			break ;

		case WIPE_OUT_OFF_ROAD :
			setMessage( "You ran off the course." ) ;
			break ;

		default :
			setMessage( "You collided with another player." ) ;
			break ;
	}
}



/*------------------------------------------------------------------------------
 * Initialize the cars.
 *----------------------------------------------------------------------------*/
void
initCarColors(
	void
	)
{
	int	i ;

	for( i = 0 ; i < MAX_PLAYERS ; i++ )
	{
		player[i].color[0] = carColors[i][0] ;
		player[i].color[1] = carColors[i][1] ;
		player[i].color[2] = carColors[i][2] ;
	}
}



/*------------------------------------------------------------------------------
 * Update the cars positions.
 *----------------------------------------------------------------------------*/
void
updateCars(
	float	eTime,
	int	timeTrial
	)
{
	int	i ;
	float	t ;
	float	f ;

	for( i = 0 ; i <  MAX_PLAYERS ; i++ )
	{
		if( cars[i].status & CAR_WIPE_OUT )
		{
			if( currentTime - cars[i].wreckTime > WRECK_WAIT_TIME )
			{
				if( i >= nPlayers )
				{
					cars[i].desiredSpeed = 0.0f ;
				}
				resetCar( i, timeTrial ) ;
			}
			else
			{
				t = ( currentTime - cars[i].wreckTime ) /
					WRECK_WAIT_TIME ;
				t = 1.0f - sqrtf( t ) ;
				player[i].x += t * cars[i].wreckVel[0] * eTime ;
				player[i].y += t * cars[i].wreckVel[1] * eTime ;
				player[i].z += t * TOP_SPEED * eTime ;
				player[i].thetaDeg += cars[i].spinRate * eTime ;
				if( player[i].thetaDeg > 360.0f )
				{
					player[i].thetaDeg -= 360.0f ;
				}
				else if( player[i].thetaDeg < 0.0f )
				{
					player[i].thetaDeg += 360.0f ;
				}
				if( player[i].roll != 0.0f )
				{
					stopRoll( cars+i, eTime ) ;
				}
			}
		}
		else if( ( cars[i].status & CAR_FINISHED ) == 0 )
		{
			if( player[i].roll != 0.0f &&
				( cars[i].status & CAR_SKID ) == 0 )
			{
				stopRoll( cars+i, eTime ) ;
			}
			f = fmodf( cars[i].lane, 1.0f ) ;
			player[i].z = f * ( 1.0f - f ) * 80.0f ;
			cars[i].track->follow( cars[i].track, cars+i, eTime ) ;
			updatePlayerView( i ) ;
		}
	}
}



/*------------------------------------------------------------------------------
 * Initialize the first car for the timing run.
 *----------------------------------------------------------------------------*/
void
setIntroCarShot(
	void
	)
{
	int	i ;

	for( i = 0 ; i < 3 ; i++ )
	{
		cars[i].status = 0 ;
		introTrack[0]->startPosition( introTrack[0], &(cars[i]), i ) ;
		if( i > 0 )
		{
			cars[i].s -= 0.075f ;
			introTrack[0]->follow( introTrack[0], &(cars[i]),
						0.0f ) ;
		}
		cars[i].desiredSpeed = TOP_SPEED ;
		updatePlayerView( i ) ;
	}
	for( ; i < MAX_PLAYERS ; i++ )
	{
		cars[i].status = 0 ;
		introTrack[0]->startPosition( introTrack[0], &(cars[i]),
						i + MAX_PLAYERS ) ;
		updatePlayerView( i ) ;
	}
}



/*------------------------------------------------------------------------------
 * Increment the lap counter for a car.
 *----------------------------------------------------------------------------*/
void
incrementLap(
	Car	*car,
	float	tOffset
	)
{
	int	l ;
	float	lapTime ;

	if( car->track->number == 1 )
	{
		car->lap++ ;
		lapTime = currentTime - tOffset - car->startLapTime ;
		car->finishTime += lapTime ;
		car->startLapTime = currentTime - tOffset ;

		if( lapTime < car->bestLapTime || car->lap == 1 )
		{
			car->bestLapTime = lapTime ;
			/*
			 * Show best lap if this is for self.
			 */
			if( car == cars+self )
			{
				setMessage( "Best lap: %s",
					raceTimeString( lapTime, &l ) ) ;
			}
		}
	}
}



/*------------------------------------------------------------------------------
 * Position the cars based on their best lap time.
 *----------------------------------------------------------------------------*/
void
placeCarsForStart(
	void
	)
{
	int	i ;
	int	ibest ;
	int	l ;
	int	place = 0 ;
	Track	*track ;
	float	bestTime ;

	track = getTrack( 0 ) ;

	for( place = 0 ; place < MAX_PLAYERS ; place++ )
	{
		ibest = -1 ;
		for( i = 0 ; i < MAX_PLAYERS ; i++ )
		{
			if( ( cars[i].status & CAR_FINISHED ) != 0 &&
				( cars[i].bestLapTime < bestTime ||
				ibest == -1 ) )
			{
				ibest = i ;
				bestTime = cars[i].bestLapTime ;
			}
		}
		cars[ibest].track = track ;
		cars[ibest].status &= ~CAR_FINISHED ;
		track->startPosition( track, &cars[ibest], place ) ;
		cars[ibest].spinAngle = 0.0f ;
		updatePlayerView( ibest ) ;
		player[ibest].place = place + 1 ;
		if( ibest == self )
		{
			informPosition( place ) ;
		}
	}
}



/*------------------------------------------------------------------------------
 * Place a car back on the track.
 *----------------------------------------------------------------------------*/
static void
resetCar(
	int	n,
	int	timeTrial
	)
{
	int	j ;
	int	tCurr ;
	int	tPrev ;
	int	lane = cars[n].lane ;
	float	lCurr ;
	float	lPrev ;
	float	dis = RESET_DISTANCE + (float)( rand() % 0x07f ) ;
	Track	*t ;

	if( cars[n].desiredSpeed > 0.10f * TOP_SPEED )
	{
		return ;
	}

	t = cars[n].track ;
	tCurr = t->number ;
	lCurr = t->trackLength( t, lane ) ;

	tPrev = ( tCurr + nTracks - 1 ) % nTracks ;
	t = getTrack( tPrev ) ;
	lPrev = t->trackLength( t, lane ) ;

	while( cars[n].s * lCurr < dis )
	{
		if( tCurr == 1 )
		{
			cars[n].s = 0.0f ;
			dis = 0.0f ;
		}
		else
		{
			dis -= cars[n].s * lCurr ;
			cars[n].track = t ;
			cars[n].s = 1.0f ;
			tCurr = tPrev ;
			lCurr = lPrev ;

			tPrev = ( tCurr + nTracks - 1 ) % nTracks ;
			t = getTrack( tPrev ) ;
			lPrev = t->trackLength( t, lane ) ;
		}
	}
	cars[n].s -= dis / lCurr ;
	cars[n].track->follow( cars[n].track, cars+n, 0.0f ) ;

	if( !timeTrial )
	{
		for( j = 0 ; j < MAX_PLAYERS ; j++ )
		{
			if( j != n && CAR_RACING( cars+j ) )
			{
				if( ( cars[j].track->number == tCurr &&
					cars[j].s < cars[n].s + 0.25f ) ||
					( cars[j].track->number == tPrev &&
					cars[j].lane == cars[n].lane ) )
				{
					return ;
				}
			}
		}
	}

	cars[n].status &= ~( CAR_WIPE_OUT | WIPE_OUT_REASON ) ;
	cars[n].spinAngle = 0.0f ;
	cars[n].speed = 0.0f ;
	player[n].z = 0.0f ;
	cars[n].track->follow( cars[n].track, cars+n, 0.0f ) ;
	updatePlayerView( n ) ;
	if( n == self )
	{
		sfxPlay( changeLaneSfx ) ;
		sfxEnable( motorSfx ) ;
		sfxPlayPitch( motorSfx, cars[self].speed / TOP_SPEED ) ;
		if( timeTrial && cars[self].lap > 0 )
		{
			setMessage( "Best lap: %s",
				raceTimeString( cars[self].bestLapTime, &j ) ) ;
		}
		else
		{
			setMessage( "Car set back on track." ) ;
		}
	}
}



/*------------------------------------------------------------------------------
 * Update the player view based on the car position.
 *----------------------------------------------------------------------------*/
void
updatePlayerView(
	int	i
	)
{
	player[i].x = cars[i].x ;
	player[i].y = cars[i].y ;
	player[i].thetaDeg = cars[i].theta * 180.0f / M_PI ;
	player[i].headingDeg = player[i].thetaDeg ;
}



/*------------------------------------------------------------------------------
 * Print a message indicating the starting position.
 *----------------------------------------------------------------------------*/
void
informPosition(
	int	place
	)
{
	if( place == 0 )
	{
		setMessage( "Congratulations!  You won the pole"
			" position.  Prepare to race." ) ;
		sfxPlay( cheerSfx ) ;
	}
	else
	{
		setMessage( "You qualified for %s position.  "
			"Prepare to race.",
			positionString[place] ) ;
	}
}



/*------------------------------------------------------------------------------
 * Start a car skidding.
 *----------------------------------------------------------------------------*/
void
startSkid(
	Car	*car,
	float	force,
	float	dt
	)
{
	float		newRoll ;
	float		oldRoll ;

	if( car->number == self && ( car->status & CAR_SKID ) == 0 )
	{
		sfxPlay( squealSfx ) ;
	}

	car->status |= CAR_SKID ;

	newRoll = force * 40.0f ;
	oldRoll = player[car->number].roll ;

	if( newRoll > oldRoll )
	{
		oldRoll = oldRoll + rollRate * dt ;
		oldRoll = MINFUNC( oldRoll, newRoll ) ;
	}
	else
	{
		oldRoll = oldRoll - rollRate * dt ;
		oldRoll = MAXFUNC( oldRoll, newRoll ) ;
	}
	player[car->number].roll = oldRoll ;
}



/*------------------------------------------------------------------------------
 * Stop a car skidding.
 *----------------------------------------------------------------------------*/
void
stopSkid(
	Car	*car
	)
{
	car->status &= ~CAR_SKID ;

	if( car->number == self )
	{
		sfxSilenceLoop( squealSfx ) ;
	}
}



/*------------------------------------------------------------------------------
 * Zero out the roll angle of a car.
 *----------------------------------------------------------------------------*/
static void
stopRoll(
	Car	*car,
	float	dt
	)
{
	float		newRoll ;
	float		oldRoll ;

	newRoll = 0.0f ;
	oldRoll = player[car->number].roll ;

	if( newRoll > oldRoll )
	{
		oldRoll = oldRoll + rollRate * dt ;
		oldRoll = MINFUNC( oldRoll, newRoll ) ;
	}
	else
	{
		oldRoll = oldRoll - rollRate * dt ;
		oldRoll = MAXFUNC( oldRoll, newRoll ) ;
	}
	player[car->number].roll = oldRoll ;
}



/*------------------------------------------------------------------------------
 * Indicate that a car has finished.
 *----------------------------------------------------------------------------*/
void
finishCar(
	int	n
	)
{
	cars[n].status |= CAR_FINISHED ;
	cars[n].status &= ~( CAR_WIPE_OUT ) ;
	cars[n].spinAngle = 0.0f ;
	player[n].z = 0.0f ;
	player[n].roll = 0.0f ;
}
