/*
 * Copyright (C) 1994, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <Xm/Xm.h>

#include <GL/glx.h>
#include <GL/GLwMDrawA.h>

#include "vroom.h"
#include "car.h"
#include "track.h"
#include "solo.h"
#include "ogl.h"
#include "playmode.h"
#include "collision.h"
#include "sound.h"

typedef	void	(*SoloFunc)( void ) ;

#define	LANES_CLEAR		0x00
#define	LANES_BLOCKED_LEFT	0x01
#define	LANES_BLOCKED_RIGHT	0x02
#define	LANES_BLOCKED		( LANES_BLOCKED_LEFT | LANES_BLOCKED_RIGHT )
#define	LANES_SHOULD_CHANGE	0x04

#define	SHOULD_GO_RIGHT( st )	((st) == \
				(LANES_SHOULD_CHANGE|LANES_BLOCKED_LEFT))
#define	SHOULD_GO_LEFT( st )	((st) == \
				(LANES_SHOULD_CHANGE|LANES_BLOCKED_RIGHT))
#define	ACCELERATION		TOP_SPEED
#define	ONE_OVER_ACC		(1.0f/ACCELERATION)
				

typedef struct {
	float	smarts ;
	float	courage ;
	float	vision ;
	float	politeness ;
	} RobotSpec ;

/* BEGIN PROTOTYPES -S solo.c */
static int      comparePositions( char *i1, char *i2 ) ;
static float    computerPickLane( int n, Track *track, int *closestN,
                    float *closestDis, int timeTrial ) ;
static float    draftSpeed( int src, int target, float eTime ) ;
static void     showPositions( void ) ;
static void     soloPostRace( void ) ;
static void     soloPostTrial( void ) ;
static void     soloPreRace( void ) ;
static void     soloPreTrial( void ) ;
static void     soloRace( void ) ;
static void     soloTrial( void ) ;
static float    topTrackSpeed( Track *track, float s, float lane,
                    float percent ) ;
static void     updateComputerCars( float eTime, int laps, int timeTrial ) ;
static void     updateSpeeds( float t ) ;
static void     verifyCarS( char *label ) ;
/* END PROTOTYPES -S solo.c */

int			carsFinished ;
int			isRobot[MAX_PLAYERS] ;
int			startLastLap ;
float			startSoundTime ;
#ifdef DEBUG
char			*robotName[MAX_PLAYERS] = {
				"me",
				"Manny Fold (green)",
				"Bernie Rubber (blue)",
				"Otto Maddock (yellow)",
				"Gus Peddle (pink)",
				"Dee Frost (cyan)",
				"Ray De Yaitor (orange)",
				"Heddi Lights (white)",
				} ;
#else
char			*robotName[MAX_PLAYERS] = {
				"me",
				"Manny Fold",
				"Bernie Rubber",
				"Otto Maddock",
				"Gus Peddle",
				"Dee Frost",
				"Ray De Yaitor",
				"Heddi Lights",
				} ;
#endif /* DEBUG */

extern int		debugOn ;
extern int		playMode ;
extern int		nTracks ;
extern int		raceLaps ;
extern int		nCars ;
extern int		nPlayers ;
extern Car		cars[] ;
extern PlayerStruct	player[] ;
extern float		currentTime ;
extern int		self ;
extern Widget		mainOgl ;
extern Widget		localCourseForm ;
extern char		*positionString[] ;
extern float		carLength ;
extern char		*myName ;
extern float		speedFactor ;
extern Sfx		cheerSfx ;
extern Sfx		lastlapSfx ;
extern Sfx		motorSfx ;
extern Sfx		startSfx ;
extern Sfx		toneSfx ;

static int		showRecord ;
static float		maxTimeStep ;
static float		soloStartTime ;
static float		soloLastTime ;
static char		soloPosition[MAX_PLAYERS] ;
static SoloFunc		soloFunc ;

static RobotSpec	robot[MAX_PLAYERS] ;

static float		smarts[MAX_PLAYERS] = {
				0.00f, 0.05f, 0.20f, 0.60f,
				1.00f, 0.10f, 0.05f, 0.50f,
				} ;
static float		courage[MAX_PLAYERS] = {
				0.77f, 0.75f, 0.78f, 0.84f,
				0.90f, 0.87f, 0.81f, 0.72f,
				} ;
static float		vision[MAX_PLAYERS] = {
				0.50f, 0.40f, 0.50f, 1.00f,
				0.80f, 0.60f, 0.40f, 0.50f,
				} ;
static float		politeness[MAX_PLAYERS] = {
				1.00f, 0.70f, 1.00f, 0.60f,
				0.90f, 0.30f, 0.80f, 0.20f,
				} ;



/*------------------------------------------------------------------------------
 * Update the computer cars.
 *----------------------------------------------------------------------------*/
static void
updateComputerCars(
	float	eTime,
	int	laps,
	int	timeTrial
	)
{
	int	i ;
	int	n ;
	int	closestN ;
	float	closestDis ;
	float	thisSpeed ;
	float	nextSpeed ;
	float	s ;
	float	dl ;
	float	dt ;
	float	lane ;
	Track	*thisTrack ;
	Track	*nextTrack ;

	for( i = 0 ; i < MAX_PLAYERS ; i++ )
	{
		if( isRobot[i] == 0 || cars[i].status & CAR_WIPE_OUT )
		{
			continue ;
		}
		else if( cars[i].lap < laps )
		{
			thisTrack = cars[i].track ;
			/*
			 * Check for changing lanes.
			 */
			lane = computerPickLane( i, thisTrack, &closestN,
					&closestDis, timeTrial ) ;

			thisSpeed = topTrackSpeed( thisTrack, cars[i].s,
					lane, robot[i].courage ) ;

			if( timeTrial == 0 && closestDis < robot[i].vision )
			{
				nextSpeed = draftSpeed( i, closestN, eTime ) ;
				thisSpeed = MINFUNC( thisSpeed, nextSpeed ) ;
			}

			nextTrack = thisTrack->nextTrack ;
			nextSpeed = topTrackSpeed( nextTrack, 0.0f, lane,
							robot[i].courage ) ;
			if( nextSpeed < thisSpeed )
			{
				dt = ( thisSpeed - nextSpeed ) * ONE_OVER_ACC ;
				dl = 200.0f + ( 1.0f - cars[i].s ) *
					thisTrack->trackLength( thisTrack,
								lane ) ;
				if( cars[i].speed * dt < dl )
				{
					cars[i].desiredSpeed = nextSpeed ;
				}
			}
			else
			{
				cars[i].desiredSpeed = thisSpeed ;
			}
		}
	}
}



/*------------------------------------------------------------------------------
 * Pick the shortest lane.
 *----------------------------------------------------------------------------*/
static float
computerPickLane(
	int	n,
	Track	*track,
	int	*closestN,
	float	*closestDis,
	int	timeTrial
	)
{
	int	i ;
	int	st = LANES_CLEAR ;
	int	l ;
	float	d ;
	float	dis ;
	float	dl ;
	float	nt ;
	float	maxDis ;
	float	lane ;
	float	*len = track->length ;
	Car	*car = &(cars[n]) ;

	if( car->headOn )
	{
		st = LANES_SHOULD_CHANGE ;
	}
	/*
	 * Check for cars in adjacent lanes if close enough.
	 */
	nt = (float)nTracks ;
	*closestDis = nt ;
	*closestN = n ;
	if( timeTrial == 0 )
	{
		d = robot[n].vision + cars[n].lapDis + nt ;
		maxDis = ( 1.0f + robot[n].politeness ) * robot[n].vision ;
		for( i = 0 ; i < MAX_PLAYERS ; i++ )
		{
			dis = fmodf( d - cars[i].lapDis, nt ) ;
			if( i == n || dis > maxDis || !CAR_RACING( cars+i ) )
			{
				continue ;
			}
			dl = car->lane - cars[i].lane ;
			if( 0.0f < dl && dl < 2.0f )
			{
				st |= LANES_BLOCKED_LEFT ;
			}
			else if( -2.0f < dl && dl < 0.0f )
			{
				st |= LANES_BLOCKED_RIGHT ;
			}
			if( ABSFUNC( dl ) < 1.0f && dis < robot[n].vision )
			{
				dis = robot[n].vision - dis ;
				if( dis < *closestDis )
				{
					*closestN = i ;
					*closestDis = dis ;
				}
				if( cars[i].speed < cars[n].speed )
				{
					st |= LANES_SHOULD_CHANGE ;
				}
			}
		}
	}

	if( car->status & CHANGE_LANE_RIGHT )
	{
		lane = ceilf( car->lane ) ;
		if( lane >= track->lanes )
		{
			lane = track->lanes ;
		}
		st = LANES_BLOCKED ;
	}
	else if( car->status & CHANGE_LANE_LEFT )
	{
		lane = floorf( car->lane ) ;
		if( lane <= 1.0f )
		{
			lane = 1.0f ;
		}
		st = LANES_BLOCKED ;
	}
	else
	{
		if( car->lane == track->lanes )
		{
			st |= LANES_BLOCKED_RIGHT ;
		}
		if( car->lane == 1.0f )
		{
			st |= LANES_BLOCKED_LEFT ;
		}
		lane = car->lane ;
	}

	/*
	 * If lanes aren't blocked on both sides.
	 */
	if( ( st & LANES_BLOCKED ) != LANES_BLOCKED )
	{
		l = (int)car->lane - 1 ;
		if( ( st & LANES_SHOULD_CHANGE ) ||
			randomChoice( robot[n].smarts + car->headOn * 0.75 ) )
		{
			if( l > 0 && ( ( ( ( st & LANES_BLOCKED_LEFT ) == 0 ) &&
				( len[l] > len[l-1] ) ) ||
				SHOULD_GO_LEFT( st ) ) )
			{
				car->status |= CHANGE_LANE_LEFT ;
				lane = car->lane - 1.0f ;
				car->headOn = 0 ;
			}
			else if( l < ( track->lanes - 1 ) &&
				( ( ( ( st & LANES_BLOCKED_RIGHT ) == 0 ) &&
				( len[l] > len[l+1] ) ) ||
				SHOULD_GO_RIGHT( st ) ) )
			{
				car->status |= CHANGE_LANE_RIGHT ;
				lane = car->lane + 1.0f ;
				car->headOn = 0 ;
			}
			else
			{
				lane = car->lane ;
			}
		}
		else
		{
			lane = car->lane ;
		}
	}

	return( lane ) ;
}



/*------------------------------------------------------------------------------
 * Return the top speed allowed on a section of track.
 *----------------------------------------------------------------------------*/
static float
topTrackSpeed(
	Track	*track,
	float	s,
	float	lane,
	float	percent
	)
{
	float	speed ;

	speed = percent * track->topSpeed( track, s, lane ) ;

	return( MINFUNC( speed, TOP_SPEED ) ) ;
}



/*------------------------------------------------------------------------------
 * Main loop for running time trials in solo mode.
 *----------------------------------------------------------------------------*/
void
soloLoop(
	void
	)
{
	soloFunc() ;

	soloLastTime = currentTime ;

	GLwDrawingAreaSwapBuffers( mainOgl ) ;
}



/*------------------------------------------------------------------------------
 * Initialize the solo cars for time trials.
 *----------------------------------------------------------------------------*/
void
initSoloCars(
	int	resetPosition
	)
{
	int		i ;

	maxTimeStep = ( 2.0f * carLength ) / TOP_SPEED ;

	for( i = 0 ; i < MAX_PLAYERS ; i++ )
	{
		if( resetPosition )
		{
			soloPosition[i] = i ;
		}

		cars[i].status = 0 ;
		cars[i].lap = -1 ;
		cars[i].headOn = 0 ;
		cars[i].number = i ;
		cars[i].startLapTime = 0.0f ;
		cars[i].bestLapTime = MAX_RACE_TIME ;
		cars[i].finishTime = 0.0f ;
		cars[i].speed = 0.0f ;
		if( i != self )
		{
			cars[i].desiredSpeed = 0.0f ;
		}
		cars[i].track = getTrack( 0 ) ;
		cars[i].track->startPosition( cars[i].track,
					&cars[i], 0 ) ;
		cars[i].spinAngle = 0.0f ;
		cars[i].totalDis = 0.0f ;
		cars[i].lapDis = 0.0f ;
		player[i].place = i + 1 ;
		player[i].x = cars[i].x ;
		player[i].y = cars[i].y ;
		player[i].z = 0.0f ;
		player[i].thetaDeg = cars[i].theta * 180.0f / M_PI ;
		player[i].roll = 0.0f ;
		player[i].headingDeg = player[i].thetaDeg ;
	}
}



/*------------------------------------------------------------------------------
 * Initialize for running the solo game.
 *----------------------------------------------------------------------------*/
void
initSoloGame(
	int	skipTrial
	)
{
	int	i ;

	setWorkProc( VROOM_WP_SOLO, 1 ) ;

	sfxEnable( motorSfx ) ;
	sfxPlayPitch( motorSfx, 0.0f ) ;
	self = 0 ;
	showRecord = 0 ;
	if( !skipTrial )
	{
		soloFunc = soloPreTrial ;
		setMessage( "Prepare to qualify." ) ;
		nCars = 1 ;
		nPlayers = 1 ;
		initSoloCars( 1 ) ;
		soloStartTime = currentTime + 10.0f ;
		startSoundTime = soloStartTime - 2.0f ;
	}
	else
	{
		initSoloCars( 0 ) ;
		for( i = 0 ; i < MAX_PLAYERS ; i++ )
		{
			cars[i].startLapTime = currentTime ;
			cars[i].finishTime = 0.0f ;
			finishCar( i ) ;
			player[soloPosition[i]].place = i + 1 ;
			cars[soloPosition[i]].bestLapTime = i + 1.0f ;
		}
		carsFinished = 0 ;
		placeCarsForStart() ;
		for( i = 0 ; i < MAX_PLAYERS ; i++ )
		{
			cars[i].lap = -1 ;
			soloPosition[player[i].place - 1 ] = i ;
			cars[i].totalDis = cars[i].s - 1.0f ;
			cars[i].lapDis = cars[i].totalDis ;
			cars[i].status = 0 ;
		}
		nCars = MAX_PLAYERS ;
		soloStartTime = currentTime + 8.0f ;
		startSoundTime = soloStartTime - 2.0f ;
		soloFunc = soloPreRace ;
	}

	strncpy( cars[self].name, myName, sizeof( cars[self].name ) ) ;

	for( i = 1 ; i < MAX_PLAYERS ; i++ )
	{
		isRobot[i] = 1 ;
	}
	isRobot[self] = 0 ;
}



/*------------------------------------------------------------------------------
 * Count down to start position qualification.
 *----------------------------------------------------------------------------*/
static void
soloPreTrial(
	void
	)
{
	int	i ;
	float	dTime ;

	dTime = soloStartTime - currentTime ;
	sfxPlayPitch( motorSfx, 0.5f * cars[self].desiredSpeed * speedFactor ) ;
	if( dTime <= 0.0f )
	{
		sfxPlay( startSfx ) ;
		sfxPlayPitch( motorSfx, cars[self].desiredSpeed *
				speedFactor ) ;
		soloFunc = soloTrial ;
		for( i = 0 ; i < MAX_PLAYERS ; i++ )
		{
			cars[i].startLapTime = currentTime ;
			cars[i].finishTime = 0.0f ;
		}
		carsFinished = 0 ;
		dTime = 0.0f ;
		setMessage( "Best of %d laps is used to qualify.",
				VROOM_TRIAL_LAPS ) ;
	}
	else if( currentTime > startSoundTime )
	{
		sfxPlay( toneSfx ) ;
		startSoundTime += 1.0f ;
	}
	drawIt( self ) ;
	drawPreTrialOverlay( dTime, soloPosition, 1, 0 ) ;
}



/*------------------------------------------------------------------------------
 * Run the time trials.
 *----------------------------------------------------------------------------*/
static void
soloTrial(
	void
	)
{
	float		dTime ;

	dTime = currentTime - soloLastTime ;

	doAllCars( dTime, VROOM_TRIAL_LAPS, 1 ) ;
	checkCarFinishes( VROOM_TRIAL_LAPS, NULL, 0 ) ;
	drawIt( self ) ;
	drawTrialOverlay( currentTime - cars[self].startLapTime,
				soloPosition ) ;
	if( cars[self].status & CAR_FINISHED )
	{
		soloFunc = soloPostTrial ;
		finishCar( self ) ;
		soloStartTime = currentTime ;
		sfxDisable( motorSfx ) ;
		stopSkid( cars+self ) ;
	}
}



/*------------------------------------------------------------------------------
 * After the time trials, wait for all to finish.
 *----------------------------------------------------------------------------*/
static void
soloPostTrial(
	void
	)
{
	int	i ;
	int	l ;
	float	eTime ;

	if( carsFinished == MAX_PLAYERS && currentTime - soloStartTime > 3.0f )
	{
		placeCarsForStart() ;
		for( i = 0 ; i < MAX_PLAYERS ; i++ )
		{
			cars[i].lap = -1 ;
			soloPosition[player[i].place - 1 ] = i ;
			cars[i].totalDis = cars[i].s - 1.0f ;
			cars[i].lapDis = cars[i].totalDis ;
			cars[i].status = 0 ;
		}
		sfxEnable( motorSfx ) ;
		i = soloPosition[0] ;
		showRecord = checkLapRecord( cars[i].bestLapTime,
						cars[i].name, 1 ) ;
		soloFunc = soloPreRace ;
		nCars = MAX_PLAYERS ;
		soloStartTime = currentTime + 5.0f ;
		startSoundTime = soloStartTime - 2.0f ;
	}
	else
	{
		/*
		 * Go ahead and end if computer cars have finished at least
		 * one lap.
		 */
		if( currentTime - soloStartTime > 3.0f )
		{
			for( i = 1 ; i < MAX_PLAYERS ; i++ )
			{
				if( ( cars[i].status & CAR_FINISHED ) == 0 &&
					cars[i].lap > 0 )
				{
					autoFinishCar( i, 0.0f ) ;
				}
			}
		}
		eTime = currentTime - soloLastTime ;
		doAllCars( eTime, VROOM_TRIAL_LAPS, 1 ) ;
		computePositions( soloPosition, VROOM_TRIAL_LAPS ) ;
		checkCarFinishes( VROOM_TRIAL_LAPS, NULL, 0 ) ;
		drawIt( self ) ;
		drawPostTrialOverlay( cars[self].bestLapTime, soloPosition ) ;
	}
}



/*------------------------------------------------------------------------------
 * Count down to start race.
 *----------------------------------------------------------------------------*/
static void
soloPreRace(
	void
	)
{
	int	i ;
	float	dTime ;

	dTime = soloStartTime - currentTime ;
	sfxPlayPitch( motorSfx, 0.5f * cars[self].desiredSpeed * speedFactor ) ;
	if( dTime <= 0.0f )
	{
		sfxPlay( startSfx ) ;
		sfxPlayPitch( motorSfx, cars[self].desiredSpeed *
				speedFactor ) ;
		soloFunc = soloRace ;
		startLastLap = 0 ;
		for( i = 0 ; i < MAX_PLAYERS ; i++ )
		{
			cars[i].startLapTime = currentTime ;
			cars[i].finishTime = 0.0f ;
		}
		carsFinished = 0 ;
		setMessage( "May the best driver win." ) ;
		dTime = 0.0f ;
	}
	else if( currentTime > startSoundTime )
	{
		sfxPlay( toneSfx ) ;
		startSoundTime += 1.0f ;
	}
	drawIt( 0 ) ;
	drawPreTrialOverlay( dTime, soloPosition, 0, showRecord ) ;
}



/*------------------------------------------------------------------------------
 * Run the race.
 *----------------------------------------------------------------------------*/
static void
soloRace(
	void
	)
{
	float	dTime ;
	float	eTime ;
	int	dLaps ;
	int	n ;

	dTime = currentTime - soloStartTime ;
	eTime = currentTime - soloLastTime ;

	doAllCars( eTime, raceLaps, 0 ) ;
	if( cars[self].lap >= raceLaps )
	{
		sfxDisable( motorSfx ) ;
		stopSkid( cars+self ) ;
		soloFunc = soloPostRace ;
		finishCar( self ) ;
		carsFinished++ ;
		cars[self].track = getTrack( 0 ) ;
		computePositions( soloPosition, raceLaps ) ;
		checkCarFinishes( raceLaps, NULL, 0 ) ;
		showFinishMessage() ;
		cars[self].track->endPosition( cars[self].track, &(cars[self]),
						player[self].place ) ;
		updatePlayerView( self ) ;
		drawIt( 0 ) ;
	}
	else
	{
		if( cars[self].lap == raceLaps - 1 && startLastLap == 0 )
		{
			startLastLap = 1 ;
			sfxPlay( lastlapSfx ) ;
		}
		computePositions( soloPosition, raceLaps ) ;
		checkCarFinishes( raceLaps, NULL, 0 ) ;
		drawIt( 0 ) ;
		if( player[self].place == 1 )
		{
			dLaps = (int)( cars[self].totalDis -
				cars[soloPosition[1]].totalDis ) / nTracks ;
			drawRaceOverlay( dTime, raceLaps - cars[self].lap - 1,
				dLaps, 1, soloPosition ) ;
		}
		else
		{
			n = soloPosition[0] ;
			if( cars[n].status & CAR_FINISHED )
			{
				dLaps = raceLaps - cars[self].lap - 1 ;
			}
			else
			{
				dLaps = (int)( cars[n].totalDis -
					cars[self].totalDis ) / nTracks ;
			}
			drawRaceOverlay( dTime, raceLaps - cars[self].lap - 1,
				dLaps, 0, soloPosition ) ;
		}
	}
}



/*------------------------------------------------------------------------------
 * After the race, wait for all to finish.
 *----------------------------------------------------------------------------*/
static void
soloPostRace(
	void
	)
{
	int	i ;
	float	d ;
	float	dTime ;
	float	eTime ;

	dTime = currentTime - soloStartTime ;

	if( carsFinished == MAX_PLAYERS &&
		dTime > cars[self].finishTime + 3.0f )
	{
		setWorkProc( VROOM_WP_SOLO, 0 ) ;
		showSoloResults( soloPosition ) ;
	}
	else
	{
		/*
		 * Go ahead and end if computer cars have finished at least
		 * one lap.
		 */
		if( dTime > cars[self].finishTime + 7.0f )
		{
			for( i = 1 ; i < MAX_PLAYERS ; i++ )
			{
				if( ( cars[i].status & CAR_FINISHED ) == 0 &&
					cars[i].lap > 0 )
				{
					autoFinishCar( i, dTime ) ;
				}
			}
		}
		eTime = currentTime - soloLastTime ;
		doAllCars( eTime, raceLaps, 0 ) ;
		computePositions( soloPosition, raceLaps ) ;
		checkCarFinishes( raceLaps, NULL, 0 ) ;
		drawIt( 0 ) ;
		drawPostRaceOverlay( cars[self].finishTime, dTime,
					soloPosition ) ;
	}
}



/*------------------------------------------------------------------------------
 * Control computer cars and update.
 *----------------------------------------------------------------------------*/
void
doAllCars(
	float	t,
	int	laps,
	int	timeTrial
	)
{
	int	i ;
	int	n ;
	float	t0 ;
	float	dt ;

	/*
	 * If timestep is too great, update at smaller intervals multiple times.
	 * Otherwise, computer cars don't look far enough ahead.
	 */
	if( t > maxTimeStep )
	{
		t0 = ceilf( t / maxTimeStep ) ;
		n = (int)t0 ;
		dt = t / t0 ;
		while( n-- )
		{
			updateComputerCars( dt, laps, timeTrial ) ;
			updateSpeeds( dt ) ;
			updateCars( dt, timeTrial ) ;
		}
	}
	else
	{
		updateComputerCars( t, laps, timeTrial ) ;
		updateSpeeds( t ) ;
		updateCars( t, timeTrial ) ;
	}

	if( !timeTrial )
	{
		checkCollisions() ;
	}
}



/*------------------------------------------------------------------------------
 * Computer the positions of the cars.
 *----------------------------------------------------------------------------*/
void
computePositions(
	char	*position,
	int	laps
	)
{
	int	i ;

	for( i = 0 ; i < MAX_PLAYERS ; i++ )
	{
		if( cars[i].lap >= laps )
		{
			cars[i].lapDis = 0.0f ;
			cars[i].totalDis = nTracks * laps + 99.0f * 60.0f -
						cars[i].finishTime ;
		}
		else
		{
			cars[i].lapDis = (float)( ( cars[i].track->number +
					nTracks - 1 ) % nTracks ) + cars[i].s ;
			cars[i].totalDis = cars[i].lapDis +
					nTracks * cars[i].lap ;
		}
	}
	qsort( position, MAX_PLAYERS, sizeof( char ),
		(int (*)(const void *, const void *))comparePositions ) ;

	for( i = 0 ; i < MAX_PLAYERS ; i++ )
	{
		player[position[i]].place = i + 1 ;
	}
}



/*------------------------------------------------------------------------------
 * Compare the positions of 2 cars (return 1 if car1 behind car2).
 *----------------------------------------------------------------------------*/
static int
comparePositions(
	char	*i1,
	char	*i2
	)
{
	float	d1 = cars[*i1].totalDis ;
	float	d2 = cars[*i2].totalDis ;

	if( d1 > d2 )
	{
		return( -1 ) ;
	}
	else if( d1 < d2 )
	{
		return( 1 ) ;
	}
	else
	{
		return( 0 ) ;
	}
}



/*------------------------------------------------------------------------------
 * Show positions.
 *----------------------------------------------------------------------------*/
static void
showPositions(
	void
	)
{
	int	i ;
	int	n ;

	printf( "\n" ) ;
	printf( "Place Car Lap Track  S    Lane\n" ) ;
	for( i = 0 ; i < MAX_PLAYERS ; i++ )
	{
		n = soloPosition[i] ;
		printf( "  %d    %d  %2d   %2d   %.2f  %.1f %d\n", i, n,
			cars[n].lap, cars[n].track->number, cars[n].s,
			cars[n].lane, player[n].place ) ;
	}
}



/*------------------------------------------------------------------------------
 * debug.
 *----------------------------------------------------------------------------*/
static void
verifyCarS(
	char	*label
	)
{
	int	i ;

	for( i = 0 ; i < MAX_PLAYERS ; i++ )
	{
		if( cars[i].s < 0.0f || cars[i].s > 1.0f )
		{
			printf( "%s: car %d has bad s (%f) track %d lap %d\n",
				label, i, cars[i].s, cars[i].track->number,
				cars[i].lap ) ;
		}
	}
}



/*------------------------------------------------------------------------------
 * Compute the speed necessary to catch the specified car.  (Assumes target is
 * ahead of src.)
 *----------------------------------------------------------------------------*/
static float
draftSpeed(
	int	src,
	int	target,
	float	eTime
	)
{
	int	ns ;
	int	nt ;
	Track	*trackS ;
	Track	*trackT ;
	float	d ;
	float	s ;

	trackS = cars[src].track ;
	trackT = cars[target].track ;

	ns = trackS->number ;
	nt = trackT->number ;

	if( ns == nt )
	{
		d = cars[target].s *
			trackT->trackLength( trackT, cars[target].lane ) -
			cars[src].s *
			trackS->trackLength( trackS, cars[src].lane ) -
			( 2.0f - robot[src].courage ) * carLength ;
	}
	else
	{
		d = cars[target].s *
			trackT->trackLength( trackT, cars[target].lane ) +
			( 1.0f - cars[src].s ) *
			trackS->trackLength( trackS, cars[src].lane ) -
			( 2.0f - robot[src].courage ) * carLength ;
	}
	if( d < 0.0f )
	{
		s = cars[target].speed + robot[src].courage * d / eTime ;
		if( s < 0.05f * TOP_SPEED )
		{
			s = cars[target].speed * 0.80f ;
		}
	}
	else
	{
		s = cars[target].speed + robot[src].courage * d / eTime ;
	}

	return( s ) ;
}



/*------------------------------------------------------------------------------
 * Update the speeds of all cars.
 *----------------------------------------------------------------------------*/
static void
updateSpeeds(
	float	t
	)
{
	int	i ;

	if( cars[self].desiredSpeed != cars[self].speed )
	{
		sfxPlayPitch( motorSfx, cars[self].desiredSpeed *
				speedFactor ) ;
	}

	for( i = 0 ; i < MAX_PLAYERS ; i++ )
	{
		if( CAR_RACING( cars+i ) )
		{
			if( cars[i].speed < cars[i].desiredSpeed )
			{
				cars[i].speed += t * ACCELERATION ;
				if( cars[i].speed > cars[i].desiredSpeed )
				{
					cars[i].speed = cars[i].desiredSpeed ;
				}
			}
			else if( cars[i].speed > cars[i].desiredSpeed )
			{
				cars[i].speed -= t * 1.00f * TOP_SPEED ;
				if( cars[i].speed < cars[i].desiredSpeed )
				{
					cars[i].speed = cars[i].desiredSpeed ;
				}
			}
		}
	}
}




/*------------------------------------------------------------------------------
 * Set computer-controlled cars' characteristics.
 *----------------------------------------------------------------------------*/
float
setRobotChars(
	float	factor
	)
{
	int		i ;
	float		retVal ;
	float		onemf = 1.0f - factor ;
	static float	lastFactor = 0.0f ;

#ifdef DEBUG
printf( "setRobotChars( %.3f )\n", factor ) ;
#endif /* DEBUG */
	for( i = 0 ; i < MAX_PLAYERS ; i++ )
	{
		robot[i].smarts = factor * smarts[i] ;
		robot[i].courage = factor * courage[i] ;
		robot[i].vision = factor * vision[i] ;
		robot[i].politeness = onemf * politeness[i] ;
	}

	retVal = lastFactor ;
	lastFactor =  factor ;

	return( retVal ) ;
}



/*------------------------------------------------------------------------------
 * Initialize robot names.
 *----------------------------------------------------------------------------*/
void
initRobotNames(
	void
	)
{
	int	i ;

	for( i = 0 ; i < MAX_PLAYERS ; i++ )
	{
		strncpy( cars[i].name, robotName[i], sizeof( cars[i].name ) ) ;
	}
}



/*------------------------------------------------------------------------------
 * Display a message indicating the finish position.
 *----------------------------------------------------------------------------*/
void
showFinishMessage(
	void
	)
{
	setMessage( "Race over -- you finished %s.",
			positionString[player[self].place-1] ) ;
	if( player[self].place == 1 )
	{
		sfxPlay( cheerSfx ) ;
	}
}



/*------------------------------------------------------------------------------
 * Give up on current race.
 *----------------------------------------------------------------------------*/
void
giveUpRace(
	void
	)
{
	if( playMode == VROOM_SOLO )
	{
		sfxDisable( motorSfx ) ;
		stopSkid( cars+self ) ;
		if( soloFunc == soloRace )
		{
			carsFinished++ ;
			cars[self].lap = -1 ;
			cars[self].s = 0.0f ;
			cars[self].finishTime = MAX_RACE_TIME ;
			cars[self].status = CAR_FINISHED ;
			player[self].z = 0.0f ;
			player[self].roll = 0.0f ;
			cars[self].track = getTrack( 0 ) ;
			soloStartTime = currentTime ;
			computePositions( soloPosition, raceLaps ) ;
			setMessage( "Race conceded." ) ;
			cars[self].track->endPosition( cars[self].track,
					&(cars[self]), player[self].place ) ;
			updatePlayerView( self ) ;
		}
		setWorkProc( VROOM_WP_SOLO, 0 ) ;
		busyCursor() ;
		pickSkillLevelCB( NULL, localCourseForm, NULL ) ;
		controlMainWindow( 0 ) ;
		setIntroCarShot() ;
		controlIntroWindow( 1 ) ;
		unbusyCursor() ;
	}
	else
	{
		setMessage( "Can't concede in multi-player mode." ) ;
	}
}



/*------------------------------------------------------------------------------
 * Automatically finish a car.
 *----------------------------------------------------------------------------*/
void
autoFinishCar(
	int	i,
	float	raceTime
	)
{
	float	d ;
	int	l ;

	d = (float)raceLaps - cars[i].totalDis / (float)nTracks ;
	cars[i].finishTime = raceTime + d * cars[i].bestLapTime ;
	carsFinished++ ;
	finishCar( i ) ;
	cars[i].track = getTrack( 0 ) ;
	cars[i].track->endPosition( cars[i].track, &(cars[i]),
					player[i].place ) ;
	updatePlayerView( i ) ;
}



/*------------------------------------------------------------------------------
 * Check for cars finishing.
 *----------------------------------------------------------------------------*/
void
checkCarFinishes(
	int	laps,
	float	*timeList,
	int	trial
	)
{
	int	i ;

	for( i = 0 ; i < MAX_PLAYERS ; i++ )
	{
		if( cars[i].lap >= laps &&
			( cars[i].status & CAR_FINISHED ) == 0 )
		{
			finishCar( i ) ;
			cars[i].track = getTrack( 0 ) ;
			cars[i].track->endPosition( cars[i].track, &(cars[i]),
							player[i].place ) ;
			updatePlayerView( i ) ;
			if( timeList )
			{
				if( trial )
				{
					timeList[i] = cars[i].bestLapTime ;
				}
				else
				{
					timeList[i] = cars[i].finishTime ;
				}
			}
			carsFinished++ ;
		}
	}
}
