/*
 * Copyright (C) 1994, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <bstring.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <dirent.h>
#include <errno.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <Xm/Xm.h>
#include <Xm/Label.h>
#include <Xm/List.h>
#include <Xm/Separator.h>
#include <Xm/PushB.h>
#include "Row.h"

#include "vroom.h"
#include "track.h"
#include "car.h"
#include "ogl.h"
#include "server.h"
#include "client.h"
#include "playmode.h"

#include "straight.h"
#include "curved.h"
#include "merge.h"

#define	MAX_SKID_FORCE	( TOP_SPEED * TOP_SPEED / 1200.0f )
#define	MIN_SKID_FORCE	( 0.7f * MAX_SKID_FORCE )
#define	SKID_FORCE_DIV	( 1.0f / ( MAX_SKID_FORCE - MIN_SKID_FORCE ) )


/* BEGIN PROTOTYPES -S track.c */
static CourseNode *         addServerCourse( char *fileName ) ;
static XmStringTable        buildCourseList( struct dirent **courseNames ) ;
static void                 coursePreviewCB( Widget w, Widget list,
                                XtPointer callData ) ;
static int                  courseSelect( struct dirent *file ) ;
static void                 courseSelectionCB( Widget w, Widget list,
                                XtPointer callData ) ;
static void                 drawTrackOutline( void ) ;
static void                 freeCourseList( void ) ;
static void                 freeCourseNames( struct dirent **courseNames ) ;
static char *               getTrackRecordFileName( char *courseName ) ;
static outCoursePacket *    loadTrackIntoPacket( char *fileName ) ;
static void                 loadTrackRecords( outCoursePacket *packet ) ;
static void                 lookForCourses( struct dirent ***courseNames ) ;
static char *               myBaseName( char *fullName ) ;
static void                 saveTrackRecords( outCoursePacket *packet ) ;
static void                 serverCoursePreviewCB( Widget w, Widget list,
                                XtPointer callData ) ;
/* END PROTOTYPES -S track.c */

GLuint			trackOutline = -1 ;
int			nTracks = 0 ;
int			raceLaps = 1 ;

extern int		baseLod ;
extern int		playMode ;
extern Widget		serverCourseLabel ;
extern Widget		localCourseForm ;
extern Widget		serverCourseForm ;
extern char		*defaultDataDir ;
extern char		*defaultRecDir ;

static Track		**course ;
static int		nCourses = 0 ;
static XmStringTable	courseList = NULL ;
static GLfloat		trackBounds[8] ;
static char		*trackDir ;
static char		*recDir ;
static int		trackDirLen ;
static int		recDirLen ;
static outCoursePacket	*currentRacePacket = NULL ;
static char		*courseSelLabels[4] = {
				"Go", "Preview", "Quit", "Help"
				} ;
static char		*trackLoadErrorMsgs[] = {
			"Read `%s'",
			"Error opening `%s'",
			"Error reading first line of `%s'",
			"Course must begin with straight piece.\n(`%s')",
			"Unknown track type in `%s'",
			} ;
static struct {
		Track	*t ;
		int	lod ;
		} *trackDisplayList = NULL ;



#define	TRACK_MIN_X		0
#define	TRACK_MAX_X		1
#define	TRACK_MIN_Y		2
#define	TRACK_MAX_Y		3
#define	TRACK_CENTER_X		4
#define	TRACK_CENTER_Y		5
#define	TRACK_HALF_SIDE_X	6
#define	TRACK_HALF_SIDE_Y	7



/*------------------------------------------------------------------------------
 * Load a track from a course file.
 *----------------------------------------------------------------------------*/
int
loadTrack(
	char	*fileName
	)
{
	int		err = 0 ;
	outCoursePacket	*packet ;

	if( ( packet = loadTrackIntoPacket( fileName ) ) != NULL )
	{
		err = loadTrackFromPacket( packet ) ;
		free( packet ) ;
	}
	else
	{
		err = 1 ;
	}
	
	return( err ) ;
}



/*------------------------------------------------------------------------------
 * Load a track from a course file.
 *----------------------------------------------------------------------------*/
int
loadTrackFromPacket(
	outCoursePacket	*cp
	)
{
	int	i ;
	int	n ;
	int	k ;
	int	dp ;
	int	trackId ;
	int	id ;
	int	nLanes ;
	float	length ;
	float	radius ;
	float	angle ;
	float	arg1 ;
	float	arg2 ;
	FILE	*f ;
	int	err = 0 ;
	Track	*first = NULL ;
	Track	*prev ;
	Track	*track ;
	Track	*t ;
	char	line[80] ;
	
	trackBounds[TRACK_MIN_X] = 0.0f ;
	trackBounds[TRACK_MAX_X] = 0.0f ;
	trackBounds[TRACK_MIN_Y] = 0.0f ;
	trackBounds[TRACK_MAX_Y] = 0.0f ;

	k = 0 ;
	for( n = 0 ; n < cp->size ; n++ )
	{
		dp = n / 4 ;
		k = n % 4 ;

		track = NULL ;
		id = (int)cp->detail[dp].id[k] ;
		arg1 = cp->detail[dp].data[k+k] ;
		arg2 = cp->detail[dp].data[k+k+1] ;

		if( n == 0 )
		{
			if( id != STRAIGHT_TRACK )
			{
				err = 3 ;
			}
			else
			{
				nLanes = (int)( arg1 ) ;
				length = arg2 ;
				track = createStartingTrack( nLanes, length ) ;
			}
		}
		else
		{
			switch( id )
			{
				case STRAIGHT_TRACK :
					nLanes = (int)( arg1 ) ;
					length = arg2 ;
					track = createStraightTrack( nLanes,
							length ) ;
					break ;

				case L_CURVE_TRACK :
					nLanes = (int)( arg1 / 360.0f );
					angle = fmodf( arg1, 360.0f ) ;
					radius = arg2 ;
					track = createLeftCurvedTrack( nLanes,
						radius, angle ) ;
					break ;

				case R_CURVE_TRACK :
					nLanes = (int)( arg1 / 360.0f );
					angle = fmodf( arg1, 360.0f ) ;
					radius = arg2 ;
					track = createRightCurvedTrack( nLanes,
						radius, angle ) ;
					break ;

				case MERGE_IN_TRACK :
					nLanes = (int)( arg1 ) ;
					length = arg2 ;
					track = createMergeInTrack( nLanes+2,
							length ) ;
					break ;

				case MERGE_OUT_TRACK :
					nLanes = (int)( arg1 ) ;
					length = arg2 ;
					track = createMergeOutTrack( nLanes,
							length ) ;
					break ;

				default :
					err = 4 ;
					break ;
			}
		}

		if( track != NULL )
		{
			if( first == NULL )
			{
				first = track ;
				track->x = 0.0f ;
				track->y = 0.0f ;
				track->theta = 0.0f ;
			}
			else
			{
				prev->add( prev, track ) ;
				if( track->x < trackBounds[TRACK_MIN_X] )
				{
					trackBounds[TRACK_MIN_X] = track->x ;
				}
				else if( track->x > trackBounds[TRACK_MAX_X] )
				{
					trackBounds[TRACK_MAX_X] = track->x ;
				}
				if( track->y < trackBounds[TRACK_MIN_Y] )
				{
					trackBounds[TRACK_MIN_Y] = track->y ;
				}
				else if( track->y > trackBounds[TRACK_MAX_Y] )
				{
					trackBounds[TRACK_MAX_Y] = track->y ;
				}
			}
			track->number = n ;
			track->added( track ) ;
			prev = track ;
		}
	}

	if( err && ( first != NULL ) )
	{
		track = first ;
		while( track )
		{
			prev = track ;
			track = prev->nextTrack ;
			free( prev->length ) ;
			free( prev ) ;
		}
	}

	if( nTracks > 0 )
	{
		for( i = 0 ; i < nTracks ; i++ )
		{
			glDeleteLists( course[i]->drawList, MAX_LOD ) ;
			course[i]->free( course[i] ) ;
		}
		free( course ) ;
		nTracks = 0 ;
		course = NULL ;
		if( trackDisplayList != NULL )
		{
			free( trackDisplayList ) ;
		}
	}

	if( !err )
	{
		nTracks = n ;

		course = myMalloc( nTracks * sizeof( Track * ) ) ;
		trackDisplayList = myMalloc( nTracks *
						sizeof( *trackDisplayList ) ) ;

		track = first ;
		i = 0 ;
		while( track )
		{
			course[i] = track ;
			track = track->nextTrack ;
			i++ ;
		}

		/*
		 * Print message if track doesn't match up at ends.
		 */
		t = createStraightTrack( first->lanes, 100.0f ) ;
		course[nTracks-1]->add( course[nTracks-1], t ) ;
		if( ABSFUNC( t->x ) > 5.0f ||
			ABSFUNC( t->y ) > 5.0f )
		{
			minorError( NULL, NULL, "last track x,y = %.2f,%.2f\n",
					t->x, t->y ) ;
		}
		t->free( t ) ;

		/*
		 * Close up last track with first track.
		 */
		course[nTracks-1]->nextTrack = first ;
		raceLaps = cp->nLaps ;

		trackBounds[TRACK_CENTER_X] = 0.5f *
			( trackBounds[TRACK_MIN_X] + trackBounds[TRACK_MAX_X] );
		trackBounds[TRACK_CENTER_Y] = 0.5f *
			( trackBounds[TRACK_MIN_Y] + trackBounds[TRACK_MAX_Y] );
		trackBounds[TRACK_HALF_SIDE_X] = 0.5f *
			( trackBounds[TRACK_MAX_X] - trackBounds[TRACK_MIN_X] );
		trackBounds[TRACK_HALF_SIDE_Y] = 0.5f *
			( trackBounds[TRACK_MAX_Y] - trackBounds[TRACK_MIN_Y] );

		if( trackOutline == -1 )
		{
			trackOutline = glGenLists( 1 ) ;
		}

		glNewList( trackOutline, GL_COMPILE ) ;
		glColor3f( 1.0f, 1.0f, 1.0f ) ;
		drawTrackOutline() ;
		glEndList() ;

		if( currentRacePacket != NULL )
		{
			free( currentRacePacket ) ;
		}
		currentRacePacket = (outCoursePacket *)myMalloc(
						sizeof( outCoursePacket ) ) ;
		bcopy( cp, currentRacePacket, sizeof( outCoursePacket ) ) ;
	}
	else
	{
		nTracks = 0 ;
	}

	return( err ) ;
}



/*------------------------------------------------------------------------------
 * Load a track from a course packet id (network mode).
 *----------------------------------------------------------------------------*/
int
loadTrackFromId(
	int	courseId
	)
{
	outCoursePacket	*cp ;
	
	if( ( cp = findCourseInListById( courseId ) ) == NULL )
	{
		return( -1 ) ;
	}

	return( loadTrackFromPacket( cp ) ) ;
}



/*------------------------------------------------------------------------------
 * Draw all the track pieces.
 *----------------------------------------------------------------------------*/
void
drawTrack(
	void
	)
{
	int	i ;
	int	j ;
	int	l ;
	int	n = 0 ;
	int	end ;
	Track	*t ;

	for( i = 0 ; i < nTracks ; i++ )
	{
		t = course[i] ;
		if( t->cullDetail == NULL )
		{
			glCallList( t->drawMarkerList + baseLod ) ;
			trackDisplayList[n].t = t ;
			trackDisplayList[n].lod = baseLod ;
			n++ ;
		}
		else if( ( l = qcCull( t->center, t->cullDetail ) )
			!= QC_CULLED )
		{
			glCallList( t->drawMarkerList + l ) ;
			trackDisplayList[n].t = t ;
			trackDisplayList[n].lod = l ;
			n++ ;
		}
	}

	glColor3f( 0.4f, 0.4f, 0.4f ) ;
	for( i = 0 ; i < n ; i++ )
	{
		glCallList( trackDisplayList[i].t->drawList +
				trackDisplayList[i].lod ) ;
	}
}



/*------------------------------------------------------------------------------
 * Return a pointer to the first track piece.
 *----------------------------------------------------------------------------*/
Track *
getTrack(
	int	n
	)
{
	if( course == NULL )
	{
		fatalError( "getTrack: track not loaded yet" ) ;
	}
	else
	{
		return( course[ n % nTracks ] ) ;
	}
}



/*------------------------------------------------------------------------------
 * Build a selection list for courses.
 *----------------------------------------------------------------------------*/
static XmStringTable
buildCourseList(
	struct dirent	**courseNames
	)
{
	int		i ;
	XmStringTable	list ;

	list = (XmStringTable)myMalloc( nCourses * sizeof( XmString ) ) ;

	for( i = 0 ; i < nCourses ; i++ )
	{
		list[i] = XmStringCreateLocalized(
				myBaseName( courseNames[i]->d_name ) ) ;
	}

	return( list ) ;
}



/*------------------------------------------------------------------------------
 * Free up the space used in the course list.
 *----------------------------------------------------------------------------*/
static void
freeCourseList(
	)
{
	int	n ;

	for( n = 0 ; n < nCourses ; n++ )
	{
		XmStringFree( courseList[n] ) ;
	}

	free( courseList ) ;

	courseList = NULL ;
}



/*------------------------------------------------------------------------------
 * Determine the number of available courses.
 *----------------------------------------------------------------------------*/
static void
lookForCourses(
	struct dirent	***courseNames
	)
{
	trackDir = defaultDataDir ;

	recDir = defaultRecDir ;

	trackDirLen = strlen( trackDir ) ;

	recDirLen = strlen( recDir ) ;

	nCourses = scandir( trackDir, courseNames, courseSelect, alphasort ) ;

	if( nCourses < 0 )
	{
		fatalError( "could not scan current directory for courses!" ) ;
	}
	else if( nCourses == 0 )
	{
		fatalError( "No race courses found in the specified directory!"
			"  The\ncurrent search directory is `%s'.  Use\n\n"
			"       setenv VROOM_DIR course_directory\n\n"
			"to specify the location of the race courses.",
			trackDir ) ;
	}
}



/*------------------------------------------------------------------------------
 * Filter mechanism for track files.
 *----------------------------------------------------------------------------*/
static int
courseSelect(
	struct dirent	*file
	)
{
	int			l ;
	static struct stat	buf ;
	int			mode = S_IFREG | S_IREAD ;
	char			*fileName ;

	l = strlen( file->d_name ) ;
	fileName = myMalloc( trackDirLen + l + 2 ) ;

	sprintf( fileName, "%s/%s", trackDir, file->d_name ) ;

	if( stat( fileName, &buf ) == -1 ||
		( buf.st_mode & mode ) != mode )
	{
		free( fileName ) ;
		return( 0 ) ;
	}
	else
	{
		free( fileName ) ;
		if( l > 4 )
		{
			return( !strcmp( ".vct", file->d_name+l-4 ) ) ;
		}
		else
		{
			return( 0 ) ;
		}
	}
}



/*------------------------------------------------------------------------------
 * Apply the course selection.
 *----------------------------------------------------------------------------*/
static void
coursePreviewCB(
	Widget		w,
	Widget		list,
	XtPointer	callData
	)
{
	int	l ;
	char	*courseFile ;
	char	*courseName ;
	char	*description ;
	char	rec[2][VROOM_NAMELEN+10] ;
	int	errSt ;

	if( ( courseFile = getListSelectedString( list ) ) == NULL )
	{
		minorError( NULL, NULL,
				"Error determining course file name." ) ;
	}
	else
	{
		courseName = myMalloc( trackDirLen + strlen( courseFile ) + 6 );
		sprintf( courseName, "%s/%s.vct", trackDir, courseFile ) ;

		if( ( errSt = loadTrack( courseName ) ) != 0 )
		{
			minorError( NULL, NULL,
				trackLoadErrorMsgs[errSt],
				courseName ) ;
			free( courseName ) ;
			free( courseFile ) ;
			return ;
		}
		else
		{
			description = myMalloc( strlen( courseFile ) + 40 +
						sizeof( rec ) ) ;
			sprintf( rec[0], "%s %s", longRaceTimeString(
				currentRacePacket->recLap, &l ),
				currentRacePacket->recLapOwner ) ;
			sprintf( rec[1], "%s %s", longRaceTimeString(
				currentRacePacket->recRace, &l ),
				currentRacePacket->recRaceOwner ) ;
			sprintf( description, "%s - %d laps\n"
				"Lap Record: %s\nRace Record: %s",
				courseFile, raceLaps, rec[0], rec[1] ) ;
			coursePreview( description, localCourseForm ) ;
			free( description ) ;
		}
		free( courseFile ) ;
		free( courseName ) ;
	}
}



/*------------------------------------------------------------------------------
 * Apply the course selection.
 *----------------------------------------------------------------------------*/
static void
courseSelectionCB(
	Widget		w,
	Widget		list,
	XtPointer	callData
	)
{
	char		*courseFile ;
	char		*courseName ;
	int		errSt ;

	busyCursor() ;

	if( ( courseFile = getListSelectedString( list ) ) == NULL )
	{
		minorError( NULL, NULL,
				"Error determining course file name." ) ;
	}
	else
	{
		courseName = myMalloc( trackDirLen + strlen( courseFile ) + 6 );
		sprintf( courseName, "%s/%s.vct", trackDir, courseFile ) ;
		free( courseFile ) ;

		if( ( errSt = loadTrack( courseName ) ) == 0 )
		{
			freeCourseList() ;
			free( courseName ) ;

			startTimeTrials( 0 ) ;
		}
		else
		{
			minorError( NULL, NULL,
				trackLoadErrorMsgs[errSt],
				courseName ) ;
			free( courseName ) ;
		}

	}

	unbusyCursor() ;
}



/*------------------------------------------------------------------------------
 * Make the chooser for selecting a local course (solo mode).
 *----------------------------------------------------------------------------*/
void
createLocalCourseSelector(
	Widget	parent
	)
{
	int		n ;
	Arg		args[10] ;
	Widget		w ;
	Widget		list ;
	Widget		courseSelector ;
	Widget		button[4] ;
	XmString	str ;
	struct dirent	**courseNames ;
	static int	init = 0 ;
	static char	*helpFiles[] = { "localcourse.help", NULL } ;

	if( init )
	{
		return ;
	}
	init = 1 ;

	lookForCourses( &courseNames ) ;

	courseList = buildCourseList( courseNames ) ;

	n = 0 ;
	XtSetArg( args[n], XmNtopAttachment, XmATTACH_FORM ) ; n++ ;
	XtSetArg( args[n], XmNleftAttachment, XmATTACH_FORM ) ; n++ ;
	XtSetArg( args[n], XmNrightAttachment, XmATTACH_FORM ) ; n++ ;
	XtSetArg( args[n], XmNbottomAttachment, XmATTACH_FORM ) ; n++ ;
	createSelectionList( "Select a course:", parent, &list, 4,
				button, courseSelLabels, args, n ) ;
	XmListDeleteAllItems( list ) ;
	XmListAddItems( list, courseList, nCourses, 0 ) ;
	XmListSelectPos( list, 1, True ) ;

	XtAddCallback( list, XmNdefaultActionCallback,
		(XtCallbackProc)courseSelectionCB, list ) ;
	XtAddCallback( button[0], XmNactivateCallback,
		(XtCallbackProc)courseSelectionCB, list ) ;
	XtAddCallback( button[1], XmNactivateCallback,
		(XtCallbackProc)coursePreviewCB, list ) ;
	XtAddCallback( button[2], XmNactivateCallback,
		(XtCallbackProc)exitCB, NULL ) ;
	XtAddCallback( button[3], XmNactivateCallback,
		(XtCallbackProc)helpCB, helpFiles ) ;

	freeCourseNames( courseNames ) ;
}



/*------------------------------------------------------------------------------
 * Make the chooser for selecting a course from the server.
 *----------------------------------------------------------------------------*/
Widget
createServerCourseSelector(
	Widget	parent
	)
{
	int		n ;
	Arg		args[10] ;
	Widget		w ;
	Widget		button[4] ;
	XmString	str ;
	static Widget	courseSelector ;
	static int	init = 0 ;
	static char	*helpFiles[] = { "netcourse.help", NULL } ;

	if( init )
	{
		XmListDeleteAllItems( courseSelector ) ;
		return( courseSelector ) ;
	}
	init = 1 ;

	str = XmStringCreateLocalized( " " ) ;
	n = 0 ;
	XtSetArg( args[n], XmNlabelString, str ) ; n++ ;
	XtSetArg( args[n], XmNalignment, XmALIGNMENT_CENTER ) ; n++ ;
	XtSetArg( args[n], XmNtopAttachment, XmATTACH_POSITION ) ; n++ ;
	XtSetArg( args[n], XmNtopPosition, 80 ) ; n++ ;
	XtSetArg( args[n], XmNleftAttachment, XmATTACH_FORM ) ; n++ ;
	XtSetArg( args[n], XmNrightAttachment, XmATTACH_FORM ) ; n++ ;
	XtSetArg( args[n], XmNbottomAttachment, XmATTACH_FORM ) ; n++ ;
	serverCourseLabel = XtCreateManagedWidget( "label", xmLabelWidgetClass,
					parent, args, n ) ;
	XmStringFree( str ) ;

	n = 0 ;
	XtSetArg( args[n], XmNorientation, XmHORIZONTAL ) ; n++ ;
	XtSetArg( args[n], XmNtopAttachment, XmATTACH_NONE ) ; n++ ;
	XtSetArg( args[n], XmNleftAttachment, XmATTACH_FORM ) ; n++ ;
	XtSetArg( args[n], XmNrightAttachment, XmATTACH_FORM ) ; n++ ;
	XtSetArg( args[n], XmNbottomAttachment, XmATTACH_WIDGET ) ; n++ ;
	XtSetArg( args[n], XmNbottomWidget, serverCourseLabel ) ; n++ ;
	w = XtCreateManagedWidget( "separator", xmSeparatorWidgetClass, parent,
					args, n ) ;

	n = 0 ;
	XtSetArg( args[n], XmNtopAttachment, XmATTACH_FORM ) ; n++ ;
	XtSetArg( args[n], XmNleftAttachment, XmATTACH_FORM ) ; n++ ;
	XtSetArg( args[n], XmNrightAttachment, XmATTACH_FORM ) ; n++ ;
	XtSetArg( args[n], XmNbottomAttachment, XmATTACH_WIDGET ) ; n++ ;
	XtSetArg( args[n], XmNbottomWidget, w ) ; n++ ;
	createSelectionList( "Select a course:", parent, &courseSelector, 4,
				button, courseSelLabels, args, n ) ;
	XmListDeleteAllItems( courseSelector ) ;
/*
	XmListAddItems( courseSelector, courseList, nCourses, 0 ) ;
	XmListSelectPos( courseSelector, 1, True ) ;
*/

	XtAddCallback( courseSelector, XmNdefaultActionCallback,
		(XtCallbackProc)serverCourseSelectionCB, courseSelector ) ;
	XtAddCallback( button[0], XmNactivateCallback,
		(XtCallbackProc)serverCourseSelectionCB, courseSelector ) ;
	XtAddCallback( button[1], XmNactivateCallback,
		(XtCallbackProc)serverCoursePreviewCB, courseSelector ) ;
	XtAddCallback( button[2], XmNactivateCallback,
		(XtCallbackProc)exitCB, NULL ) ;
	XtAddCallback( button[3], XmNactivateCallback,
		(XtCallbackProc)helpCB, helpFiles ) ;

	return( courseSelector ) ;
}



/*------------------------------------------------------------------------------
 * Add a server course to the selection list.
 *----------------------------------------------------------------------------*/
void
addCourseToSelectionList(
	Widget	list,
	char	*name
	)
{
	XmString	str ;
	int		*positionList ;
	int		positionCount ;

	str = XmStringCreateLocalized( name ) ;
	XmListAddItem( list, str, 0 ) ;
	XmStringFree( str ) ;

	if( !XmListGetSelectedPos( list, &positionList, &positionCount ) )
	{
		XmListSelectPos( list, 1, True ) ;
	}
	else
	{
		XtFree( (char *)positionList ) ;
	}
}



/*------------------------------------------------------------------------------
 * Scan for all courses and load up for the server.
 *----------------------------------------------------------------------------*/
int
readServerCourses(
	CourseNode	**courseHead
	)
{
	int		i ;
	int		courseCount = 0 ;
	CourseNode	*course = NULL ;
	CourseNode	*newCourse ;
	struct dirent	**courseNames ;

	*courseHead = NULL ;

	lookForCourses( &courseNames ) ;

	for( i = 0 ; i < nCourses ; i++ )
	{
		if( ( newCourse = addServerCourse( courseNames[i]->d_name ) )
			!= NULL )
		{
			if( course )
			{
				course->next = newCourse ;
			}
			else
			{
				*courseHead = newCourse ;
			}
			course = newCourse ;
			newCourse->packet->courseId = courseCount ;
			courseCount++ ;
		}
		
	}

	freeCourseNames( courseNames ) ;

	return( courseCount ) ;
}



/*------------------------------------------------------------------------------
 * Add a course to the server list.
 *----------------------------------------------------------------------------*/
static CourseNode *
addServerCourse(
	char	*fileName
	)
{
	outCoursePacket	*packet ;
	CourseNode	*node ;
	char		*courseName ;
	
	courseName = myMalloc( trackDirLen + strlen( fileName ) + 2 );
	sprintf( courseName, "%s/%s", trackDir, fileName ) ;

	if( ( packet = loadTrackIntoPacket( courseName ) ) != NULL )
	{
		node = (CourseNode *)myMalloc( sizeof( CourseNode ) ) ;
		node->packet = packet ;
		node->next = NULL ;
	}
	else
	{
		node = NULL ;
	}

	free( courseName ) ;
	
	return( node ) ;
}



/*------------------------------------------------------------------------------
 * Free memory allocated for course names during directory scan.
 *----------------------------------------------------------------------------*/
static void
freeCourseNames(
	struct dirent	**courseNames
	)
{
	int	i ;

	for( i = 0 ; i < nCourses ; i++ )
	{
		free( courseNames[i] ) ;
	}

	free( courseNames ) ;

	nCourses = 0 ;
}



/*------------------------------------------------------------------------------
 * Callback for previewing a server course.
 *----------------------------------------------------------------------------*/
static void
serverCoursePreviewCB(
	Widget		w,
	Widget		list,
	XtPointer	callData
	)
{
	int		l ;
	char		*courseName ;
	char		*description ;
	int		errSt ;
	char		rec[2][VROOM_NAMELEN+10] ;
	outCoursePacket	*course ;

	if( ( courseName = getListSelectedString( list ) ) == NULL )
	{
		minorError( NULL, NULL,
				"Error determining course file name." ) ;
	}
	else
	{
		course = findCourseInList( courseName ) ;

		if( course == NULL )
		{
			minorError( NULL, NULL, "Could not find selected "
					"course `%s' in list.", courseName ) ;
		}

		if( ( errSt = loadTrackFromPacket( course ) ) != 0 )
		{
			minorError( NULL, NULL,
				trackLoadErrorMsgs[errSt],
				courseName ) ;
		}
		else
		{
			description = myMalloc( strlen( courseName ) + 40 +
						sizeof( rec ) ) ;
			sprintf( rec[0], "%s %s", longRaceTimeString(
				currentRacePacket->recLap, &l ),
				currentRacePacket->recLapOwner ) ;
			sprintf( rec[1], "%s %s", longRaceTimeString(
				currentRacePacket->recRace, &l ),
				currentRacePacket->recRaceOwner ) ;
			sprintf( description, "%s - %d laps\n"
				"Lap Record: %s\nRace Record: %s",
				courseName, raceLaps, rec[0], rec[1] ) ;
			coursePreview( description, serverCourseForm ) ;
			free( description ) ;
		}

		XtFree( courseName ) ;
	}
}



/*------------------------------------------------------------------------------
 * Draw the track outline.
 *----------------------------------------------------------------------------*/
static void
drawTrackOutline(
	void
	)
{
	int	n ;

	glBegin( GL_LINE_STRIP ) ;
	for( n = 0 ; n < nTracks ; n++ )
	{
		course[n]->drawOutline( course[n], TRACK_LEFT_SIDE ) ;
	}
	glEnd() ;

	glBegin( GL_LINE_STRIP ) ;
	for( n = 0 ; n < nTracks ; n++ )
	{
		course[n]->drawOutline( course[n], TRACK_RIGHT_SIDE ) ;
	}
	glEnd() ;
}



/*------------------------------------------------------------------------------
 * Return the array of info on the track boundaries (approximate).
 *----------------------------------------------------------------------------*/
void
getTrackBounds(
	GLfloat	bnds[8]
	)
{
	bnds[0] = trackBounds[0] ;
	bnds[1] = trackBounds[1] ;
	bnds[2] = trackBounds[2] ;
	bnds[3] = trackBounds[3] ;
	bnds[4] = trackBounds[4] ;
	bnds[5] = trackBounds[5] ;
	bnds[6] = trackBounds[6] ;
	bnds[7] = trackBounds[7] ;
}



/*------------------------------------------------------------------------------
 * Load a outCoursePacket with a race course.
 *----------------------------------------------------------------------------*/
static outCoursePacket *
loadTrackIntoPacket(
	char	*fileName
	)
{
	int		i ;
	int		k ;
	int		dp ;
	int		n ;
	int		f ;
	int		nLaps ;
	int		err = 0 ;
	char		line[80] ;
	outCoursePacket	*packet ;
	
	packet = (outCoursePacket *)myMalloc( sizeof( outCoursePacket ) ) ;

	strncpy( packet->name, myBaseName( fileName ), sizeof( packet->name ) );

	if( ( f = open( fileName, O_RDONLY ) ) < 0 )
	{
		minorError( NULL, NULL, "%s: %s.", fileName,
			strerror( errno ) ) ;
		free( packet ) ;
		return( NULL ) ;
	}
	else if( read( f, &nLaps, sizeof( nLaps ) ) != sizeof( nLaps ) )
	{
		minorError( NULL, NULL, "%s: bad header.", fileName ) ;
		free( packet ) ;
		close( f ) ;
		return( NULL ) ;
	}
	else if( read( f, &packet->size, sizeof( packet->size ) ) !=
		sizeof( packet->size ) )
	{
		minorError( NULL, NULL, "%s: bad header.", fileName ) ;
		free( packet ) ;
		close( f ) ;
		return( NULL ) ;
	}

	packet->nLaps = nLaps ;

	for( n = 0 ; !err && n < packet->size ; n++ )
	{
		k = n % 4 ;

		if( k == 0 )
		{
			dp = n / 4 ;
			if( read( f, &packet->detail[dp],
				sizeof( TrackDetailQuad ) ) !=
				sizeof( TrackDetailQuad ) )
			{
				err = 1 ;
				break ;
			}
		}
	}

	if( packet->detail[0].id[0] != STRAIGHT_TRACK )
	{
		minorError( NULL, NULL, "%s: first track piece is not"
				" a straight piece.", fileName ) ;
		err = 1 ;
	}

	close( f ) ;
	if( err )
	{
		free( packet ) ;
		packet = NULL ;
	}
	else
	{
		loadTrackRecords( packet ) ;
	}

	return( packet ) ;
}



/*------------------------------------------------------------------------------
 * Create a quickcull detail structure for a piece of track.
 *----------------------------------------------------------------------------*/
qcDetail
trackDetail(
	float	radius
	)
{
	return( qcCreateDetail( 0, MAX_LOD - 1, 1000.0f, 4000.0f, radius ) ) ;
}



/*------------------------------------------------------------------------------
 * Set the minimum LOD for all tracks.
 *----------------------------------------------------------------------------*/
void
setTrackBaseLod(
	int	lod
	)
{
	int	i ;

	for( i = 0 ; i < nTracks ; i++ )
	{
		if( course[i]->cullDetail )
		{
			qcSetLodRange( course[i]->cullDetail, lod,
					MAX_LOD - 1 ) ;
		}
	}
}



/*------------------------------------------------------------------------------
 * Calculate the centrifugal force on a car, given the speed and radius of
 * curvature.  A return value of 0 mean okay.  A return value of 1 or greater
 * means a wipe out.  A value in between means skidding.
 *----------------------------------------------------------------------------*/
float
centrifugalForce(
	float	speed,
	float	radius
	)
{
	float		f ;

	f = speed * speed / radius ;

	if( f < MIN_SKID_FORCE )
	{
		return( 0.0f ) ;
	}
	else
	{
		return( ( f - MIN_SKID_FORCE ) * SKID_FORCE_DIV ) ;
	}
}



/*------------------------------------------------------------------------------
 * Return the maximum speed at a given radius of curvature.
 *----------------------------------------------------------------------------*/
float
maxSpeed(
	float	radius
	)
{
	if( radius > 0.0f )
	{
		return( sqrtf( MAX_SKID_FORCE * radius ) ) ;
	}
	else
	{
		return( TOP_SPEED * 1.5f ) ;
	}
}



/*------------------------------------------------------------------------------
 * No-op function for entering a track segment.
 *----------------------------------------------------------------------------*/
void
enterTrack(
	Track	*track,
	Car	*car
	)
{
}



/*------------------------------------------------------------------------------
 * Strip the directory and extension off of a string.
 *----------------------------------------------------------------------------*/
static char *
myBaseName(
	char	*fullName
	)
{
	int		i = 0 ;
	char		*e ;
	char		*s ;
	static char	name[256] ;

	if( ( s = strrchr( fullName, '/' ) ) == NULL )
	{
		s = fullName ;
	}
	else
	{
		s++ ;
	}

	if( ( e = strrchr( s, '.' ) ) == NULL )
	{
		return( fullName ) ;
	}
	else
	{
		i = (int)( e - s ) / sizeof( char ) ;
		if( i > sizeof( name ) - 1 )
		{
			i = sizeof( name ) / sizeof( char ) ;
		}
		strncpy( name, s, i ) ;
		name[i] = '\0' ;
		return( name ) ;
	}
}



/*------------------------------------------------------------------------------
 * Get the name of the track record file.
 *----------------------------------------------------------------------------*/
static char *
getTrackRecordFileName(
	char	*courseName
	)
{
	char	*recName ;

	recName = myMalloc( recDirLen + strlen( courseName ) + 6 );
	sprintf( recName, "%s/%s.rec", recDir, courseName ) ;

	return( recName ) ;
}



/*------------------------------------------------------------------------------
 * Load up the track records.
 *----------------------------------------------------------------------------*/
static void
loadTrackRecords(
	outCoursePacket	*packet
	)
{
	int	f ;
	float	tLap ;
	float	tRace ;
	char	nLap[VROOM_NAMELEN] ;
	char	nRace[VROOM_NAMELEN] ;
	char	*recName ;

	packet->recLap = MAX_RACE_TIME ;
	packet->recRace = MAX_RACE_TIME ;
	strcpy( packet->recLapOwner, "nobody" ) ;
	strcpy( packet->recRaceOwner, packet->recLapOwner ) ;

	if( ( recName = getTrackRecordFileName( packet->name ) ) == NULL )
	{
		return ;
	}

	if( ( f = open( recName, O_RDONLY ) ) == -1 )
	{
		free( recName ) ;
		return ;
	}

	if( read( f, &tLap, sizeof( tLap ) ) == sizeof( tLap ) &&
		read( f, nLap, sizeof( nLap ) ) == sizeof( nLap ) &&
		read( f, &tRace, sizeof( tRace ) ) == sizeof( tRace ) &&
		read( f, nRace, sizeof( nRace ) ) == sizeof( nRace ) )
	{
		if( tLap < MAX_RACE_TIME )
		{
			packet->recLap = tLap ;
			strncpy( packet->recLapOwner, nLap,
				sizeof( packet->recLapOwner ) ) ;
		}
		if( tRace < MAX_RACE_TIME )
		{
			packet->recRace = tRace ;
			strncpy( packet->recRaceOwner, nRace,
				sizeof( packet->recRaceOwner ) ) ;
		}
	}

	close( f ) ;
	free( recName ) ;
}



/*------------------------------------------------------------------------------
 * Save the track records.
 *----------------------------------------------------------------------------*/
static void
saveTrackRecords(
	outCoursePacket	*packet
	)
{
	int	f ;
	char	*recName ;

	if( ( recName = getTrackRecordFileName( packet->name ) ) == NULL )
	{
		return ;
	}

	if( ( f = open( recName, O_WRONLY | O_CREAT | O_TRUNC, 0666 ) ) == -1 )
	{
		perror( recName ) ;
		printf("/var/IndiZone/vroom does not exist!  You must either use inst to install vroom, or create /var/IndiZone/vroom as root in order to save high scores.\n");
		free( recName ) ;
		return ;
	}

	if( write( f, &packet->recLap, sizeof( packet->recLap ) )
		!= sizeof( packet->recLap ) ||
		write( f, packet->recLapOwner, sizeof( packet->recLapOwner ) )
		!= sizeof( packet->recLapOwner ) ||
		write( f, &packet->recRace, sizeof( packet->recRace ) )
		!= sizeof( packet->recRace ) ||
		write( f, packet->recRaceOwner, sizeof( packet->recRaceOwner ) )
		!= sizeof( packet->recRaceOwner ) )
	{
		perror( recName ) ;
	}

	free( recName ) ;
	close( f ) ;
}



/*------------------------------------------------------------------------------
 * Check if qualifying lap sets a new record.
 *----------------------------------------------------------------------------*/
int
checkLapRecord(
	float	t,
	char	*name,
	int	update
	)
{
	outCoursePacket	*cp ;

	if( t < currentRacePacket->recLap )
	{
		currentRacePacket->recLap = t ;
		strncpy( currentRacePacket->recLapOwner, name,
			sizeof( currentRacePacket->recLapOwner ) ) ;
		if( update )
		{
			saveTrackRecords( currentRacePacket ) ;
		}
		if( playMode == VROOM_TEAM )
		{
	
			if( ( cp = findCourseInListById(
				currentRacePacket->courseId ) ) != NULL )
			{
				cp->recLap = t ;
				strncpy( cp->recLapOwner, name,
					sizeof( cp->recLapOwner ) ) ;
			}
		}
		return( 1 ) ;
	}
	else
	{
		return( 0 ) ;
	}
}



/*------------------------------------------------------------------------------
 * Check (and save) for new race record.
 *----------------------------------------------------------------------------*/
int
checkRaceRecord(
	float	t,
	char	*name,
	int	update
	)
{
	outCoursePacket	*cp ;

	if( t < currentRacePacket->recRace )
	{
		currentRacePacket->recRace = t ;
		strncpy( currentRacePacket->recRaceOwner, name,
			sizeof( currentRacePacket->recRaceOwner ) ) ;
		if( update )
		{
			saveTrackRecords( currentRacePacket ) ;
		}
		if( playMode == VROOM_TEAM )
		{
	
			if( ( cp = findCourseInListById(
				currentRacePacket->courseId ) ) != NULL )
			{
				cp->recRace = t ;
				strncpy( cp->recRaceOwner, name,
					sizeof( cp->recRaceOwner ) ) ;
			}
		}
		return( 1 ) ;
	}
	else
	{
		return( 0 ) ;
	}
}



/*------------------------------------------------------------------------------
 * Return the selected character string of a list widget.
 *----------------------------------------------------------------------------*/
char *
getListSelectedString(
	Widget	list
	)
{
	int		n ;
	Arg		args[2] ;
	char		*s ;
	XmStringTable	selections ;
	int		nSelections ;

	n = 0 ;
	XtSetArg( args[n], XmNselectedItems, &selections ) ; n++ ;
	XtSetArg( args[n], XmNselectedItemCount, &nSelections ) ; n++ ;
	XtGetValues( list, args, n ) ;

	if( nSelections > 0 && selections != NULL )
	{
		s = extractNormalString( selections[0] ) ;
	}
	else
	{
		s = NULL ;
	}

	return( s ) ;
}



/*------------------------------------------------------------------------------
 * Return the track lap record time.
 *----------------------------------------------------------------------------*/
float
getTrackLapRecordTime(
	void
	)
{
	return( currentRacePacket->recLap ) ;
}



/*------------------------------------------------------------------------------
 * Return the track race record time.
 *----------------------------------------------------------------------------*/
float
getTrackRaceRecordTime(
	void
	)
{
	return( currentRacePacket->recRace ) ;
}
