/*
 * Copyright (C) 1994, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */
#ifndef _TRACK_H_
#define _TRACK_H_

#include <GL/gl.h>
#include "car.h"
#include "server.h"
#include "quickcull.h"

#define	LANE_CHANGE_SPEED (4.0f)

#define	SIDE_MARKER_LENGTH	(100.0f)
#define	SIDE_MARKER_WIDTH	(20.0f)

#define	TRACK_RIGHT_SIDE	0
#define	TRACK_LEFT_SIDE		1

typedef void	(*TrackFollowProc)( struct _track *t, Car *c, float dt ) ;
typedef void	(*TrackAddProc)( struct _track *o, struct _track *n ) ;
typedef void	(*TrackAddedProc)( struct _track *t ) ;
typedef float	(*TrackSpeedProc)( struct _track *t, float s, float l ) ;
typedef void	(*StartPositionProc)( struct _track *t, Car *c, int place ) ;
typedef void	(*EndPositionProc)( struct _track *t, Car *c, int place ) ;
typedef void	(*FreeTrackProc)( struct _track *t ) ;
typedef void	(*DrawOutlineProc)( struct _track *t, int side ) ;
typedef float	(*TrackLenProc)( struct _track *t, float lane ) ;
typedef void	(*TrackEnterProc)( struct _track *t, Car *c ) ;
typedef float	(*LaneWidthProc)( struct _track *t, float s, float lane1,
                    float lane2 ) ;

typedef struct _track {
	int			lanes ;
	int			number ;
	int			type ;
	struct _track		*nextTrack ;
	GLuint			drawList ;
	GLuint			drawMarkerList ;
	float			x ;
	float			y ;
	float			theta ;
	float			midLane ;
	float			*length ;
	float			c[20] ;
	float			center[3] ;
	qcDetail		cullDetail ;
	TrackFollowProc		follow ;
	TrackAddProc		add ;
	TrackAddedProc		added ;
	TrackSpeedProc		topSpeed ;
	StartPositionProc	startPosition ;
	EndPositionProc		endPosition ;
	FreeTrackProc		free ;
	DrawOutlineProc		drawOutline ;
	TrackLenProc		trackLength ;
	TrackEnterProc		enter ;
	LaneWidthProc		laneWidth ;
	} Track ;

typedef struct _courseNode {
	outCoursePacket		*packet ;
	struct _courseNode	*next ;
	} CourseNode ;


#define	STRAIGHT_TRACK		(0)
#define	L_CURVE_TRACK		(1)
#define	R_CURVE_TRACK		(2)
#define	MERGE_IN_TRACK		(3)
#define	MERGE_OUT_TRACK		(4)

/* BEGIN PROTOTYPES track.c */
void        addCourseToSelectionList( Widget list, char *name ) ;
float       centrifugalForce( float speed, float radius ) ;
int         checkLapRecord( float t, char *name, int update ) ;
int         checkRaceRecord( float t, char *name, int update ) ;
void        createLocalCourseSelector( Widget parent ) ;
Widget      createServerCourseSelector( Widget parent ) ;
void        drawTrack( void ) ;
void        enterTrack( Track *track, Car *car ) ;
char *      getListSelectedString( Widget list ) ;
Track *     getTrack( int n ) ;
void        getTrackBounds( GLfloat bnds[8] ) ;
float       getTrackLapRecordTime( void ) ;
float       getTrackRaceRecordTime( void ) ;
int         loadTrack( char *fileName ) ;
int         loadTrackFromId( int courseId ) ;
int         loadTrackFromPacket( outCoursePacket *cp ) ;
float       maxSpeed( float radius ) ;
int         readServerCourses( CourseNode **courseHead ) ;
void        setTrackBaseLod( int lod ) ;
qcDetail    trackDetail( float radius ) ;
/* END PROTOTYPES track.c */

#endif /* !_TRACK_H_ */
